// FILE: CalCmd.c
// COPYRIGHT: (c), Symmetric Research, 2006-2008
//
// Source file for the USBxCH calibration CMD application

#if defined( SROS_WINXP ) || defined( SROS_WIN2K )
#include <windows.h>
#include <process.h>    // needed for _spawnlp (spawn a new process)
#elif defined( SROS_LINUX )
#endif  // SROS_xxxxx

#include <stdio.h>
#include <stdlib.h>
#include <string.h>              // strcpy

#include "CalLib.h"              // Common Cal defines
#include "SrIniParse.h"          // for PrepareFullMsg, SetErrorIniFunction etc
#include "SrHelper.h"            // for SrKeyboardNonBlock, SrGetKeyWait, SrFlushPrint, etc


// Defines and global variables

#define CAL_MENU_MIN     0
#define CAL_MENU_HELP    1
#define CAL_MENU_SHOW    2
#define CAL_MENU_EDIT    3
#define CAL_MENU_TITLE   4
#define CAL_MENU_UNITS   5
#define CAL_MENU_FORMAT  6
#define CAL_MENU_SAVE    7
#define CAL_MENU_QUIT    8
#define CAL_MENU_MAX     9

char *CalMenuText[CAL_MENU_MAX] = {
        "Menu entry 0",
        "Show help text in a separate window",
        "Show current calibration data",
        "Edit calibration data for one channel",
        "Enter title strings for all channels",
        "Enter units strings for all channels",
        "Enter display formats for all channels",
        "Save calibration data to a file",
        "Exit"
        };

char TempStr[MAXSTR];
int  UnsavedChanges;
HANDLE EditorHelpHandle;
        


// Prototypes

void ShowMenu( void );
int  ReadMenuChoice( void );
int  DoShowHelp( void );
int  DoDisplayResults( void );
int  DoEditChannel( void );
int  AskForChannel( void );
void AskForTitleText( int Channel );
void AskForUnitText( int Channel );
void AskForUnitFormat( int Channel );
void AskForValue( int Channel, int AtodInput );
void MeasureValue( int Channel, int AtodInput );
int  DoAskForTitles( void );
int  DoAskForUnits( void );
int  DoAskForFormat( void );
int  DoSaveCalibration( void );
int  DoQuit( void );

void OpenFileInEditor( char *editor, char *filename,
                       HANDLE *EditorHandle, char *ErrString );
int  IsEditorClosed( HANDLE EditorHandle );

void ParseErrorMessage( char *msg );
void ParseWarnMessage( char *msg );
void DisplayLogMessage( int MsgLevel, int Fatal, char *ExtraMsg );

int CleanExit( int ErrLevel );


//------------------------------------------------------------------------------
// ROUTINE: CalCmd Main
// PURPOSE: The main program for the command line version of CAL, the USBxCH
//          calibration program.  This program provides a menu letting the user
//          perform various calibration related functions such as editing the
//          identifying data for a channel and measuring new calibrations values, 
//          viewing the results, and saving them to a file.
//------------------------------------------------------------------------------
int main( int argc, char *argv[] ) {
        
        int   Choice, Success;
        char *CmdLine;



        // Set keyboard behavior to allow for checking a key

        SrKeyboardMode( KEYMODE_NONBLOCK );


        // Initialize parser message functions
        
        SetErrorIniFunction( ParseErrorMessage );
        SetWarnIniFunction( ParseWarnMessage );


        // Save command line argument

        if ( argc == 2 )
                CmdLine = argv[1];
        else
                CmdLine = NULL;


        // Initialize

        if ( !DvmInit( DisplayLogMessage, CmdLine, "CalCmd" ) )  CleanExit( 1 );
        CalInit( );
        UnsavedChanges   = 0;
        EditorHelpHandle = NULL;


        // Open A/D in preparation for perform calibration functions
        
        printf( "Beginning calibration program ...\n" );

        
        // Run menu loop

        while( 1 ) {

                ShowMenu( );
                Choice = ReadMenuChoice( );

                switch ( Choice ) {

                        case CAL_MENU_HELP:
                                Success = DoShowHelp( );
                                break;

                        case CAL_MENU_SHOW:
                                Success = DoDisplayResults( );
                                break;

                        case CAL_MENU_EDIT:
                                Success = DoEditChannel( );
                                break;

                        case CAL_MENU_TITLE:
                                Success = DoAskForTitles( );
                                break;

                        case CAL_MENU_UNITS:
                                Success = DoAskForUnits( );
                                break;

                        case CAL_MENU_FORMAT:
                                Success = DoAskForFormat( );
                                break;

                        case CAL_MENU_SAVE:
                                Success = DoSaveCalibration( );
                                break;

                        case CAL_MENU_QUIT:
                                Success = DoQuit( );
                                break;

                        default:
                                printf( "\nInvalid menu choice.  Please try again.\n" );
                        }

                } // end while 1

        return( 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: ShowMenu
// PURPOSE: Print menu with available actions.
//------------------------------------------------------------------------------
void ShowMenu( void ) {

        int i;

        printf( "\nWhat would you like to do next?\n\n" );

        for ( i = CAL_MENU_MIN+1 ; i < CAL_MENU_MAX ; i++ ) 
                printf( "  %d. %s\n", i, CalMenuText[i] );

        printf( "\n" );
}

//------------------------------------------------------------------------------
// ROUTINE: ReadMenuChoice
// PURPOSE: Read and check users menu choice.
//------------------------------------------------------------------------------
int ReadMenuChoice( void ) {

        int Choice, Nscan;

        printf( "Enter number of desired menu option: " );   SrFlushPrint( );
        SrGetLine( TempStr, MAXSTR );
        Nscan = sscanf( TempStr, "%d", &Choice );
        if ( ( Nscan != 1 )             ||
             ( Choice <= CAL_MENU_MIN ) ||
             ( Choice >= CAL_MENU_MAX ) )
                return( CAL_MENU_MIN );
        else
                return( Choice );
}




//------------------------------------------------------------------------------
// ROUTINE: DoShowHelp
// PURPOSE: Show help screen.
//------------------------------------------------------------------------------
int DoShowHelp( void ) {

#if defined( SROS_WINXP ) || defined( SROS_WIN2K )

        OpenFileInEditor( DvmInfo.EditorName,
                          "\\SR\\USBXCH\\DVM\\Calibrate\\CalHelp.txt",
                          &EditorHelpHandle,
                          "CalHelp.txt" );
#else
        printf( "****      OpenFileInEditor not available under Linux yet       ****\n" );
        printf( "**** Please cat /usr/local/SR/USBXCH/DVM/Calibrate/CalHelp.txt ****\n" );

#endif // SROS_xxx

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: DoDisplayResults
// PURPOSE: Print the calibration results on the screen.
//------------------------------------------------------------------------------
int DoDisplayResults( void ) {

        int  i;
        char SampleFormat[MAXSTR];

        // Display calibration info

        for ( i = 0 ; i < DvmInfo.AtodAnalogChannels ; i++ ) {

                CalMakeFormatSample( DvmInfo.Channel[i].UnitsPlaces,
                                     DvmInfo.Channel[i].UnitsDigits,
                                     SampleFormat, MAXSTR-1 );
        
                printf("\nCalibration for channel %d = %s:\n",
                                                       i, DvmInfo.Channel[i].Title );
                printf("   Input A ... Counts %10ld -> %s = %lg\n",
                                                       CalInfo.UserCountA[i],
                                                       DvmInfo.Channel[i].UnitsName,
                                                       CalInfo.UserValueA[i] );
                printf("   Input B ... Counts %10ld -> %s = %lg\n",
                                                       CalInfo.UserCountB[i],
                                                       DvmInfo.Channel[i].UnitsName,
                                                       CalInfo.UserValueB[i] );
                printf("   Display format is %d.%d with Places = %d, Digits = %d, eg %s\n",
                                                       DvmInfo.Channel[i].UnitsPlaces,
                                                       DvmInfo.Channel[i].UnitsDigits,
                                                       DvmInfo.Channel[i].UnitsPlaces,
                                                       DvmInfo.Channel[i].UnitsDigits,
                                                       SampleFormat );
                printf("   Slope = %lg, Offset = %lg\n",
                                                       DvmInfo.Channel[i].UnitsSlope,
                                                       DvmInfo.Channel[i].UnitsOffset );
                }


        printf( "\nPress a key to return to menu\n" );
        SrGetKeyWait( );
        
        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: DoEditChannel
// PURPOSE: Prompt user to edit the calibration data for a channel.  This
//          involves selecting the desired channel, providing the title and
//          unit text, and the units format, and finally entering two values 
//          and measuring their input signals.
//------------------------------------------------------------------------------
int DoEditChannel( void ) {

        int i, Channel;

        Channel = AskForChannel( );
        if ( Channel == -1 ) {
                printf( "Invalid channel.  Please select again.\n" );
                return( 0 );
                }

        AskForTitleText( Channel );
        AskForUnitText( Channel );
        AskForUnitFormat( Channel );

        printf( "\nSpecify and measure %d input values for Channel %d (%s in %s) ...\n",
                  NINPUTS, Channel,
                  DvmInfo.Channel[Channel].Title, DvmInfo.Channel[Channel].UnitsName );
        for ( i = 0 ; i < NINPUTS ; i++ ) {
                AskForValue( Channel, i );
                MeasureValue( Channel, i );
                }

        UnsavedChanges = 1;

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: AskForChannel
// PURPOSE: Prompt user to select the desired channel.
//------------------------------------------------------------------------------
int AskForChannel( void ) {
        
        int Channel, Nscan, Count;

        Channel = 0;
        Count   = 0;

        // Prompt for and read in channel number

        while ( Count < 3 ) {

                printf( "\nEnter number of channel to edit (0 to %d): ",
                          DvmInfo.AtodAnalogChannels-1 );  SrFlushPrint( );
                SrGetLine( TempStr, MAXSTR );
                Nscan = sscanf( TempStr, "%d", &Channel );

                // If channel is valid, then return it

                if ( ( Nscan == 1 )   &&
                     ( Channel >= 0 ) &&
                     ( Channel < DvmInfo.AtodAnalogChannels ) ) {

                        return( Channel );
                        }

                // Otherwise give warning and allow another try

                else {
                        printf( "Invalid channel number.  Must be between 0 and %d\n",
                                  DvmInfo.AtodAnalogChannels-1 );
                        Count++;
                        }

                } // end while Count < 3

        return( -1 );
}

//------------------------------------------------------------------------------
// ROUTINE: AskForTitleText
// PURPOSE: Prompt user to enter the display title (eg Voltage, Rheostat, etc)
//          for the specified channel.
//------------------------------------------------------------------------------
void AskForTitleText( int Channel ) {
        
        int i, len;

        i = Channel;
        
        // Prompt for and read in title

        printf( "\nEnter channel %d title (Voltage, Temperature, etc) [%s]: ",
                  i, DvmInfo.Channel[i].Title );  SrFlushPrint( );
        SrGetLine( TempStr, MAXSTR );
        len = strlen( TempStr );


        // Check for exit request

        if ( len == 1 && TempStr[0] == 'q' )
                CleanExit( 1 );

        // Use read in string

        else if ( len > 0 )
                strcpy( DvmInfo.Channel[i].Title, TempStr );

        // Else len == 0 so use original (default) value

}

//------------------------------------------------------------------------------
// ROUTINE: AskForUnitText
// PURPOSE: Prompt user to enter the calibration units (eg Volts, Inches, etc)
//          for the specified channel.
//------------------------------------------------------------------------------
void AskForUnitText( int Channel ) {
        
        int i, len;

        i = Channel;
        
        // Prompt for and read in units

        printf( "\nEnter channel %d units (Volts, Ohms, etc) [%s]: ",
                  i, DvmInfo.Channel[i].UnitsName );  SrFlushPrint( );
        SrGetLine( TempStr, MAXSTR );
        len = strlen( TempStr );

        // Check for exit request

        if ( len == 1 && TempStr[0] == 'q' )
                CleanExit( 1 );

        // Use read in string

        else if ( len > 0 )
                strcpy( DvmInfo.Channel[i].UnitsName, TempStr );

        // Else len == 0 so use original (default) value
}

//------------------------------------------------------------------------------
// ROUTINE: AskForUnitFormat
// PURPOSE: Prompt user to enter the units places and digits format numbers.  
//          Places specifies the total number of places (including the decimal) 
//          and digits specifies the number of digits after the decimal.  For
//          example, entering 7.2 means the data values would be displayed in
//          1234.12 format.
//------------------------------------------------------------------------------
void AskForUnitFormat( int Channel ) {
        
        int i, len, Place, Digit, Nscan;

        i = Channel;
        
        // Prompt for and read in units format places.digits

        printf( "\nEnter channel %d (%s in %s) format (7.5, 6.2, etc) [%d.%d]: ",
                  i, DvmInfo.Channel[i].Title, DvmInfo.Channel[i].UnitsName,
                  DvmInfo.Channel[i].UnitsPlaces, DvmInfo.Channel[i].UnitsDigits );
                 SrFlushPrint( );
        SrGetLine( TempStr, MAXSTR );
        len = strlen( TempStr );

        // Check for exit request

        if ( len == 1 && TempStr[0] == 'q' )
                CleanExit( 1 );

        // Use read in string

        else if ( len > 0 ) {
                Nscan = sscanf( TempStr, "%d.%d", &Place, &Digit );
                if ( (Nscan == 2) && DvmFormatIsValid( Place, Digit ) ) {
                        DvmInfo.Channel[i].UnitsPlaces = Place;
                        DvmInfo.Channel[i].UnitsDigits = Digit;
                        }
                else
                        printf( "Bad format string entered.  Using default of %d.%d.\n",
                                 DvmInfo.Channel[i].UnitsPlaces,
                                 DvmInfo.Channel[i].UnitsDigits );
                }

        // Else len == 0 so use original (default) value
}

//------------------------------------------------------------------------------
// ROUTINE: AskForValue
// PURPOSE: Prompt user to specify the values of the known USBxCH inputs
//          which will be used in the calibration computation.
//------------------------------------------------------------------------------
void AskForValue( int Channel, int AtodInput ) {

        int    i, Nscan;
        double Value;
        char  *InputName[] = {"A", "B"};

        i = Channel;

        printf( "\nEnter channel %d input value %s in %s: ",
                  Channel, InputName[AtodInput], DvmInfo.Channel[i].UnitsName );
        SrFlushPrint( );
        
        SrGetLine( TempStr, MAXSTR );
        Nscan = sscanf( TempStr, "%lg", &Value );
        
        if ( AtodInput == INPUT_A )
                CalInfo.UserValueA[i] = Value;
        else
                CalInfo.UserValueB[i] = Value;

}

//------------------------------------------------------------------------------
// ROUTINE: MeasureValue
// PURPOSE: Prompt user to press a key before measuring raw A/D counts for a
//          known USBxCH input.
//------------------------------------------------------------------------------
void MeasureValue( int Channel, int AtodInput ) {

        int Ans, Ok;

        // Wait for a keypress before beginning measurement
        
        printf( "\nHook up input and press a key to measure signal\n" );
        Ans = SrGetKeyWait( );
        if (Ans == 'q')
                return;

        // Measure the raw counts for the current A/D input

        printf("Measurement beginning ...\n");
        Ok = CalMeasure( Channel, AtodInput );
        printf("Measurement done ...\n");

        if ( !Ok )
                DvmInfo.ShowLogMessage( MSGLEVEL_ERR, ERROR_FATAL, 
                                "Failed to measure values ...  Exiting now" );
}

//------------------------------------------------------------------------------
// ROUTINE: DoAskForTitles
// PURPOSE: Prompt user to specify the display titles (eg Voltage,Rheostat,etc).
//------------------------------------------------------------------------------
int DoAskForTitles( void ) {
        
        int i;
        
        for ( i = 0 ; i < DvmInfo.AtodAnalogChannels ; i++ )

                AskForTitleText( i );

        UnsavedChanges = 1;

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: DoAskForUnits
// PURPOSE: Prompt user to specify the calibration units (eg Volts, Inches, etc).
//------------------------------------------------------------------------------
int DoAskForUnits( void ) {
        
        int i;
        
        for ( i = 0 ; i < DvmInfo.AtodAnalogChannels ; i++ )

                AskForUnitText( i );

        UnsavedChanges = 1;

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: DoAskForFormat
// PURPOSE: Prompt user to specify the units format (eg 7.5, 6.2, etc).
//------------------------------------------------------------------------------
int DoAskForFormat( void ) {
        
        int i;
        
        for ( i = 0 ; i < DvmInfo.AtodAnalogChannels ; i++ )

                AskForUnitFormat( i );

        UnsavedChanges = 1;

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: DoSaveCalibration
// PURPOSE: Save calibration data to dvmcal.ini file.
//------------------------------------------------------------------------------
int DoSaveCalibration( void ) {

        CalCompute( );
        CalIniOutput( );
        UnsavedChanges = 0;

        printf( "\nCalibration results saved in file %s\n", CalInfo.CalFileName );

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: DoQuit
// PURPOSE: Close A/D and exit after checking for unsaved changes.
//------------------------------------------------------------------------------
int DoQuit( void ) {

        int WantToQuit;


        // Initialize

        WantToQuit = 1;



        if ( !IsEditorClosed( EditorHelpHandle ) ) {  // Verify editor is closed
                printf( "\nCalHelp.txt is still running!  Do you really want to quit (Y or N)? " );
                SrFlushPrint( );
                SrGetLine( TempStr, MAXSTR );
                if ( TempStr[0] != 'y' && TempStr[0] != 'Y' )
                        WantToQuit = 0;
                }
        
        
        // If there are unsaved changes, verify that the user really wants to quit

        if ( UnsavedChanges == 1 ) {
                printf( "\nUnsaved changes!  Do you really want to quit (Y or N)? " );
                SrFlushPrint( );
                SrGetLine( TempStr, MAXSTR );
                if ( TempStr[0] != 'y' && TempStr[0] != 'Y' )
                        WantToQuit = 0;
                }


        // Really quit

        if ( WantToQuit ) {
                printf( "Exiting ...\n" );
                CleanExit( 0 );
                }

        else {
                printf( "\nPlease close help file or save changes before quitting.\n" );
                }

        return( WantToQuit );
}

//------------------------------------------------------------------------------
// ROUTINE: OpenFileInEditor
// PURPOSE: To open the specified file in the specified editor program.
//------------------------------------------------------------------------------
void OpenFileInEditor( char *editor, char *filename,
                       HANDLE *EditorHandle, char *ErrString ) {

#if defined( SROS_WINXP ) || defined( SROS_WIN2K )
        int      error;
        intptr_t hEditor;
        DWORD    ExitCode;
        BOOL     Ok;


        // Ensure we are only working with one editor open at a time

        if ( EditorHandle != NULL ) {
                ExitCode = STILL_ACTIVE;
                Ok = GetExitCodeProcess( *EditorHandle, &ExitCode );

                // Tell user to use existing editor

                if ( Ok && (ExitCode == STILL_ACTIVE) ) {
                        printf( "CalHelp.txt is still running.  Please use existing version\nor close it down and try again.\n" );
                        return;
                        }

                // Clean up remains of previous editor

                else {
                        CloseHandle( *EditorHandle );
                        *EditorHandle = NULL;
                        }
                }

        
        // Start editor showing help file

        error   = 0;
        hEditor = _spawnlp( _P_NOWAIT, editor, editor, filename, NULL );

        if ( hEditor == -1 ) {
                _get_errno( &error );
                sprintf( TempStr, "Could not open %s with editor %s\n(error# %d)", filename, editor, error );
                DisplayLogMessage( MSGLEVEL_WARN, ERROR_NOT_FATAL, TempStr );
                }
        else {
                if (EditorHandle)  *EditorHandle = (HANDLE)hEditor;
                }
        
#else
//FIX   printf( "******* OpenFileInEditor not available under Linux yet ********\n" );

#endif // SROS_xxx
        
}

//------------------------------------------------------------------------------
// ROUTINE: IsEditorClosed
// PURPOSE: Check to be sure editor isn't running.  Return 1 if editor is
//          closed or the user has said it is ok anyway.
//------------------------------------------------------------------------------
int IsEditorClosed( HANDLE EditorHandle ) {

#if defined( SROS_WINXP ) || defined( SROS_WIN2K )
        BOOL  Ok;
        DWORD ExitCode;
        
        // Check to be sure editor isn't running

        if ( EditorHandle != NULL ) {
                ExitCode = STILL_ACTIVE;
                Ok = GetExitCodeProcess( EditorHandle, &ExitCode );
                if ( Ok && (ExitCode == STILL_ACTIVE) ) {
                        printf( "IsEditorClosed Exit STILL running\n" );
                        return( 0 );
                        }
                else {
                        if ( !Ok )
                                printf( "Editor failed to get exit code\n" );
                        // continue to exit anyway
                        }
                }

#else
        printf( "************** OpenFileInEditor not available under Linux yet ***********************\n" );

#endif // SROS_xxx

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: ParseErrorMessage
// PURPOSE: Error message used by parsing functions.
//------------------------------------------------------------------------------
void ParseErrorMessage( char *msg ) {
        
        char *ExtraMsg;
        ExtraMsg = PrepareFullMsg( msg );
        DisplayLogMessage( MSGLEVEL_ERR, ERROR_FATAL, ExtraMsg );
}

//------------------------------------------------------------------------------
// ROUTINE: ParseWarnMessage
// PURPOSE: Warning message used by parsing functions.
//------------------------------------------------------------------------------
void ParseWarnMessage( char *msg ) {
        
        char *ExtraMsg;
        ExtraMsg = PrepareFullMsg( msg );
        DisplayLogMessage( MSGLEVEL_WARN, ERROR_NOT_FATAL, ExtraMsg );
}

//------------------------------------------------------------------------------
// ROUTINE: DisplayLogMessage
// PURPOSE: General message facility.
//------------------------------------------------------------------------------
void DisplayLogMessage( int MsgLevel, int Fatal, char *ExtraMsg ) {

        // Prepare log message
        
        DvmLogMessagePrep( MsgLevel, ExtraMsg );

        

        // Output the log message
        
        printf( DvmInfo.LogMessage );



        // If error is fatal, then stop A/D and quit
        
        if ( Fatal == ERROR_FATAL ) {
        
                if ( DvmInfo.IsRunning == 1 )
                        DvmAtodStop( );
        
                CleanExit( 1 );
                }
}

//------------------------------------------------------------------------------
// ROUTINE: CleanExit
// PURPOSE: Leave the program, while ensuring the keyboard mode is restored.
//------------------------------------------------------------------------------
int CleanExit( int ErrLevel ) {
        
        SrKeyboardMode( KEYMODE_BLOCK );  // Restore blocking keyboard behavior.

        exit( ErrLevel );
}

