// FILE: CalLib.c
// COPYRIGHT: (c), Symmetric Research, 2007-2009
//
// Source file for the USBxCH calibration application shared functions.

#if defined( SROS_WINXP ) || defined( SROS_WIN2K )
#include <windows.h>
#endif // SROS_xxx

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "CalLib.h"     // Common Cal defines
#include "SrHelper.h"   // SrSleep and other OS dependent functions



// Global variables all collected up together in one structure

CAL_VARIABLES CalInfo;
char          TempStr[MAXSTR];

//------------------------------------------------------------------------------
// ROUTINE: CalInit
// PURPOSE: Initialize the CalInfo structure with reasonable values based on
//          the calibration slope + offset in the DvmInfo structure.
//------------------------------------------------------------------------------
void CalInit( void ) {

        int i;

        strcpy(  CalInfo.CalFileName, CAL_DEFAULT_INI );
        
        for ( i = 0 ; i < DvmInfo.AtodAnalogChannels ; i++ ) {

                CalInfo.UserCountA[i] = 0;
                CalInfo.UserCountB[i] = 8000000;

                CalInfo.UserValueA[i]  = DvmInfo.Channel[i].UnitsOffset;
                CalInfo.UserValueB[i]  = DvmInfo.Channel[i].UnitsSlope *
                                         CalInfo.UserCountB[i] +
                                         DvmInfo.Channel[i].UnitsOffset;
                }
}

//------------------------------------------------------------------------------
// ROUTINE: CalMeasure
// PURPOSE: Acquire raw count data and save it in the CalInfo structure.
//          Once two inputs have been measured, the calibration slope + offset
//          results can be computed.
//------------------------------------------------------------------------------
int CalMeasure( int Channel, int AtodInput ) {

        int i, Nsamples, Ok, ReadTry;

printf("CalMeasure Enter with timer interval %d\n", DvmInfo.TimerInterval );

        // Call Dvm library functions to acquire raw count data
        
        DvmDataInit( );


        // Open the A/D driver and start acquiring data
        
        Ok = DvmAtodOpen( );  if ( !Ok )  return( Ok );
        DvmAtodStart( );

        
        ReadTry = 0;
        i       = 0;

#define MAXTRY 5        
        while ( ReadTry < MAXTRY ) {

printf("CalMeasure read try %d starting sleep\n", ReadTry );

                SrSleep( DvmInfo.TimerInterval * 2 ); // wait twice as long as it should take

printf("CalMeasure read try %d finished sleep\n", ReadTry );

                Ok = DvmAtodRead( &Nsamples );
printf("CalMeasure DvmAtodRead result %d, nsamples %d\n", Ok, Nsamples );

                if ( Ok == 1 ) {       // all good
                        DvmDataFilter( Nsamples );
                        break;
                        }
                else if ( Ok == -1 ) { // no data yet
                        ReadTry++;
                        }
                else if ( Ok == 0 ) {  // failure
                        DvmAtodStop( );
                        return( Ok );
                        }
                else { // Ok == ? {    // unknown
                        ReadTry++;
                        }
                }


        // Stop A/D and close its driver

        DvmAtodStop( );
        DvmAtodClose( );

printf("CalMeasure read result %d after %d tries\n", Ok, ReadTry );

        // Save the acquired data in the CalInfo structure

        if ( AtodInput == INPUT_A )
                CalInfo.UserCountA[Channel] = DvmInfo.DataCounts[Channel];
        else
                CalInfo.UserCountB[Channel] = DvmInfo.DataCounts[Channel];

printf("CalMeasure Exit\n" );
        return( Ok );
}

//------------------------------------------------------------------------------
// ROUTINE: CalCompute
// PURPOSE: Compute the slope and offset calibration results based on the
//          provided user values and measured raw counts for two inputs.
//------------------------------------------------------------------------------
int CalCompute( void ) {

        int i, Ok;


printf("CalCompute Enter\n" );

        // Indicate if divide by zero prevention adjustment done
        
        Ok = 1;


        // Compute calibration

        for ( i = 0 ; i < DvmInfo.AtodAnalogChannels ; i++ ) {

                if ( CalInfo.UserCountB[i] == CalInfo.UserCountA[i] ) {
                        Ok = 0;
                        CalInfo.UserCountB[i]++;
                        }
                
                DvmInfo.Channel[i].UnitsSlope  =
                        (CalInfo.UserValueB[i] - CalInfo.UserValueA[i]) /
                        (CalInfo.UserCountB[i] - CalInfo.UserCountA[i]);
                DvmInfo.Channel[i].UnitsOffset =
                        CalInfo.UserValueB[i] -
                        DvmInfo.Channel[i].UnitsSlope * CalInfo.UserCountB[i];
                }

printf("CalCompute Exit\n" );
        return( Ok );
}

//------------------------------------------------------------------------------
// ROUTINE: CalIniOutput
// PURPOSE: Output the computed calibration results to a default ini file.
//          This is done using the DvmIniWrite function, so this file is 
//          guaranteed to be in the format the DVM program requires.  To use 
//          the output ini file with DVM, just copy it to the DVM directory and 
//          rename it.
//------------------------------------------------------------------------------
int CalIniOutput( void ) {

        FILE* FileCal;
        int   i;

        // Open calibration output file
        
        FileCal = fopen( CalInfo.CalFileName, "w" );
        if ( !FileCal ) {
                DvmInfo.ShowLogMessage( MSGLEVEL_ERR, ERROR_NOT_FATAL,
                                        "Could not open calibration output file" );
                return( 0 );
                }


        // Write it
        
        DvmIniWrite( FileCal, INIPART_GENERAL, 0,
                     "; DVM Initialization Data Output by CalGui or CalCmd\n\n", NULL, "\n" );
        
        for ( i = 0 ; i < DvmInfo.AtodAnalogChannels ; i++ ) {
                sprintf( TempStr, "; Parameters for Channel %d\n\n", i );
                DvmIniWrite( FileCal, INIPART_DISPLAY, i,
                             TempStr, NULL, "\n" );
                }


        // Close it

        fclose( FileCal );

        return( 1 );
}
//------------------------------------------------------------------------------
// ROUTINE: CalMakeFormatSample
// PURPOSE: Use the units places and digits format information to generate a
//          string with a number using this format.
//------------------------------------------------------------------------------
int CalMakeFormatSample( int Places, int Digits, char *DataString, int Nchar ) {

        int  i, len, Before, After;
        char DigitChars[] = "01234567890123456789"; // extras in case Places > 9


        // Error check

        if ( DataString == NULL )
                return( 0 );

        len = strlen( DigitChars );
        
        // Check for invalid values

        if ( !DvmFormatIsValid( Places, Digits ) ||
             Places > len                        ||
             Places > Nchar-1                    ) {
                if ( Nchar > 7 )
                        strcpy( DataString, "Invalid" );
                else if ( Nchar > 4 )
                        strcpy( DataString, "Bad" );
                return( 0 );
                }
        

        // Initialize

        DataString[0] = '\0';

        Before = Places - 1 - Digits;
        After  = Digits;


        // Fill string

        for ( i = 0 ; i < Before ; i++ )
                DataString[i] = DigitChars[Places-i];
        
        if ( Digits > 0 )
                DataString[i] = '.';
        else
                DataString[i] = DigitChars[Places-i];
                

        for ( i = 0 ; i < After ; i++ )
                DataString[Before+i+1] = DigitChars[i+1];
        
        DataString[Places] = '\0';

        return( 1 );
}
