// FILE: DvmCmd.c
// COPYRIGHT: (c), Symmetric Research, 2009-2010
//
// Source file for DvmCmd, a simple command line version of a Digital
// Volt Meter application written in C/C++ to acquire data from a USBxCH
// A/D connected to a USB port.
//
// DvmCmd displays sample values on the screen and writes them out to an
// ASCII text file.  An .ini initialization file is used to specify various
// the settings such as the name of the output file and the names and scaling
// constants for each channel.

#if defined( SROS_WINXP ) || defined( SROS_WIN2K )
#include <windows.h>
#elif defined( SROS_LINUX )
#endif  // SROS_xxxxx

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "DvmLib.h"             // Common DVM defines
#include "SrIniParse.h"         // for PrepareFullMsg, SetErrorIniFunction etc
#include "SrHelper.h"           // for SrKeyboardNonBlock, SrGetKeyWait, SrSleep, etc


// Program copyright and screen title:

char *Copyright =

          "\n\nDvmCmd.exe : Copyright (c), Symmetric Research, 2009-2010\n\n";

char *ScreenTitle =

      "USBxCH DvmCmd [/?]: (c) SR 2009 : Rev {2010/08/30}\n\n";


// Defines and global variables

#define MAX_DATA_LINES 20

char TempStr[MAXSTR];
int  NumLines;


// Message arrays

char *UsageMsg[] = {
        "Acquires and saves 1 Hz data from the command line         ",
        "                                                           ",
        "USAGE: DvmCmd [parameters.ini] [/?]                        ",
        "                                                           ",
        "    parameters.ini = file with program settings            ",
        "                /? = show help screen and quit             ",
        "                                                           ",
        "Examples:                                                  ",
        "                                                           ",
        "    DvmCmd /?                 < show help screen and quit  ",
        "    DvmCmd DvmSetupVolts.ini  < run with Volts settings    ",
        "                                                           ",
        "See DvmHelpIniSyntax.txt for descriptions of the allowed   ",
        "parameter settings.                                        ",
        "                                                           ",

        NULL,
        };


// Prototypes

int  CheckForQuit( void );
void DrawTitles( void );
void UpdateDisplay( void );
int  OfferRemoveOutFile( void );

int  ProcessCommandLine( int argc, char **argv );
void Usage( char *errmsg );

void ParseErrorMessage( char *msg );
void ParseWarnMessage( char *msg );
void DisplayLogMessage( int MsgLevel, int Fatal, char *ExtraMsg );

int  CleanExit( int ErrLevel );



//------------------------------------------------------------------------------
// ROUTINE: DvmCmd main
// PURPOSE: The main program for the command line version DVM, the USBxCH
//          Digital Volt Meter acquistion program.  This program reads input
//          parameters from the file dvm.ini and acquires and displays data
//          until ENTER is pressed.
//------------------------------------------------------------------------------
int main( int argc, char *argv[] ) {
        
        int   Nsamples, IniArg;
        char *IniFileName;


        // Announce program

        printf( ScreenTitle );



        // Set keyboard behavior to allow for checking a key

        SrKeyboardMode( KEYMODE_NONBLOCK );



        // Initialize .ini parser message functions
        
        SetErrorIniFunction( ParseErrorMessage );
        SetWarnIniFunction( ParseWarnMessage );



        // Get command line arguments

        IniArg = ProcessCommandLine( argc, argv );

        if ( IniArg > 0 )
                IniFileName = argv[IniArg];
        else
                IniFileName = NULL;

        

        // Initialize from .ini file, create output file header, and show titles
        
        NumLines = MAX_DATA_LINES + 1;

        if ( !DvmInit( DisplayLogMessage, IniFileName, "DvmCmd" ) )
                CleanExit( 1 );

        if ( !DvmFileHeaderWrite( ) ) {        // Save header to out file
                if ( !OfferRemoveOutFile() )   // remove old file
                        CleanExit( 1 );
                if ( !DvmFileHeaderWrite( ) )  // try save again
                        CleanExit( 1 );
                }

        DrawTitles(  );



        // Open driver, initialize A/D, and start running

        if ( !DvmAtodOpen() )                          CleanExit( 1 );
        DvmAtodStart( );



        // Begin acquisition loop after getting a keypress

        printf("\nPress ENTER to start acquisition, press %s to quit\n", SR_QENTER );
        if ( SrGetKeyWait() == 'q' ) // allow instant exit
                CleanExit( 1 );

        while ( 1 ) {

                // Wait a little for data to accumulate

                SrSleep( DvmInfo.TimerInterval - DVM_TIMER_LESS );


                // Acquire a sample

                if ( !DvmAtodRead( &Nsamples ) )  break; // Read more data
                DvmDataFilter( Nsamples );               // Update results w/ new data


                // Do something with that sample

                if ( !DvmFileDataWrite() )        break; // Save data to outfile
                UpdateDisplay( );                        // Show updated data


                // See if the user wants to continue
                
                if ( CheckForQuit() )             break; // Found 'q' keypress to exit
                
                } // end while 1


        // Stop A/D and close driver

        DvmAtodStop();
        DvmAtodClose();


        CleanExit( 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: CheckForQuit
// PURPOSE: Check if the user has asked to quit by pressing the 'q' key.  If so,
//          return 1.  Otherwise return 0.  If the user pressed a key, but it
//          was NOT a 'q', then give them a warning before returning 0.
//------------------------------------------------------------------------------
int CheckForQuit( void ) {

        int c;

        c = SrGetKeyCheck();

        // No key pressed, return 0

        if ( c == EOF )
                return( 0 );

        // A 'q' was pressed, return 1

        else if ( c == 'q' || c == 'Q' ) {

                printf( "\n ... Keypress 'q' abort ... \n");
                return( 1 );
                }

        // Some unknown key was pressed, return 0
        else {
                printf( "To abort press %s ... \n", SR_QENTER );
                return( 0 );
                }
}

//------------------------------------------------------------------------------
// ROUTINE: DrawTitles
// PURPOSE: Print the constant header titles to the screen.
//------------------------------------------------------------------------------
void DrawTitles( void ) {

        int   i;
        FILE *OutFp;


        // Output ini name

        printf( "Input file = %s\n", DvmInfo.IniFileNameShort );


        // Update titles

        DvmTitleInit();

        
        // Draw constant header titles

        for ( i = 0 ; i < TITLE_MAX ; i++ )
                printf( "%s\n", DvmInfo.Title[i] );
        printf( "\n" );

        
        // Draw general and per display header info

        OutFp = stdout;
                
        if ( DvmInfo.OutFileShowHeader == INISWITCH_ON ) {

                DvmIniWrite( OutFp, INIPART_GENERAL, 0, NULL, "  ", NULL );
                
                for ( i = 0 ; i < DvmInfo.AtodAnalogChannels ; i++ )
                        DvmIniWrite( OutFp, INIPART_DISPLAY, i, NULL, "  ", NULL );
                }
}

//------------------------------------------------------------------------------
// ROUTINE: UpdateDisplay
// PURPOSE: This is called each time new data is acquired to update the display.
//          Print out data titles in single sample mode or after a page of data.
//          Then print data values out to the screen.  
//------------------------------------------------------------------------------
void UpdateDisplay( void ) {

        int    i, PowerGoodH;
        double TempC, TempF;
        char   TitleFmt[MAXSTR];


        // Initialize variables
        
        PowerGoodH = SrUsbXchIsPowerBitGood( DvmInfo.AtodPowerInfo,
                                             USBXCH_POWER_HISTORY );
        
        SrDatUsbTemperatureCompute( DvmInfo.AtodTemperature, &TempC, &TempF );


        

        // Draw channel names and units titles at start of new page

        if ( NumLines >= MAX_DATA_LINES ) {

                NumLines    = 0;
                TitleFmt[0] = '\0';
                
                // First line contains sample number, channel titles,
                // digital, power, and temperature columns

                for ( i = 0 ; i < DvmInfo.AtodAnalogChannels ; i++ ) {
                        sprintf( TempStr, " %10s", DvmInfo.Channel[i].Title );
                        strcat( TitleFmt, TempStr );
                        }

#ifdef SHOW_ALL_DVMINFO
                printf( "  Sample #%s D0 D1 D2 D3 PwrH DegC   DegF   Time in Seconds     Time as YMD HMS    Frac Sec\n",
                          TitleFmt );
#else // SHOW_FIT_DVMINFO
                printf( "  Sample #%s D0 D1 D2 D3 PwrH DegC\n", TitleFmt );
#endif // SHOW_ALL_DVMINFO


                // Second line contains channel units

                sprintf( TitleFmt, "          " );

                for ( i = 0 ; i < DvmInfo.AtodAnalogChannels ; i++ ) {
                        sprintf( TempStr, " %10s", DvmInfo.Channel[i].UnitsName );
                        strcat( TitleFmt, TempStr );
                        }

                printf( "%s\n", TitleFmt );
                }

        NumLines++;



        // Draw sample number, data values, and time values if requested
        
        printf( "%10ld", DvmInfo.SampleNumber );
        
        for ( i = 0 ; i < DvmInfo.AtodAnalogChannels ; i++ ) {
                DvmDataFormat( i, TempStr, MAXSTR-1 );
                printf( " %10s", TempStr );
                }

        printf( "  %d  %d  %d  %d   ",
                   (DvmInfo.AtodDigitalIn & 0x01),
                  ((DvmInfo.AtodDigitalIn & 0x02)>>1),
                  ((DvmInfo.AtodDigitalIn & 0x04)>>2),
                  ((DvmInfo.AtodDigitalIn & 0x08)>>3) );

#ifdef SHOW_ALL_DVMINFO
        printf( "%1d %5.1lf  %5.1lf  %17.6lf  %4d/%02d/%02d %02d:%02d:%02d  0.%06ld\n",
                  PowerGoodH, TempC, TempF, DvmInfo.TimeSecSince1970,
                  DvmInfo.TimeYear, DvmInfo.TimeMonth, DvmInfo.TimeDay,
                  DvmInfo.TimeHour, DvmInfo.TimeMin,   DvmInfo.TimeSec,
                  DvmInfo.TimeUsec );
        
#else // SHOW_FIT_DVMINFO - just what fits
        printf( "%1d %5.1lf\n", PowerGoodH, TempC );
        
#endif // SHOW_ALL_DVMINFO        
}

//------------------------------------------------------------------------------
// ROUTINE: OfferRemoveOutFile
// PURPOSE: Offer to delete the existing output file, so a new one can be 
//          acquired.  Returns 1 if the file was removed, 0 otherwise.
//------------------------------------------------------------------------------
int OfferRemoveOutFile( void ) {

        int Answer;

        // Quick exit if not using output file

        if ( DvmInfo.OutFileOn == INISWITCH_OFF )
                return( 1 );


        // Offer user the chance to delete the old file

        printf( "\nDo you want to delete the existing out file named\n%s? (y or n)  ", DvmInfo.OutFileName );
        SrFlushPrint();
        Answer = SrGetKeyWait();
        printf( "\n" );


        // User wants to try deleting

        if ( Answer == 'y' || Answer == 'Y' ) {

                // Successful delete

                if ( remove( DvmInfo.OutFileName ) != -1 )
                        return( 1 );

                // Not deleted because delete failed
                
                printf( "Could not delete existing file.\nDelete from OS or check permissions and\nsharing and try again." );
                return( 0 );
                }

        // Not deleted because the user does not want to

        else if ( Answer == 'n' || Answer == 'N' )
                return( 0 );


        // Not deleted because the user didn't select a valid option

        else
                return( 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: ProcessCommandLine
// PURPOSE: Get the name of the ini file with the control settings from the
//          command line.
//------------------------------------------------------------------------------
int ProcessCommandLine( int argc, char *argv[] ) {

        int i, IniArg;


        // Check the number of command line arguments.

        if ( argc < 1 || argc > 3 )

                Usage( "wrong number of command line arguments" );

        // Process args

        IniArg = 0;
        for ( i = 1; i < argc ; i++ ) {
                if ( strcmp( argv[i], "/?" ) == 0 )         // help requested
                        Usage( "?" );
                else if ( IniArg == 0 )                     // ini name found
                        IniArg = i;
                else                                        // dup ini name
                        Usage( "Too many .ini filenames" );
                }

        return( IniArg );

}


//------------------------------------------------------------------------------
// ROUTINE: Usage
// PURPOSE: Print a message showing the command line options.
//------------------------------------------------------------------------------
void Usage( char *errmsg ) {

        int i;

        // Print the command line usage message and quit.

        // Select first string depending on
        // 1) no message, 2) help requested, or 3) message given

        if ( errmsg == NULL )
                printf( "\nERROR: Command line usage.\n\n" );

        else if ( strcmp( errmsg, "?" ) == 0 )
                ;

        else
                printf( "\nERROR: Command line usage (%s).\n\n", errmsg );


        for ( i = 0 ; UsageMsg[i] ; i++ )
                printf( "%s\n", Chop( UsageMsg[i] ) );

        CleanExit( 1 );
}


//------------------------------------------------------------------------------
// ROUTINE: ParseErrorMessage
// PURPOSE: Display error message when called by parsing functions.
//------------------------------------------------------------------------------
void ParseErrorMessage( char *msg ) {
        
        char *ExtraMsg;
        ExtraMsg = PrepareFullMsg( msg );
//        ExtraMsg += "\nDVM will exit";
        DisplayLogMessage( MSGLEVEL_ERR, ERROR_FATAL, ExtraMsg );
}

//------------------------------------------------------------------------------
// ROUTINE: ParseWarnMessage
// PURPOSE: Display warning message when called by parsing functions.
//------------------------------------------------------------------------------
void ParseWarnMessage( char *msg ) {
        
        char *ExtraMsg;
        ExtraMsg = PrepareFullMsg( msg );
        DisplayLogMessage( MSGLEVEL_WARN, ERROR_NOT_FATAL, ExtraMsg );
}

//------------------------------------------------------------------------------
// ROUTINE: DisplayLogMessage
// PURPOSE: General message function that does clean up and exits if fatal.
//------------------------------------------------------------------------------
void DisplayLogMessage( int MsgLevel, int Fatal, char *ExtraMsg ) {

        // Prepare log message
        
        DvmLogMessagePrep( MsgLevel, ExtraMsg );

        

        // Stop A/D execution
        
        if ( DvmInfo.IsRunning == 1 )
                DvmAtodStop();


        
        // Output the log message
        
        printf( DvmInfo.LogMessage );



        // If error is fatal, then quit
        
        if ( Fatal == ERROR_FATAL )
                CleanExit( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: CleanExit
// PURPOSE: Leave the program, while ensuring the keyboard mode is restored.
//------------------------------------------------------------------------------
int CleanExit( int ErrLevel ) {
        
        SrKeyboardMode( KEYMODE_BLOCK );  // Restore blocking keyboard behavior.

        exit( ErrLevel );
}
