// FILE: SrParse.c
// COPYRIGHT: (c), Symmetric Research, 2009
//
// INI FILE PARSING AND OUTPUT HELPER FUNCTIONS:
//
// These helper functions are used in parsing the .ini input file and
// displaying the .ini values.  The parsing functions are more
// complicated than simple fscanf to allow for free format syntax and
// better error reporting.
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "SrDefines.h"
#include "SrIniParse.h"


FILE *IniFp = NULL;

int fline = 1;

char TOKEN[MAXTOKEN], *token;
char SaveFileName[MAXTOKEN];
char msg[MAXTOKEN];
char ErrorMsgString[MAXTOKEN];

int ttype;

FNPTR ErrorIni = DefaultErrorIni; // Pointer to error function
FNPTR WarnIni  = DefaultWarnIni;  // Pointer to warning function


//------------------------------------------------------------------------------
// ROUTINE: DefaultErrorIni
// PURPOSE: Display an .ini error and quit.  This function is used as the
//          default error function if none was supplied by the caller.
//------------------------------------------------------------------------------
void DefaultErrorIni( char *msg ) {

        printf( "%s", PrepareFullMsg( msg ) );

        exit( 1 );

}

//------------------------------------------------------------------------------
// ROUTINE: DefaultWarnIni
// PURPOSE: Display an .ini warning.  This function is used as the default
//          warning function if none was supplied by the caller.
//------------------------------------------------------------------------------
void DefaultWarnIni( char *msg ) {

        printf( "%s", PrepareFullMsg( msg ) );
}



//------------------------------------------------------------------------------
// ROUTINE: IniOpen
// PURPOSE: Open an ini file for reading.
//------------------------------------------------------------------------------
FUNCTYPE( int ) IniOpen( char *IniFilename ) {

        // Save file name and open ini file
        
        strcpy( SaveFileName, IniFilename );
        
        IniFp = fopen( IniFilename, "r" );
        
        if ( !IniFp ) {
                ErrorIni( "unable to open .ini file" );
                return( 0 );
                }


        // Initialize line count and token
        
        fline    = 1;
        TOKEN[0] = '\0';

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: IniClose
// PURPOSE: Close the open ini file.
//------------------------------------------------------------------------------
FUNCTYPE( int ) IniClose( void ) {

        if ( fclose( IniFp ) == EOF ) {
                ErrorIni( "unable to close .ini file" );
                return( 0 );
                }

        IniFp = NULL;
        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: SetErrorIniFunction
// PURPOSE: Allow caller to specify the function to be called when a parsing
//          error occurs.
//------------------------------------------------------------------------------
FUNCTYPE( void ) SetErrorIniFunction( FNPTR NewErrorFunction ) {

        ErrorIni = NewErrorFunction;
        
}

//------------------------------------------------------------------------------
// ROUTINE: SetWarnIniFunction
// PURPOSE: Allow caller to specify the function to be called when a parsing
//          warning occurs.
//------------------------------------------------------------------------------
FUNCTYPE( void ) SetWarnIniFunction( FNPTR NewWarnFunction ) {

        WarnIni = NewWarnFunction;
        
}

//------------------------------------------------------------------------------
// ROUTINE: PrepareFullMsg
// PURPOSE: Prepare a full error message with file name, line num and user msg.
//------------------------------------------------------------------------------
FUNCTYPE( char* ) PrepareFullMsg( char *msg ) {

        sprintf( ErrorMsgString,
                 "ERROR[%s] : line %d : %s",
                 SaveFileName, fline, msg );

        return( ErrorMsgString );

}

////////////////////////////////////////////////////////////////////////////////
//                                                                            //
//                       GENERAL PARSING HELPER FUNCTIONS                     //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// ROUTINE: AddToToken
// PURPOSE: Add c to TOKEN, checking for token buffer overflow.
//------------------------------------------------------------------------------
FUNCTYPE( int ) AddToToken( int c ) {

        if ( (token - TOKEN) < MAXTOKEN )

                *token++ = c;

        else {
                ErrorIni( "internal token length exceeded" );
                ttype = ERROR;
                TOKEN[MAXTOKEN-1] = '\0';
                return( 0 );
                }

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: SkipSpace
// PURPOSE: Skip over white space, including comments.
//------------------------------------------------------------------------------
FUNCTYPE( int ) SkipSpace( int c ) {

skip:

        if ( isspace( c ) ) {

                if ( c == '\n' )

                        fline++;


                c = fgetc( IniFp );

                goto skip;

                }


        else if ( c == ';' ) { // ; comment ...

                while ( c != '\n' && c != EOF )

                        c = fgetc( IniFp );

                goto skip;

                }


        return( c );
}

//------------------------------------------------------------------------------
// ROUTINE: GetToken
// PURPOSE: Gather up next token and save it and its type.  Tokens include
//          things like strings, symbols, floats, integers, etc.
//------------------------------------------------------------------------------
FUNCTYPE( int ) GetToken( void ) {

        static int c = ' ';
        int ok;



        // Skip over any white space.

        c = SkipSpace( c );



        // Gather the next token.

        token = TOKEN;


        if ( isalpha( c ) || c == '_' || c == '/' ) {

                ttype = SYMBOL;

                while ( isalnum( c ) || c == '_' || c == '/' ) {

                        ok = AddToToken( c );  if ( !ok ) return( ERROR );
                        c  = fgetc( IniFp );
                        }
                }


        else if ( c == '=' ) {

                ttype = EQUALS;

                ok = AddToToken( c );  if ( !ok ) return( ERROR );
                c  = fgetc( IniFp );

                }


        else if ( isdigit( c ) || c == '-' || c == '+' ) {

                // ttype could be HEX, INTEGER, or FLOAT ...

                ok = AddToToken( c );  if ( !ok ) return( ERROR );
                c  = fgetc( IniFp );



                if ( TOKEN[0] == '0' && c == 'x' ) {

                        ttype = HEX;
                        token = TOKEN;

                        c = fgetc( IniFp );

                        if ( isxdigit( c ) ) {

                                while ( isxdigit( c ) ) {

                                        ok = AddToToken( c );  if ( !ok ) return( ERROR );
                                        c  = fgetc( IniFp );
                                        }
                                }

                        else

                                ttype = UNKNOWN;
                        }



                else {

                        ttype = INTEGER;

                        while ( isdigit( c ) ) {

                                ok = AddToToken( c );  if ( !ok ) return( ERROR );
                                c  = fgetc( IniFp );
                                }



                        if ( c == '.' ) {

                                ttype = FLOAT;

                                ok = AddToToken( c );  if ( !ok ) return( ERROR );
                                c  = fgetc( IniFp );

                                while ( isdigit( c ) ) {

                                        ok = AddToToken( c );  if ( !ok ) return( ERROR );
                                        c  = fgetc( IniFp );
                                        }

                                if ( c == 'e' || c == 'E' ) {

                                        ttype = EXPFLOAT;

                                        ok = AddToToken( c );  if ( !ok ) return( ERROR );
                                        c  = fgetc( IniFp );

                                        if ( c == '+' || c == '-' ) {

                                                ok = AddToToken( c );  if ( !ok ) return( ERROR );
                                                c  = fgetc( IniFp );

                                                } // end c + or -
                                          

                                        while ( isdigit( c ) ) {

                                                ok = AddToToken( c );  if ( !ok ) return( ERROR );
                                                c  = fgetc( IniFp );
                                                }
                                        
                                        } // end c exponent

                                } // end c .

                        } // end else not HEX

                } // end else if c isdigit + -


        else if ( c == '"' ) {

                ttype = STRING;

                c = fgetc( IniFp );

                  while ( c != '"' && c != '\n' && c != EOF ) {

                        ok = AddToToken( c );  if ( !ok ) return( ERROR );
                        c  = fgetc( IniFp );
                        }



                if ( c == '"' )

                        c = fgetc( IniFp );

                else {
                        ErrorIni( "missing quote on string" );
                        ttype = ERROR;
                        return( ERROR );
                        }

                }



        else if ( c == EOF ) {

                ttype = EOF;
                c     = ' ';  // reset for next pass
                }


        else {

                ttype = UNKNOWN;

                ok = AddToToken( c );  if ( !ok ) return( ERROR );
                c  = fgetc( IniFp );
                }




        // Null terminate the TOKEN and return its type.

        AddToToken( '\0' );

        return( ttype );
}

//------------------------------------------------------------------------------
// ROUTINE: GetTokenPtr
// PURPOSE: Allow calling program to access the current token.
//------------------------------------------------------------------------------
FUNCTYPE( char* ) GetTokenPtr( void ) {
        return( TOKEN );
}

//------------------------------------------------------------------------------
// ROUTINE: GetFilePtr
// PURPOSE: Allow calling program to access the current ini file.
//------------------------------------------------------------------------------
FUNCTYPE( FILE* ) GetFilePtr( void ) {
        return( IniFp );
}

//------------------------------------------------------------------------------
// ROUTINE: Match
// PURPOSE: Check for a match between s and TOKEN.
//------------------------------------------------------------------------------
FUNCTYPE( int ) Match( char *s ) {

        char *t;

        t = TOKEN;

        while ( *s && *t && toupper( *s ) == toupper( *t ) )

                { s++; t++; }


        return ( *s == '\0' && *t == '\0' ? 1 : 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: MatchTable
// PURPOSE: Check in the specified match table for a match between the string
//          for each table entry and the current token.  If found, return both
//          the value associated with the matching entry and the index of that
//          entry in the table.
//------------------------------------------------------------------------------
FUNCTYPE( int ) MatchTable( MATCHTABLE *table, int *value, int *index ) {

        MATCHTABLE *mt;
        int         mi;

        // Error check
        
        if ( !table ) {
                if (value)  *value = -1;
                if (index)  *index = -1;
                return( 0 );
                }

        // For each entry in the match table

        mt = table;
        mi = 0;

        while ( mt->String != NULL ) {
                
                // Check for a match between s and TOKEN.
                
                if ( Match( mt->String ) ) {
                        if (value)  *value = mt->Value;
                        if (index)  *index = mi;
                        return( 1 );
                        }

                mt++;
                mi++;
                }

        // No match found
        
        return ( 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: KeywordEquals
// PURPOSE: Check for the specified keyword followed by an equal sign.
//------------------------------------------------------------------------------
FUNCTYPE( int ) KeywordEquals( char *keyword ) {

#define KEYWORD_EQUALS_ERROR   "expected = sign after %s keyword"

        if ( Match( keyword ) ) {

                if ( GetToken() == EQUALS )

                        return( 1 );

                else {
                        sprintf( msg, KEYWORD_EQUALS_ERROR, keyword );
                        ErrorIni( msg );
                        return( 0 );
                        }
                }


        return( 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: ChannelKeywordEquals
// PURPOSE: Check for the specified keyword followed by a "channel" number
//          and then an equal sign.
//------------------------------------------------------------------------------
FUNCTYPE( int ) ChannelKeywordEquals( char *keyword, int minchan, int maxchan, int *chan ) {

#define CHAN_KEYWORD_EQUALS_ERROR   "expected = sign after %s # keyword"
#define CHAN_KEYWORD_NUMBER_ERROR   "expected channel number after %s keyword"
#define CHAN_KEYWORD_RANGE_ERROR    "channel number must be > %d and < %d"

        int Channel;

        if ( Match( keyword ) ) {

                // Get and verify channel number
                
                GetInt( &Channel, CHAN_KEYWORD_NUMBER_ERROR );

                if ( Channel < minchan || maxchan < Channel ) {
                        sprintf( msg, CHAN_KEYWORD_RANGE_ERROR, minchan, maxchan );
                        ErrorIni( msg );
                        return( 0 );
                        }

                if ( GetToken() == EQUALS ) {
                        if (chan)  *chan = Channel;
                        return( 1 );
                        }

                else {
                        sprintf( msg, CHAN_KEYWORD_EQUALS_ERROR, keyword );
                        ErrorIni( msg );
                        return( 0 );
                        }

                }


        return( 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: GetOnOff
// PURPOSE: Get an {ON,OFF} parameter value.
//------------------------------------------------------------------------------
FUNCTYPE( int ) GetOnOff( int *parameter, char *error ) {


        GetToken();

        if ( ttype == ERROR )
                return( 0 );


        if ( Match( "ON" ) )

                *parameter = ON;

        else if ( Match( "OFF" ) )

                *parameter = OFF;

        else {
                ErrorIni( error );
                return( 0 );
                }


        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: GetInt
// PURPOSE: Get a decimal int parameter value.
//------------------------------------------------------------------------------
FUNCTYPE( int ) GetInt( int *parameter, char *error ) {


        GetToken();

        if ( ttype != INTEGER ) {

                ErrorIni( error );
                return( 0 );
                }

        sscanf( TOKEN, "%d", parameter );
        
        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: GetUint
// PURPOSE: Get a decimal int parameter value.
//------------------------------------------------------------------------------
FUNCTYPE( int ) GetUint( unsigned int *parameter, char *error ) {

        GetToken();

        if ( ttype != INTEGER ) {

                ErrorIni( error );
                return( 0 );
                }

        sscanf( TOKEN, "%u", parameter );

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: GetUlong
// PURPOSE: Get an unsigned long parameter value.
//------------------------------------------------------------------------------
FUNCTYPE( int ) GetUlong( unsigned long *parameter, char *error ) {

        GetToken();

        if ( ttype != INTEGER ) {

                ErrorIni( error );
                return( 0 );
                }

        sscanf( TOKEN, "%lu", parameter );

        return( 1 );
}


//------------------------------------------------------------------------------
// ROUTINE: GetHex
// PURPOSE: Get a hex parameter value.
//------------------------------------------------------------------------------
FUNCTYPE( int ) GetHex( int *parameter, char *error ) {

        GetToken();

        if ( ttype != HEX ) {

                ErrorIni( error );
                return( 0 );
                }


        sscanf( TOKEN, "%x", parameter );

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: GetFloat
// PURPOSE: Get a floating point parameter value.
//------------------------------------------------------------------------------
FUNCTYPE( int ) GetFloat( double *parameter, char *error ) {

        GetToken();

        if ( (ttype != EXPFLOAT) &&
             (ttype != FLOAT)    &&
             (ttype != INTEGER) ) {

                ErrorIni( error );
                return( 0 );
                }


        sscanf( TOKEN, "%lg", parameter );

        return( 1 );
}
