// FILE: ScopeCmd.c
// COPYRIGHT: (c), Symmetric Research, 2009-2010
//
// Acquire USBxCH and PARGPS data with a simple text only display and disk
// save.
//
// For an overview of the program flow, see the main function located
// at the end of this file after all the helper functions.
//


#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "ScopeLib.h"
#include "SrIniParse.h"         // for PrepareFullMsg, SetErrorIniFunction etc
#include "SrHelper.h"           // for SrKeyboardNonBlock, SrGetKeyWait, SrSleep, etc


// Program copyright and screen title:
//
// 2010/08/25 Added keyboard blocking mode restore
// 2009/08/10 First official version

char *Copyright =

          "\n\nScopeCmd.exe : Copyright (c), Symmetric Research, 2009-2010\n\n";

char *ScreenTitle =

      "USBxCH ScopeCmd [/?]: (c) SR 2009 : Rev {2010/08/25}\n\n";


// Global variables

char cmdmsg[MAXSTR];



// Prototypes needed to satisfy forward references

void Rev( void );
void DisplayValues( void );
int  ProcessCommandLine( int argc, char **argv );
void Usage( char *errmsg );
void StartMsg( void );
void KeyWaitForRun( void );
int  KeyProcess( int key );
int  KeyCheckForQuit( void );
void CleanExit( int value );
void ParseErrorMessage( char *msg );
void ParseWarnMessage( char *msg );
void DisplayLogMessage( int MsgLevel, int DlgShow, int Fatal, char *ExtraMsg );


// Runtime error handling ...

#define ErrorFatal( msg )       DisplayLogMessage( MSGLEVEL_ERR,  MSGDLG_NO_SHOW, ERROR_FATAL, msg )
#define ErrorWarning( msg )     DisplayLogMessage( MSGLEVEL_WARN, MSGDLG_NO_SHOW, ScopeInfo.ProgramExitOnError, msg )
#define LogMessage( msg )       DisplayLogMessage( MSGLEVEL_MSG,  MSGDLG_NO_SHOW, ERROR_NOT_FATAL, msg )


// Message arrays

char *UsageMsg[] = {
        "Acquires and saves data from the command line                 ",
        "                                                              ",
        "USAGE: ScopeCmd [parameters.ini] [/?]                         ",
        "                                                              ",
        "    parameters.ini = file with program settings               ",
        "                /? = show help screen and quit                ",
        "                                                              ",
        "Examples:                                                     ",
        "                                                              ",
        "    ScopeCmd /?                   < show help screen and quit ",
        "    ScopeCmd ScopeSetup130Hz.ini  < run with 130Hz settings   ",
        "                                                              ",
        "See ScopeHelpIniSyntax.txt for descriptions of the allowed    ",
        "parameter settings.                                           ",
        "                                                              ",

        NULL,
        };

char *StartMsgFmt[] = {

        "                                                              \n",
        "Initializing USBxCH with: (see Scope.rpt for more details)    \n",
        "                                                              \n",
        "       AtodDriverName = %s                                    \n",
        "        AtodModelName = %s                                    \n",
        "         GpsModelName = %s                                    \n",
        "     TrueSamplingRate = %f Hz                                 \n",
        "     OutputFileNaming = %s                                    \n",
        "          TriggerType = %s                                    \n",
        "                                                              \n",

        NULL,

        };


// REV MESSAGE:
//

void Rev( void ) {

        // Print the rev info.

        sprintf( cmdmsg, ScreenTitle );
        ScopeDisplayInfo( cmdmsg );
}




//------------------------------------------------------------------------------
// ROUTINE: ScopeCmd main program
// PURPOSE: Read settings in from ini file and acquire data which can be
//          saved in a file and/or displayed on the screen.
//------------------------------------------------------------------------------
int main( int argc, char *argv[] ) {

        int    IniArg, ScreenWidth, ScreenHeight;
        char   CurrDir[MAXSTR], *IniFileName;
        time_t StartTime;
        long   StartUsec;


        // Macros computing the number of remaining unfilled buffer points
        // and resetting the current pointers to the buffer beginning


        // Allow getchar to return even if no character is waiting.

        SrKeyboardMode( KEYMODE_NONBLOCK );



        ScreenWidth  = 800;
        ScreenHeight = 640;
        sprintf( CurrDir, "." );

        sprintf( ScopeInfo.ProgramName, "ScopeCmd" );
        ScopeInitScopeInfo( ScreenWidth, ScreenHeight, CurrDir,
                            &DisplayLogMessage );




        // Get command line arguments

        IniArg = ProcessCommandLine( argc, argv );




        // Initialize INI parameters ...
        //
        //    Setup defaults
        //    Parse the user's .ini file
        //    Fill in additional computed values
        //    Generate the (ini) .rpt file
        //
        //

        // This function checks if the specified file exists.  If so, it
        // just returns.  Otherwise, it writes out a default ini file.  In
        // either case, the ScopeInfo.IniFileName variable is set, so that's
        // what we use as input to the ini parsing function.

        if ( IniArg > 0 )
                IniFileName = argv[IniArg];
        else
                IniFileName = NULL;

        if ( !ScopeIniCheck( IniFileName ) )
                Usage( "Could not process ini file on command line" );

        
        ScopeIniParse( ScopeInfo.IniFileName );
        Rev();  // Announce program version as soon as ScopeInfo.InfoDisplay is known.
        ScopeIniReport();





        // Initialize HDR parameters ...
        //
        //    Setup defaults
        //
        //

        ScopeHdrDefaults();




        // Display the starting message

        StartMsg();





        // Open and initialize the USBxCH and PARGPS.

        if ( !ScopeAtodInit() )
                CleanExit( 1 );





        // Allocate PC acquisition and GPS buffer memory.

        ScopeBuffersAllocate(
                             &ScopeInfo.UsbXchPacketArray,
                             &ScopeInfo.UsbXchAnalogStruct,
                             &ScopeInfo.UsbXchAnalogBuffer,
                             &ScopeInfo.UsbXchAnalogBufferEnd,
                             &ScopeInfo.UsbXchSerialStruct,
                             &ScopeInfo.UsbXchSerialBuffer,
                             &ScopeInfo.UsbXchSerialBufferEnd,
                             &ScopeInfo.UsbXchEquipStruct
                            );






        // Start executing:
        //
        // The USBxCH executes continuously, saving data in its FIFO.
        // This FIFO is drained when it is partly full, so acquisition
        // continues uninterrupted while the PC saves to disk.
        //

        if ( ScopeInfo.ProgramKeypressToRun == INISWITCH_ON ) {
                printf( "\nPress ENTER to start, then ENTER again to quit ... \n" );
                KeyWaitForRun( );
                }
        else
                ScopeDisplayInfo( "\n" );




        // Start acquisition.

        ScopeAtodStart( );


        // Get approximate start time from the PC and set first file name

        GetCurrentSystemTime( &StartTime, &StartUsec );
        ScopeInitTimes( StartTime, StartUsec );
        ScopeTmpFileComputeFinalName( );


        // If using PCTIME style timing, send a timestamp to the hardware

        ScopeAtodPcTimeStamp();





        // Poll the USBxCH for available data.
        //
        // This loop repeats until a complete buffer worth of analog
        // data has been read from the USBxCH.  Then the buffer is written
        // to disk and the loop started again.
        //
        //

        sprintf( cmdmsg, "\nWaiting %4.2f sec for first buffer ...\n", ScopeInfo.BufferTime );
        ScopeDisplayInfo( cmdmsg );


polling_loop:

        // Read data from the USBxCH.  This gets both analog and serial data.
        // The routine does not return until a full buffer of analog data has
        // been acquired.

        if ( !ScopeAtodGetReadyData() ) {

                if ( KeyCheckForQuit( ) )
                        CleanExit( 0 );  // shuts down USBxCH + PARGPS

                SrSleep( ScopeInfo.SleepMs );
                goto polling_loop;
                }


        // Else current USBxCH data buffer is complete so
        // code continues on below to process the data.


        // Save the time when this complete buffer was ready
        // (which is approximately the start of the NEXT buffer).

        GetCurrentSystemTime( &ScopeInfo.ReadyTime, NULL );



//FIX - should really send timestamp every second, not just every buffer
//      but for right now, they are about equal

        // If using PCTIME style timing, send a timestamp to the hardware

        ScopeAtodPcTimeStamp();



        // If trigger condition is satisfied

        if ( ScopeTriggerCondition( ) ) {

                // Do trigger action, typically saving file (ScopeWriteData)

                ScopeTriggerAction( );

//FIX - need to check for trigger action failure
                }


        // Otherwise, trigger is done

        else {

//FIX - need IsTriggered ?
                ScopeTriggerDone(); // Done typically turns off digital bits

                sprintf( cmdmsg, "Buffer[%08lX] discarded\n", ScopeInfo.CurrBuffer );
                ScopeDisplayInfo( cmdmsg );
                }


//FIX - check for overflow ??
//                if ( ScopeAtodCheckOverflow() ) {
//                        ScopeDisplayInfo( "A/D Overflow occurred after write." );
//                        if (ScopeInfo.ProgramKeypressToRun == INISWITCH_OFF &&
//                            ScopeInfo.ProgramExitOnError == INISWITCH_OFF)
//                                ScopeAtodOverflowRecovery();
//                        return;
//                        }


        // Display the acquired data.

        DisplayValues();



        // Prepare for reading new data, increment the number of buffers
        // and points processed and roughly update the next buffer time.
        // For a more accurate value, use the GPS data to compute time.

        ScopeAtodDataPrepareToRead();
        ScopeCurrentBufferIncrement();
        ScopePtCountersInitNext();




        // Has the USBxCH DRAM FIFO overflowed ?
        // If error was not fatal, reset and continue

        if ( ScopeAtodCheckOverflow() ) {

                ErrorWarning( "FIFO overflow during PC processing ..." );
                ScopeOverflowRecovery(  );
                }



        // Check for a 'q' to quit, and then continue polling.

        if ( KeyCheckForQuit( ) )
                CleanExit( 0 );  // shuts down USBxCH + PARGPS

    goto polling_loop;


    CleanExit( 0 );
}


//------------------------------------------------------------------------------
// ROUTINE: DisplayValues
// PURPOSE: Print out acquired data values to the screen.
//------------------------------------------------------------------------------
void DisplayValues( void ) {

        int    i, j, jstart, digin, ggacount, MaxNmea;
        char  *VoltageString;
        double TempC, TempF;


        // Analog and digital channels

        if ( ScopeInfo.ProgramDisplayText == INISWITCH_ON ) {

                // Print title line

                for ( j = 0 ; j < ScopeInfo.AtodAnalogChannels; j++ )
                        printf( " %-8s", ScopeInfo.ChannelTitle[j] );
                for ( j = 0 ; j < ScopeInfo.AtodDigitalChannels; j++ )
                        printf( "D%d", j );
                for ( j = 0 ; j < ScopeInfo.AtodGpsChannels; j++ )
                            printf( "G%d", j );
                for ( j = 0 ; j < ScopeInfo.AtodCounterChannels; j++ )
                            printf( "     OBC%d", j );
                printf( "\n" );


                // Print out the values, using one hex digit for the digital channel

                for ( i = 0 ; i < ScopeInfo.NextPtAnalog ; i ++ ) {

                    // Analog channels
                    for ( j = 0 ; j < ScopeInfo.AtodAnalogChannels; j++ ) {
                        printf( " %08lX", ScopeInfo.UsbXchAnalogStruct[i].AnalogData[j] );
                        }

                    // Digital channels
                    digin = (ScopeInfo.UsbXchAnalogStruct[i].DigitalIn & 0x000F );
                    for ( j = 0 ; j < ScopeInfo.AtodDigitalChannels; j++ ) {
                            printf( " %1X", digin );
                            }

                    // Gps channels
                    ggacount = (ScopeInfo.UsbXchAnalogStruct[i].GGACount & 0x000F);
                    for ( j = 0 ; j < ScopeInfo.AtodGpsChannels; j++ ) {
                            printf( " %1X", ggacount );
                            }

                    // Counter (OBC) channels
                    for ( j = 0 ; j < ScopeInfo.AtodCounterChannels; j++ ) {
                            printf( " %8lX", ScopeInfo.UsbXchAnalogStruct[i].OnBoardCount );
                            }

                    // Finalize
                    printf( "\n" );

                    } // end for i

                } // end if ProgramDisplayText



        // GPS serial info

        if ( ScopeInfo.DisplaySerial == INISWITCH_ON ) {

                for ( i = 0 ; i < ScopeInfo.NextPtSerial ; i ++ ) {

                        printf( "Serial: PpsEvent=%u, NMEA string count = %d\n",
                                  ScopeInfo.UsbXchSerialStruct[i].PpsEventNum,
                                  ScopeInfo.UsbXchSerialStruct[i].NmeaCount
                              );

                        jstart = 0;
                        MaxNmea = ScopeInfo.UsbXchSerialStruct[i].NmeaCount;
                        for ( j = 0 ; j < MaxNmea ; j++ ) {
                                printf( "%s",
                                       &ScopeInfo.UsbXchSerialStruct[i].NmeaMsg[jstart] );
                                jstart += SRDAT_NMEA_MAX_SIZE;
                                }
                        }
                if (ScopeInfo.NextPtSerial > 0)  printf( "\n" );

                } // end if DisplaySerial


        // Equipment info

        if ( ScopeInfo.DisplayEquip == INISWITCH_ON ) {

                for ( i = 0 ; i < ScopeInfo.NextPtEquip ; i ++ ) {

                        SrDatUsbTemperatureCompute( ScopeInfo.UsbXchEquipStruct[i].Temperature,
                                                    &TempC, &TempF );

                        if ( ScopeInfo.UsbXchEquipStruct[i].VoltageGood )
                                VoltageString = "Good";
                        else
                                VoltageString = "Bad ";

                        printf( "Equipment: PpsEvent=%u\n", ScopeInfo.UsbXchEquipStruct[i].PpsEventNum );
                        printf( "Temperature %.0lf F or %.1lf C, Voltage %s\n",
                                  (TempF+0.5L), TempC, VoltageString );

                        } // for i < NextPtEquip

                } // end if DisplayEquip

}


//------------------------------------------------------------------------------
// ROUTINE: ProcessCommandLine
// PURPOSE: Get the name of the ini file with the control settings from the
//          command line.
//------------------------------------------------------------------------------
int ProcessCommandLine( int argc, char *argv[] ) {

        int i, IniArg;


        // Check the number of command line arguments.

        if ( argc < 1 || argc > 3 )

                Usage( "wrong number of command line arguments" );

        // Process args

        IniArg = 0;
        for ( i = 1; i < argc ; i++ ) {
                if ( strcmp( argv[i], "/?" ) == 0 )         // help requested
                        Usage( "?" );
                else if ( IniArg == 0 )                     // ini name found
                        IniArg = i;
                else                                        // dup ini name
                        Usage( "Too many .ini filenames" );
                }

        return( IniArg );

}


//------------------------------------------------------------------------------
// ROUTINE: Usage
// PURPOSE: Print a message showing the command line options.
//------------------------------------------------------------------------------
void Usage( char *errmsg ) {

        int i;

        // Print rev message

        ScopeInfo.ProgramDisplayInfo = INISWITCH_ON;
        Rev();


        // Print the command line usage message and quit.

        // Select first string depending on
        // 1) no message, 2) help requested, or 3) message given

        if ( errmsg == NULL )
                printf( "\nERROR: Command line usage.\n\n" );

        else if ( strcmp( errmsg, "?" ) == 0 )
                ;

        else
                printf( "\nERROR: Command line usage (%s).\n\n", errmsg );


        for ( i = 0 ; UsageMsg[i] ; i++ )
                printf( "%s\n", Chop( UsageMsg[i] ) );

        CleanExit( 1 );
}


//------------------------------------------------------------------------------
// ROUTINE: StartMsg
// PURPOSE: Print a message showing some of the current settings.
//------------------------------------------------------------------------------
void StartMsg( void ) {

        char **fmt, *OutputFileNamingStr, *TriggerTypeStr;

        LogMessage( "Beginning ScopeCmd." );

        if ( ScopeInfo.ProgramDisplayInfo == INISWITCH_OFF )
                return;



        // Display a few of the initialization parameters.

        OutputFileNamingStr = SrDatFileNameTypeString( ScopeInfo.OutputFileNaming   );
        TriggerTypeStr      = SrDatTriggerTypeString( ScopeInfo.TriggerType );

        fmt = StartMsgFmt;

            printf( Chop( *fmt++ )                                       );
            printf( Chop( *fmt++ )                                       );
            printf( Chop( *fmt++ )                                       );
            printf( Chop( *fmt++ ),        ScopeInfo.AtodDriverName      );
            printf( Chop( *fmt++ ),        ScopeInfo.AtodXchModelName    );
            printf( Chop( *fmt++ ),        ScopeInfo.GpsModelName        );
            printf( Chop( *fmt++ ),        ScopeInfo.AcqSamplingRateTrue );
            printf( Chop( *fmt++ ),        OutputFileNamingStr           );
            printf( Chop( *fmt++ ),        TriggerTypeStr                );
            printf( Chop( *fmt++ )                                       );

}







// KEYBOARD FUNCTIONS:
//
//
//
//

//------------------------------------------------------------------------------
// ROUTINE: KeyWaitForRun
// PURPOSE: Wait for a keypress and quit if it is Ctrl+C.
//------------------------------------------------------------------------------
void KeyWaitForRun( void ) {

        int c;

        // Wait for a keypress.

        c = SrGetKeyWait( );


        // Quit if it was Ctrl+C.

        if ( c == 0x3 ) // exit Ctrl+C
                CleanExit( 0 );
}


//------------------------------------------------------------------------------
// ROUTINE: KeyProcess
// PURPOSE: Do any specific processing associated with a particular keypress.
//------------------------------------------------------------------------------
int KeyProcess( int key ) {

        // Special key y is found

        if ( key == 'y' || key == 'Y' ) {

                // Turn off a latched trigger that is staYing on

                ScopeTriggerDone();


                // Now flush any remaining keys like the ENTER required
                // under Linux for the original key to be recognized

                while ( SrGetKeyCheck() != EOF )
                        ;


                // Return EOF so our special key won't cause a quit

                return( EOF );
                }


        // Add any other special key processing here ...
        // else if (key == )


        // No special key found, pass it on through untouched

        else
                return( key );
}

//------------------------------------------------------------------------------
// ROUTINE: KeyCheckForQuit
// PURPOSE: Check for a keypress but don't wait if one is not found.
//------------------------------------------------------------------------------
int KeyCheckForQuit( void ) {

        int c;

        // When running interactively, check for a quit keypress.

        if ( ScopeInfo.ProgramKeypressToRun == INISWITCH_ON ) {

                c = SrGetKeyCheck( );

                c = KeyProcess( c );

                if ( c != EOF )
                        return( 1 );
                }

        return( 0 );
}


//------------------------------------------------------------------------------
// ROUTINE: CleanExit
// PURPOSE: Ensure the A/D is stopped before exiting.
//------------------------------------------------------------------------------
void CleanExit( int value ) {

        // Perform orderly shutdown and exit.

        if ( ScopeInfo.UsbXchHandle != BAD_DEVHANDLE )
                ScopeAtodStop( );


        // Free acquisition and gps buffers

        ScopeBuffersFree(
                         &ScopeInfo.UsbXchPacketArray,
                         &ScopeInfo.UsbXchAnalogStruct,
                         &ScopeInfo.UsbXchAnalogBuffer,
                         &ScopeInfo.UsbXchAnalogBufferEnd,
                         &ScopeInfo.UsbXchSerialStruct,
                         &ScopeInfo.UsbXchSerialBuffer,
                         &ScopeInfo.UsbXchSerialBufferEnd,
                         &ScopeInfo.UsbXchEquipStruct
                        );


        LogMessage( "Ending ScopeCmd." );

        SrKeyboardMode( KEYMODE_BLOCK );  // Restore blocking keyboard behavior.

        exit( value );

}


//------------------------------------------------------------------------------
// ROUTINE: ParseErrorMessage
// PURPOSE: Display error message when called by parsing functions.
//------------------------------------------------------------------------------
void ParseErrorMessage( char *msg ) {

        char *ExtraMsg;
        ExtraMsg = PrepareFullMsg( msg );
        DisplayLogMessage( MSGLEVEL_ERR, MSGDLG_NO_SHOW, ERROR_FATAL, ExtraMsg );
}

//------------------------------------------------------------------------------
// ROUTINE: ParseWarnMessage
// PURPOSE: Display warning message when called by parsing functions.
//------------------------------------------------------------------------------
void ParseWarnMessage( char *msg ) {

        char *ExtraMsg;
        ExtraMsg = PrepareFullMsg( msg );
        DisplayLogMessage( MSGLEVEL_WARN, MSGDLG_NO_SHOW, ERROR_NOT_FATAL, ExtraMsg );
}

//------------------------------------------------------------------------------
// ROUTINE: DisplayLogMessage
// PURPOSE: General message function that does clean up and exits if fatal.
//------------------------------------------------------------------------------
void DisplayLogMessage( int MsgLevel, int DlgShow, int Fatal, char *ExtraMsg ) {

        // Prepare log message and append it to log file

        ScopeLogMessagePrep( MsgLevel, ExtraMsg );



        // Stop A/D execution ??



        // Output the log message to the screen/dialog if requested

        if ( DlgShow == MSGDLG_SHOW )
                printf( ScopeInfo.LogMessage );



        // If error is fatal, then quit

        if ( Fatal == ERROR_FATAL )
                CleanExit( 1 );
}
