// FILE: Dat2Asc.c
// COPYRIGHT: (c), Symmetric Research, 2009-2010
//      
// Dat2Asc converts USBxCH binary DAT files into ASCII text with
// formatted columns for the data and time.
//


// Include files:

#if defined( SROS_WINXP ) || defined( SROS_WIN2K )
#include <windows.h>
#elif defined( SROS_LINUX )
#endif  // SROS_xxxxx

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>       // for isdigit(), toupper(), etc
#include <time.h>

#include "SrUsbXch.h"    // for SrUsbXch library functions
#include "SrHelper.h"    // for SrFileExist, SrSecTimeSplit, SrSecTimeCombine



// Program copyright and screen title:

char *Copyright =

        "\n\nDat2Asc.exe : Copyright (c), Symmetric Research, 2009-2010\n\n";

char *ScreenTitle =

  "USBxCH Dat2Asc Data Conversion Program: (c) SR 2010 : Rev {2010/08/15}\n\n";




// Maximum FILENAME and generic string lengths:
//
// An input filename is comprised of a path, basename, and extension.  The
// path and extension are optional.  In the string:
//
//   c:\directory\foo.dat
//
// the leading c:\directory\ is the path, foo is the basename, and dat
// is the extension.  In Windows, the last '\' denotes the end of the
// path, and the last '.' denotes the beginning of the extension.
//
// A complete filename must be less than MAXFILENAME characters long.
// Even if buffers for the path, basename, and ext are all individually
// dimensioned at MAXFILENAME, the total filename constructed from their
// parts must still be less than MAXFILENAME.
//
//

#define MAXFILENAME             1024
#define MAXSTR                  256   // < generic string length


// Allowed output formats:

#define OUTPUT_FORMAT_NONE     0
#define OUTPUT_FORMAT_FW       1
#define OUTPUT_FORMAT_FW_HEX   2
#define OUTPUT_FORMAT_FW_DBG   3
#define OUTPUT_FORMAT_CSV      4
#define OUTPUT_FORMAT_CSV_HEX  5
#define OUTPUT_FORMAT_CSV_DBG  6
#define OUTPUT_FORMAT_RAW      7
#define OUTPUT_FORMAT_MAX      8

char *OutputFormatName[] = {
        "",
        "fw",
        "fw_hex",
        "fw_debug",
        "csv",
        "csv_hex",
        "csv_debug",
        "raw",
        "",

        NULL
        };




// Dat2Asc RUN TIME PARAMETERS:
//
// Both command line and fixed hardwired parameters are located here.
//    ... access this structure with statements like: DP.HexFormat ...
//
//

struct Dat2AscParametersLayout {

        int  OutFormat;
        int  CSV;
        int  Hex;
        int  Debug;

        char *FileNameTimeSample;
        char *FileNameTimeStatus;
        char *FileNameTimeEquip;
        char *FileNameTimeTotal;
        char *FileNameOutNmea;
        char *FileNameOutData;
        char *FileNameOutCsv;
        char *FileNameRawOut;

        double SamplingRate;

        int  AtodGpsModel;

        int  Sequential;
        int  nSequential;

        int  TimeMethod;
        int  ShowTime;

        int  CurrentPt;


        // command line input file ...

        int InputFileFlag;

        FILE *fpInput;
        char InputFileName[MAXFILENAME];  // base + ext, no path ...

        FILE *fpTimestamp;

        char InputFilePath[MAXFILENAME];
        char InputFileBase[MAXFILENAME];
        char InputFileExt[MAXFILENAME];


        // generic full path work area ...

        char FullPathName[MAXFILENAME];

            char *pFullPathNameBase; // -> base+ext part of FullPathName[] ...


        // string to hold converted Id

        char cDatId[5];
            

       // Time stamps

        TS TS0;
        TS TS1;


        } Dat2AscParameters = {


                // Output format argument values:

                OUTPUT_FORMAT_NONE,     // format, default = NONE
                0,                      // CSV, default = 0
                0,                      // Hex, default = 0
                0,                      // Debug, default = 0


                // Intermediate fixed filenames for each step:

                "Dat2Asc-100-TimeInfoSample.asc",  // FileNameTimeSample
                "Dat2Asc-101-TimeInfoStatus.asc",  // FileNameTimeStatus
                "Dat2Asc-102-TimeInfoEquip.asc",   // FileNameTimeEquip
                "Dat2Asc-200-TimeInfoAll.asc",     // FileNameTimeTotal
                "Dat2Asc-300-Nmea.asc",            // FileNameOutNmea
                "Dat2Asc-301-Data.asc",            // FileNameOutData
                "Dat2Asc-301-Data.csv",            // FileNameOutCsv
                "Dat2Asc-RAW.asc",                 // FileNameRawOut


                // SamplingRate: extracted from header
                
                1.0L,


                // GPS model:  extracted from header

                SRDAT_GPSMODEL_GARMIN,  // AtodGpsModel


                // Sequential processing flags:

                0,      // Sequential on/off, default = off
                0,      // nSequential


                // Time method: selected based on GpsModel

                SRDAT_TIME_METHOD_SPS,  // TimeMethod


                // Display time columns (updated during processing)

                1,      // ShowTime on/off, default = on
                

                // CurrentPt
                
                0,       // default = 0


                // Input file parameters:

                0,      // InputFileFlag = 0, input file required

                // all trailing filename areas initialized to 0 by compiler ...
                
                // all trailing timestamp areas initialized to 0 by compiler ...

                // all trailing header and tag structures initialized to 0 by compiler ...
                
                };



#define DP  Dat2AscParameters  // < DP = shorthand for Dat2AscParameters ...




// USAGE MESSAGE ARRAY:

char* UsageMsg[] = {
    "Converts USBxCH binary DAT files from Scope into ASCII text columns        ",
    "                                                                           ",
    "USAGE: Dat2Asc filename format [/?]                                        ",
    "                                                                           ",
    "    filename = DAT file to convert  (required)                             ",
    "      format = output file format  (required)                              ",
    "          /? = show help screen and quit                                   ",
    "                                                                           ",
    "where:                                                                     ",
    "                                                                           ",
    "    filename must be of the form:                                          ",
    "                                                                           ",
    "       nnnnnnnn.dat  for sequential processing                             ",
    "       alpha numeric for single file processing                            ",
    "                                                                           ",
    "                                                                           ",
    "    format must be one of the following:                                   ",
    "      ( use csv for importing into Excel, Matlab, etc ... )                ",
    "                                                                           ",
    "         fw         fixed width ASCII columns                              ",
    "         fw_hex     fw, but with analog and digital in hex                 ",
    "         fw_debug   fw, but includes lower level debug info                ",
    "                                                                           ",
    "         csv        comma separated ASCII values                           ",
    "         csv_hex    csv, but with analog and digital in hex                ",
    "         csv_debug  csv, but includes lower level debug info               ",
    "                                                                           ",
    "         raw        show raw DAT records for debugging only                ",
    "                                                                           ",
    "                                                                           ",
    " Examples:                                                                 ",
    "                                                                           ",
    "    Dat2Asc /?                < show help screen and quit                  ",
    "    Dat2Asc test.dat fw       < convert single file to fixed width ASCII   ",
    "    Dat2Asc test.dat csv      < convert to csv format                      ",
    "    Dat2Asc 00000000.dat fw   < sequential processing, starting at 0       ",
    "    Dat2Asc 02.dat fw         < not sequential, must be nnnnnnnn.dat!      ",
    "                                                                           ",

    NULL
    };





// Defines and global variables:

#define MAXDATAFMT     10
#define MAXTIMESTR     30
#define MAXCONVERT  80000
#define MAXANALOG   10000   // 10K sps is max rate on USBxCH
#define MAXSERIAL      50
#define MAXEQUIP       50
#define MASKVAL        0x00FFFFFF


// Prototypes

  void Step1( void );
  void Step2( void );
  void Step3( void );
  FILE *FileOpen( char *filename, char *modestring );
   int ReadInputFileHdr( SRDAT_HDR *Hdr );
   int ReadInputFileRecord( SRDAT_TAG *Tag, char *Buffer, int *Skip );

   int PrepareFileNames( void );
   int FileReadTimeSample( FILE *fp, TS *TSpartial );
   int FileReadTimeStatus( FILE *fp, TS *TSpartial, NI *NmeaInfo );
   int FileReadTimeEquip( FILE *fp, TS *TSpartial );
   int FileUpdateTimeSample( FILE *fp, TS *TsPartial, int RequiredPpsCount );
   int FileUpdateTimeStatus( FILE *fp, TS *TsPartial, int RequiredPpsCount,
                             NI *NmeaInfo );
   int FileUpdateTimeEquip( FILE *fp, TS *TsPartial, int RequiredPpsCount );
  void SelectTimeMethod( SRDAT_HDR *Hdr );
   int ComputeTime( long Sample, char *TimeStrSec,
                    char *TimeStrYmd, char *TimeStrHms, char *TimeStrFracSec );
   int ExtractNmeaInfo( char *NmeaGroup, int NmeaCount, NI *NmeaInfo );
  void DisplayAnalogData( FILE *fp, SRDAT_USBXCH_ANALOG *UsbxchAnalogData,
                         int TotalPt, SRDAT_HDR *Hdr, int ShowTitle );
  void DisplaySerialData( FILE *fp, SRDAT_USBXCH_SERIAL *UsbxchSerialData );
  void DisplayEquipData( FILE *fp, SRDAT_USBXCH_EQUIPMENT *UsbxchEquipData );
  void DisplayUnknownData( FILE *fp, char *Unknown, int DataType, 
                          int MaxPerLine, unsigned long Nbytes );
  void DisplayConvertedPackets( FILE *fp,
                                SRUSBXCH_PACKET *UsbPacketArray,
                                int *PacketCount,
                                int *PacketCountSample,
                                int *PacketCountStatus,
                                SRDAT_HDR *Hdr );
  void ShowHdr( FILE *fp, SRDAT_HDR *Hdr );
  void ShowTag( FILE *fp, SRDAT_TAG *Tag );
  void ShowData( FILE *fpOutData, FILE *fpOutNmea, SRDAT_HDR *Hdr, void *Buffer,
                 long DataType, unsigned long Nbytes );
 char *PrepareTriggerParm1String( long TriggerType, long TriggerParm1 );
 char *PrepareTriggerParm2String( long TriggerType, long TriggerParm2 );
  void ErrorCmd( char *msg );
  void ErrorExec( char *msg );
  void WarningExec( char *msg );
   int ProcessCommandLine( int argc, char *argv[] );
  void Usage( void );
   int MatchFormat( char *argument );
   int StrEqu( char *s1, char *s2 );
   int StrIsEightDecimal( char *s );
   int ParseFileName( char *FileName );
  void PrintRunTimeSummary( void);
  char *ChopLong( char *s);
  void ProcessRawRecords( void );
  void ShowRawData( FILE *fpRawOut, char *Buffer, unsigned long TagId,
                    unsigned long Nbytes, unsigned long RecordNum, 
                    unsigned long nChan, unsigned long *nSamples );
  char *IdString( unsigned long Id );




//------------------------------------------------------------------------------
// ROUTINE: main()  Dat2Asc
// PURPOSE: Main entry point for Dat2Asc ... convert DAT files to ASCII.
//------------------------------------------------------------------------------
int main( int argc, char *argv[] ) {


        // Announce program version.

        printf( ScreenTitle );


        // Process command line arguments

        ProcessCommandLine( argc, argv );       // << fills in the DP structure


        // Print a summary of the run time arguments.

        PrintRunTimeSummary( );



        // Check for "raw" Dat2Asc processing:
        //
        //   ... this processing option is primarily for debugging ...
        //
        //

        if ( DP.OutFormat == OUTPUT_FORMAT_RAW )

                ProcessRawRecords(); // << output the raw records and quit



        // Normal Dat2Asc processing:
        //
        // Make multiple passes to align analog records and
        // (GPS/NMEA) records to form the final ASCII output.
        //
        //

        // Prepare Sample/Status/Equip info
        
        Step1();

        
        // Combine S/S/E info into full timestamp
        
        Step2();


        // Convert to ASCII

        Step3();

        
        return( 0 );
}


//------------------------------------------------------------------------------
// ROUTINE: Step1
// PURPOSE: Initialize time stamp information that will be used to determine the
//          time of each sample point.  This is done by reading through all the
//          input .dat files gathering up the USB Analog and USB Gps information
//          associated with each GPS PPS signal and associating them all together
//          into a time stamp structure.  This time stamp info is saved in a
//          temporary file to remove any hardwired limits on how many files can
//          be processes at one time.
//------------------------------------------------------------------------------
void Step1( void ) {

        long         StartSample, SeqNumSave;
        FILE         *fpSample, *fpStatus, *fpEquip;
        int          i, Npts, Nser, Nequip, Skip;
        int          nSampleTS, nStatusTS, nEquipTS;
        int          ObcCount, PpsCount, MaskPpsCount;
        int          PpsEvent, PpsEventLast, PpsEventCorrected;
        NI           NmeaInfo;
        SRDAT_HDR    Hdr;
        SRDAT_TAG    Tag;
        SRDAT_USBXCH_ANALOG    *UsbAnalogData;
        SRDAT_USBXCH_SERIAL    *UsbSerialData;
        SRDAT_USBXCH_EQUIPMENT *UsbEquipData;
        static char Buffer[MAXCONVERT];


// These format strings are used by both printf and scanf, 
// so don't give any precisions in the format
        
#define FMT_TSAMPTITLE "  PpsCount      Sample    ObcCount     ExtraSec\n"
//                      NNNNNNNNNN  NNNNNNNNNN  0xNNNNNNNN FFFFFFFFFFFF
#define FMT_TSAMP      "%10d  %10d  0x%08X %12lf\n"

#define FMT_TSTATTITLE "  PpsCount    SecSince1970    Sat Ysrc Hsrc   PpsEvent      Time as YMD HMS\n"
//                      NNNNNNNNNN FFFFFFFFFFFFFFFFF   NN  NN  NN   NNNNNNNNNN NNNN/NN/NN NN:NN:NN.NNNNNN
#define FMT_TSTAT      "%10d %17lf   %2d  %2d  %2d   %10d %4d/%02d/%02d %02d:%02d:%02d.%06ld\n"
        
#define FMT_TSEQTITLE "  PpsCount Pwr  Temp\n"
//                     NNNNNNNNNN  N  0xXXXX
#define FMT_TSEQ      "%10d  %1d  0x%04X\n"



        // Announce STEP 1 is processing.

        printf( "\nSTEP 1 PROCESSING ...\n" );

        
        // Open temporary time info files, for sample (analog), status (NMEA), and equip

        fpSample = FileOpen( DP.FileNameTimeSample, "w" );
        fpStatus = FileOpen( DP.FileNameTimeStatus, "w" );
        fpEquip  = FileOpen( DP.FileNameTimeEquip, "w" );

        fprintf( fpSample, FMT_TSAMPTITLE );
        fprintf( fpStatus, FMT_TSTATTITLE );
        fprintf( fpEquip,  FMT_TSEQTITLE );
        


        // Save starting input file sequence number
        
        SeqNumSave = DP.nSequential;


        // Initialize values

        nSampleTS    = 0;
        nStatusTS    = 0;
        nEquipTS     = 0;
        PpsEventLast = 0;



        // Loop for all input files

        while ( ReadInputFileHdr( &Hdr ) ) {

                StartSample = Hdr.FileStartPtNum;

                // Loop through the entire file reading data records.
                // If a record is one of the known types, extract out
                // time stamp info into the appropriate temporary file

                while ( ReadInputFileRecord( &Tag, Buffer, &Skip )  ) {

                        if ( Skip )
                                continue;

                        else if ( Tag.TagId == SRDAT_TAGID_USBANALOG ) {

                                Npts = Tag.Nbytes / sizeof( SRDAT_USBXCH_ANALOG );
                                UsbAnalogData = (SRDAT_USBXCH_ANALOG*)Buffer;

                                for ( i = 0 ; i < Npts ; i++ ) {

                                        ObcCount = UsbAnalogData[i].OnBoardCount;
                                        PpsEvent = UsbAnalogData[i].GGACount;

                                        if ( PpsEvent == PpsEventLast )
                                                PpsEventCorrected = PpsEvent+1;
                                        else
                                                PpsEventCorrected = PpsEvent;
   
                                        if ( ObcCount != 0L ) {

                                                fprintf( fpSample, FMT_TSAMP,
                                                         (PpsEventCorrected & 0x0F),
                                                         (StartSample + i),
                                                         ObcCount,
                                                         (((double)ObcCount) / USBXCH_OBC_FREQ) );

                                                nSampleTS++;
                                                }

                                        PpsEventLast = PpsEvent;

                                        } // end for i < Npts
                                
                                StartSample += Npts;
                                }

                        else if ( Tag.TagId == SRDAT_TAGID_USBSERIAL ) {

                                Nser = Tag.Nbytes / sizeof( SRDAT_USBXCH_SERIAL );
                                UsbSerialData = (SRDAT_USBXCH_SERIAL*)Buffer;

                                for ( i = 0 ; i < Nser ; i++ ) {

                                        ExtractNmeaInfo( UsbSerialData[i].NmeaMsg,
                                                         UsbSerialData[i].NmeaCount,
                                                        &NmeaInfo );

                                        PpsCount     = UsbSerialData[i].PpsEventNum;
                                        MaskPpsCount = PpsCount & 0x0F;

                                        fprintf( fpStatus, FMT_TSTAT,
                                                MaskPpsCount,             // PpsCount
                                                NmeaInfo.SecSince1970,    // SecSince 1970
                                                NmeaInfo.Nsat,            // Sat
                                                NmeaInfo.YmdSource,       // Ysrc
                                                NmeaInfo.HmsSource,       // Hsrc
                                                PpsCount,                 // PpsEvent
                                                NmeaInfo.Year,            // Time as YMD HMS ...
                                                NmeaInfo.Month,
                                                NmeaInfo.Day,
                                                NmeaInfo.Hour,
                                                NmeaInfo.Minute,
                                                NmeaInfo.Second,
                                                NmeaInfo.MicroSecond
                                                );

                                        nStatusTS++;

                                        } // end for i < Nser
                                }


                        else if ( Tag.TagId == SRDAT_TAGID_USBEQUIP ) {

                                Nequip = Tag.Nbytes / sizeof( SRDAT_USBXCH_EQUIPMENT );
                                UsbEquipData = (SRDAT_USBXCH_EQUIPMENT*)Buffer;

                                for ( i = 0 ; i < Nequip ; i++ ) {

                                        fprintf( fpEquip, FMT_TSEQ,
                                                (UsbEquipData[i].PpsEventNum & 0x0F),
                                                 UsbEquipData[i].VoltageGood,
                                                 UsbEquipData[i].Temperature );

                                        nEquipTS++;

                                        } // end for i < Nequip
                                }

                        else if ( Tag.TagId == SRDAT_TAGID_EOF )
                                break;

                        }  // end of while ReadInputFileRecord


                // Close current input file

                fclose( DP.fpInput );


                // Prepare for next one

                DP.nSequential++;
                if ( !PrepareFileNames() )
                        break;
                
                } // end of while ReadInputFileHdr


        
        // Restore original sequence number and input file name

        DP.nSequential = SeqNumSave;
        PrepareFileNames();



        // Close the temporary files

        fclose( fpSample );
        fclose( fpStatus );
        fclose( fpEquip );

        printf( "Step 1: Done, Sample/Status/Equip info prepared with\n\n" );
        printf( "        Sample timestamps found = %d\n", nSampleTS );
        printf( "        Status timestamps found = %d\n", nStatusTS );
        printf( "        Equip  timestamps found = %d\n", nEquipTS );
        printf( "\n" );

}


//------------------------------------------------------------------------------
// ROUTINE: Step2
// PURPOSE: Combine information from the various time stamp files into one
//          master time stamp file.
//------------------------------------------------------------------------------
void Step2( void ) {

        FILE *fpSample, *fpStatus, *fpEquip;
        int  nTime, TsTitle;
        TS   TsSample, TsStatus, TsEquip, TsTotal;
        NI   NmeaInfo;


        // Announce STEP 2 is processing.

        printf( "\nSTEP 2 PROCESSING ...\n" );


        // Open the temporary files

        fpSample       = FileOpen( DP.FileNameTimeSample, "r" );
        fpStatus       = FileOpen( DP.FileNameTimeStatus, "r" );
        fpEquip        = FileOpen( DP.FileNameTimeEquip,  "r" );
        DP.fpTimestamp = FileOpen( DP.FileNameTimeTotal,  "w" );

        SrDatSkipTitleLine( fpSample );
        SrDatSkipTitleLine( fpStatus );
        SrDatSkipTitleLine( fpEquip );


        // Combine points from all intermediate files (fpSample + fpStatus + fpEquip)

        nTime   = 0;
        TsTitle = 1;    // Write output title the first time

        while ( FileReadTimeSample( fpSample, &TsSample ) &&
                FileReadTimeStatus( fpStatus, &TsStatus, &NmeaInfo ) &&
                FileReadTimeEquip(  fpEquip,  &TsEquip ) ) {

                // If the PpsCounts don't match, update the low file

                if ( TsSample.PpsCount < TsStatus.PpsCount ) {
                        if ( !FileUpdateTimeSample( fpSample,
                                                    &TsSample, TsStatus.PpsCount ) )
                                break;
                        }

                if ( TsStatus.PpsCount < TsSample.PpsCount ) {
                        if ( !FileUpdateTimeStatus( fpStatus, &TsStatus,
                                                    TsSample.PpsCount, &NmeaInfo ) )
                                break;
                        }

// FIX - check into what happens if TsEquip and TsSample don't have same PpsCount
                
                if ( TsEquip.PpsCount < TsSample.PpsCount ) {
                        if ( !FileUpdateTimeEquip( fpEquip,
                                                    &TsEquip, TsSample.PpsCount ) )
                                break;
                        }



                // Once the PpsCounts match, combine the time info
                // and write it out

                if ( TsSample.PpsCount == TsStatus.PpsCount ) {

                        TsTotal.PpsCount       = TsSample.PpsCount;
                        TsTotal.Sample         = TsSample.Sample;
                        TsTotal.TotalTime      = NmeaInfo.SecSince1970 + TsSample.ExtraSec;
                        TsTotal.ExtraSec       = TsSample.ExtraSec;
                        TsTotal.ObcCount       = TsSample.ObcCount;
                        TsTotal.NumSat         = NmeaInfo.Nsat;
                        TsTotal.YmdSource      = NmeaInfo.YmdSource;
                        TsTotal.HmsSource      = NmeaInfo.HmsSource;
                        
                        if ( TsSample.PpsCount == TsEquip.PpsCount ) {
                                TsTotal.PowerInfo   = TsEquip.PowerInfo;
                                TsTotal.Temperature = TsEquip.Temperature;
                                }
                        else {
                                TsTotal.PowerInfo   = USBXCH_POWER_UNKNOWN;
                                TsTotal.Temperature = USBXCH_TEMP_UNKNOWN;
                                }


                        SrDatTimestampWriteFile( DP.fpTimestamp, &TsTotal, TsTitle );
                        TsTitle = 0;

                        nTime++;
                        }
                else {
                        printf( "Trouble with alignment\n" );
                        exit( 1 );
                        }

                }

        
        // Close timestamp and temporary files
        
        fclose( fpSample );
        fclose( fpStatus );
        fclose( fpEquip );
        fclose( DP.fpTimestamp );

        printf( "Step 2: Done, time info combined into timestamps\n\n" );
        printf( "        Total timestamps found = %d\n", nTime );
        printf( "\n" );

}

//------------------------------------------------------------------------------
// ROUTINE: Step3
// PURPOSE: Process all files reading in records and writing them out.
//------------------------------------------------------------------------------
void Step3( void ) {

        int          Skip, nFiles, nRecords, nSkip;
        double       TpSecTime;
        SRDAT_TIMEPT Tp;
        SRDAT_HDR    Hdr;
        SRDAT_TAG    Tag;
        FILE        *fpOutData, *fpOutNmea;
        TS           TsTmp;
        char        *FileNameOut;
        
        static char Buffer[MAXCONVERT];

        
        // Announce STEP 3 is processing.

        printf( "\nSTEP 3 PROCESSING ...\n" );


        if ( DP.CSV )
                FileNameOut = DP.FileNameOutCsv;
        else
                FileNameOut = DP.FileNameOutData;



        // Open output files

        fpOutNmea = FileOpen( DP.FileNameOutNmea, "w" );
        fpOutData = FileOpen( FileNameOut,        "w" );




        // Determine time method

        ReadInputFileHdr( &Hdr ); // opens DP.fpInput
        fclose( DP.fpInput );
        
        SelectTimeMethod( &Hdr ); // sets DP.TimeMethod



        // Prepare default timestamps from header

        // Get PC time in seconds from the header

        Tp = Hdr.TimePoint[0];
        SrSecTimeCombine( &TpSecTime,
                        (int)Tp.Year, (int)Tp.Month, (int)Tp.Day,
                        (int)Tp.Hour, (int)Tp.Minute, (int)Tp.Second,
                         Tp.Microsecond );


        // Fill in one timestamp and make a dummy

        DP.TS0.PpsCount    = SRDAT_INVALID_PPSEVENT;
        DP.TS0.Sample      = Hdr.FileStartPtNum;;
        DP.TS0.TotalTime   = TpSecTime;
        DP.TS0.ExtraSec    = 0.0L;
        DP.TS0.ObcCount    = 0;
        DP.TS0.NumSat      = 0;
        DP.TS0.YmdSource   = Tp.YMDSource;
        DP.TS0.HmsSource   = Tp.HMSSource;
        DP.TS0.PowerInfo   = USBXCH_POWER_UNKNOWN;
        DP.TS0.Temperature = USBXCH_TEMP_UNKNOWN;

        DP.TS1             = DP.TS0;
        DP.TS1.Sample      = 0x7FFFFFFF;

        
        
        // If working with GPS, open time stamp file and get two timestamps
        
        if ( DP.TimeMethod == SRDAT_TIME_METHOD_OBC ) {
        
                DP.fpTimestamp = FileOpen( DP.FileNameTimeTotal, "r" );

                // Get first timestamp

                if ( !SrDatTimestampReadFile( DP.fpTimestamp, &TsTmp, 1 ) ) { // skip title
                        WarningExec( "Not enough timestamps for GPS time.  Using header time instead.\n" );
                        fclose( DP.fpTimestamp );
                        DP.TimeMethod = SRDAT_TIME_METHOD_SPS;
                        // Ignore TsTmp and use default TS0 and TS1
                        }

                else {
                        DP.TS0 = TsTmp; // Use valid first timestamp

                        // Get second timestamp

                        if ( !SrDatTimestampReadFile( DP.fpTimestamp, &TsTmp, 0 ) ) {
                                WarningExec( "Not enough timestamps for full GPS time.  Using sps time instead.\n" );
                                fclose( DP.fpTimestamp );
                                DP.TimeMethod = SRDAT_TIME_METHOD_SPS;
                                // Ignore TsTmp and update default TS1
                                DP.TS1        = DP.TS0;
                                DP.TS1.Sample = 0x7FFFFFFF;
                                }
                        else
                                DP.TS1 = TsTmp; // Use valid second timestamp
                        }
                }


        // Show title line (no data)

        DisplayAnalogData( fpOutData, NULL, 0, &Hdr, 1 );



        // Loop until no more dat files requested or found

        nFiles   = 0;
        nRecords = 0;
        nSkip    = 0;

        while ( ReadInputFileHdr( &Hdr ) ) {

                // Initialize point counts that will be updated by ShowData
                
                DP.CurrentPt    = (int)Hdr.FileStartPtNum;
                DP.SamplingRate = Hdr.AtodSamplingRate;


                // Read and write data records until reaching the end
                
                while ( ReadInputFileRecord( &Tag, Buffer, &Skip )  ) {

                        if ( Skip ) {
                                nSkip++;
                                continue;
                                }

                        else if ( Tag.TagId == SRDAT_TAGID_EOF )
                                break;

                        else {
                                ShowData( fpOutData,
                                          fpOutNmea,
                                          &Hdr,
                                          Buffer,
                                          Tag.TagId,
                                          Tag.Nbytes );
                                nRecords++;
                                }

                        } // end while ReadInputFileRecord


                // Close current input file

                fclose( DP.fpInput );
                nFiles++;


                // Prepare for next ones

                DP.nSequential++;
                if ( !PrepareFileNames() )
                        break;

                }  // end while ReadInputFileHdr (more files to process)


        // Close output files

        fclose( fpOutData );
        fclose( fpOutNmea );

        
        
        printf( "Step 3: Done, ASCII conversion complete\n\n"             );
        printf( "        Files processed = %d\n", nFiles                  );
        printf( "      Records processed = %d\n", nRecords                );
        printf( "        Records skipped = %d\n", nSkip                   );
        printf( "        Final data file = %s\n", ChopLong( FileNameOut ) );
        printf( "\n\n" );

        printf( "SUCCESS: final ASCII data >> %s\n\n", FileNameOut );
}

//------------------------------------------------------------------------------
// ROUTINE: FileOpen
// PURPOSE: Open the specified filename.
//------------------------------------------------------------------------------
FILE *FileOpen( char *filename, char *modestring ) {

        FILE *fp;
        char MsgBuffer[MAXSTR], *CreateOpenStr;


        // Form the full path name.
        //
        // Here filename is the base + ext.  Append it to the path for the
        // full pathname.
        //
        //

        sprintf( DP.pFullPathNameBase, "%s", filename );


        // Check mode string, can be w, r, or rb

        if ( StrEqu( modestring, "w" ) )
                CreateOpenStr = "creating";
        
        else // StrEqu( modestring, "r" ) || StrEqu( modestring, "rb" ) )
                CreateOpenStr = " opening";


        // Try to open it.

        fp = fopen( DP.FullPathName, modestring );

        if ( fp ) { // success, return the file pointer ...

                printf( "%s ... %s\n", CreateOpenStr, ChopLong( filename ) );
                return( fp );

                }

        else { // fp == NULL, fatal error ...

                sprintf( MsgBuffer, "failed to open %s", ChopLong( filename ) );

                    ErrorExec( MsgBuffer );

                }

        return( NULL ); // quiet the compiler, should never get here
}

//------------------------------------------------------------------------------
// ROUTINE: ReadInputFileHdr
// PURPOSE: Open an input file and read and validate its header.  If the current
//          file does not exist, return 0.  If an error occurs opening or
//          reading the file, exit.  Otherwise return 1 for success.
//------------------------------------------------------------------------------
int ReadInputFileHdr( SRDAT_HDR *Hdr ) {

        if ( !Hdr ) {
                WarningExec( "Bad input to ReadInputFileHdr function." );
                return( 0 );
                }

        // Verify that file exists
        
        sprintf( DP.pFullPathNameBase, "%s", DP.InputFileName );

        if ( !SrFileExist( DP.FullPathName ) ) {
                printf( "No more files to process.  Stopping at %s.\n", DP.InputFileName );
                return( 0 );
                }


        // Open input file

        DP.fpInput = FileOpen( DP.InputFileName, "rb" );


        // Read header, gather useful fields and check them
        
        if ( !DP.fpInput )
                ErrorExec( "Can't open input file" );

        if ( !SrDatHdrRead( DP.fpInput, Hdr ) )
                ErrorExec( "Could not read .DAT header" );

        if ( !SrDatHdrValidate( Hdr ) )
                ErrorExec( "This is not a valid .DAT file" );

        if ( Hdr->AtodSamplingRate <= 0.000001 ) {
                WarningExec( "Bad sampling rate.  Reset to 1 sps." );
                Hdr->AtodSamplingRate = (double)1.0;
                }


        // Success

        return( 1 );
}


//------------------------------------------------------------------------------
// ROUTINE: PrepareFileNames
// PURPOSE: Compute an input file name based on the current sequential number.
//------------------------------------------------------------------------------
int PrepareFileNames( void ) {
   
        // If working with sequential files, compute the current name

        if ( DP.Sequential ) {
                sprintf( DP.InputFileName,  "%08ld.dat", DP.nSequential );
                return( 1 );
                }


        // Otherwise only a single file was requested and we're done

        else
                return( 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: ReadInputFileRecord
// PURPOSE: Read record tag and data from the open input file.  The tag, data,
//          and number of bytes are returned.  If the data is too big to fit
//          in the buffer, it is skipped.
//------------------------------------------------------------------------------
int ReadInputFileRecord( SRDAT_TAG *Tag, char *Buffer, int *Skip ) {

        unsigned int nbytes;
        char         MsgBuffer[MAXSTR];

        // Check arguments

        if ( !Tag || !Buffer || !Skip )
                ErrorExec( "Bad arguments passed to ReadInputFileRecord\n" );


        // Read tag

        if ( !SrDatTagRead( DP.fpInput, Tag ) )
                ErrorExec( "Could not read record tag\n" );

        // Check data size

        nbytes = (unsigned int)Tag->Nbytes;
        if ( (unsigned long)nbytes != Tag->Nbytes )
                WarningExec( "Data exceeds size allowed by unsigned int.");


        // Read data if it will fit

        if ( nbytes < MAXCONVERT ) {
                *Skip = 0;
                if ( !SrDatDataRead( DP.fpInput, Buffer, nbytes ) )
                        ErrorExec( "Could not read .DAT data" );
                }

        // Otherwise skip it

        else {
                *Skip = 1;
                SrDatDataSkip( DP.fpInput, nbytes );
                sprintf( MsgBuffer, "Exceeded allocated buffer size with %lu.",
                         nbytes );
                WarningExec( MsgBuffer );
                }

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: SelectTimeMethod
// PURPOSE: Select the best method for computing the time for each data point.
//          If any GPS model or PC time is specified and there are enough
//          timestamps available, use the OBC method, otherwise, use the SPS
//          method.
//------------------------------------------------------------------------------
void SelectTimeMethod( SRDAT_HDR *Hdr ) {

        if ( Hdr                                     &&
             Hdr->DatGpsModel != SRDAT_GPSMODEL_NONE )

                DP.TimeMethod = SRDAT_TIME_METHOD_OBC;
        else
                DP.TimeMethod = SRDAT_TIME_METHOD_SPS;
}


//------------------------------------------------------------------------------
// ROUTINE: ComputeTime
// PURPOSE: Use information in the previously constructed time stamp structures
//          to determine the time of the specified sample point.
//------------------------------------------------------------------------------
int ComputeTime( long Sample, char *TimeStrSec,
                 char *TimeStrYmd, char *TimeStrHms, char *TimeStrFracSec ) {

        int    Ok, Year, Month, Day, Hour, Minute, Second;
        long   Microsecond;
        double Time, x1, x2, y1, y2, xk, yk;
        TS     TSx;


        // Compute time using onboard count info
        
        if ( DP.TimeMethod == SRDAT_TIME_METHOD_OBC ) {

                // Update timestamps if current sample number is
                // beyond the right-most timestamp sample number

                while ( Sample > DP.TS1.Sample ) {
                        if ( SrDatTimestampReadFile( DP.fpTimestamp, &TSx, 0 ) ) {
                                DP.TS0 = DP.TS1;
                                DP.TS1 = TSx;
                                }
                        else
                                break;
                        }


                // Prepare 2 points to go into straight line equation

                x1 = (double)DP.TS0.Sample;
                x2 = (double)DP.TS1.Sample;
                y1 = DP.TS0.TotalTime;
                y2 = DP.TS1.TotalTime;
                xk = (double)Sample;  // CurrentPt

                Ok = SrDatEquation2pt( x1, y1, x2, y2, xk, &yk );

                if ( Ok )
                        Time = yk;
                else {
                        if ( TimeStrSec ) strcpy( TimeStrSec, "Time error" );
                        if ( TimeStrYmd ) strcpy( TimeStrYmd, "Time error" );
                        return( 0 );
                        }
                }


        // Compute time using sample per second method

        else { // DP.TimeMethod == SRDAT_TIME_METHOD_SPS

                // Get the first time point from the header and use it plus the
                // sampling rate to determine the time.  Note that we checked
                // SamplingRate in main to ensure it is valid so we won't divide
                // by zero here.

                Time = DP.TS0.TotalTime +
                       (double)(Sample - DP.TS0.Sample) / DP.SamplingRate;
                
                }



        // Split time in seconds to YMD HMS fields

        SrSecTimeSplit( Time, &Year, &Month, &Day, 
                        &Hour, &Minute, &Second, &Microsecond );


        // Fill time strings with selected format
        
        if ( TimeStrSec )
                sprintf( TimeStrSec, "%17lf", Time );

        if ( TimeStrYmd )
                sprintf( TimeStrYmd, "%4d/%02d/%02d", Year, Month, Day );

        if ( TimeStrHms )
                sprintf( TimeStrHms, "%02d:%02d:%02d", Hour, Minute, Second );

        if ( TimeStrFracSec )
                sprintf( TimeStrFracSec, "0.%06ld", Microsecond );
                
        return( 1 );
}


//------------------------------------------------------------------------------
// ROUTINE: ExtractNmeaInfo
// PURPOSE: 
//------------------------------------------------------------------------------
int ExtractNmeaInfo( char *NmeaGroup, int NmeaCount, NI *NmeaInfo ) {

        int Ok;

        if ( !NmeaGroup || NmeaCount < 0 || !NmeaInfo )
                return( 0 );


        // Get the best NMEA info possible from the provided group of NMEA strings

        Ok = SrDatRequestNmeaInfo2( NmeaGroup, NmeaCount,
                                    &NmeaInfo->Year,
                                    &NmeaInfo->Month,
                                    &NmeaInfo->Day,
                                    &NmeaInfo->YmdSource,
                                    &NmeaInfo->YmdIsValid,
                                    &NmeaInfo->Hour,
                                    &NmeaInfo->Minute,
                                    &NmeaInfo->Second,
                                    &NmeaInfo->MicroSecond,
                                    &NmeaInfo->HmsSource,
                                    &NmeaInfo->HmsIsValid,
                                    &NmeaInfo->Latitude,
                                    &NmeaInfo->Longitude,
                                    &NmeaInfo->LocSource,
                                    &NmeaInfo->LocIsValid,
                                    &NmeaInfo->Altitude,
                                    &NmeaInfo->PosSource,
                                    &NmeaInfo->PosIsValid,
                                    &NmeaInfo->Nsat,
                                    &NmeaInfo->SatSource,
                                    &NmeaInfo->SatIsValid
                                  );

        // If all went well, output the data to the specified file
        
        if ( Ok ) {
                Ok = SrDatSecTimeCombine( NmeaInfo->Year, NmeaInfo->Month, NmeaInfo->Day,
                                          NmeaInfo->Hour, NmeaInfo->Minute, NmeaInfo->Second, 
                                          NmeaInfo->MicroSecond, &NmeaInfo->SecSince1970 );

                NmeaInfo->SecIsValid = Ok;
                
                if ( Ok && NmeaInfo->YmdIsValid && NmeaInfo->HmsIsValid )

                        return( 1 );

//                else {
//                        printf( "Warning NMEA time/date is not valid\n" );
//                        }
                }
        else
                WarningExec("SrDatRequestNmeaInfo2 failed." );

        return( 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: ShowHdr
// PURPOSE: Write header information to the ASCII output file.
//------------------------------------------------------------------------------
void ShowHdr( FILE *fp, SRDAT_HDR *Hdr ) {

        int i;

        
        // Check header size
        
        if (sizeof(*Hdr) != SRDAT_HDR_LAYOUTSIZE)
                ErrorExec( "Header size not right" );

        
        // Show header info
        
        fprintf( fp, "# Header:\n");

        fprintf( fp, "# DatId               = 0x%lX %s\n", Hdr->DatId, IdString(Hdr->DatId) );
        fprintf( fp, "# DatRev              = %ld\n", Hdr->DatRev);
        fprintf( fp, "# DatProduct          = %s\n",  Hdr->DatProduct);
        fprintf( fp, "# DatApplication      = %s\n",  Hdr->DatApplication);
        fprintf( fp, "# DatAtodModel        = %ld", Hdr->DatAtodModel);
        fprintf( fp, " (%s)\n", SrDatAtodModelString(Hdr->DatAtodModel) );
        fprintf( fp, "# DatGpsModel         = %ld", Hdr->DatGpsModel);
        fprintf( fp, " (%s)\n", SrDatGpsModelString(Hdr->DatGpsModel) );

        fprintf( fp, "# FileName            = %s\n",  Hdr->FileName);
        fprintf( fp, "# FileCreateTimeHi    = %ld\n", Hdr->FileCreateTimeHi);
        fprintf( fp, "# FileCreateTime      = %ld\n", Hdr->FileCreateTime);
        fprintf( fp, "# FileSeqNum          = %ld\n", Hdr->FileSeqNum);
        fprintf( fp, "# FileStartPtNum      = %ld\n", Hdr->FileStartPtNum);
        fprintf( fp, "# FileNameType        = %ld",   Hdr->FileNameType);
        fprintf( fp, " (%s)\n", SrDatFileNameTypeString(Hdr->FileNameType) );
        fprintf( fp, "# FileNameInc         = %ld\n", Hdr->FileNameInc);

        fprintf( fp, "# ChannelsTotal       = %ld\n", Hdr->Channels);
        fprintf( fp, "# ChannelsAnalog      = %ld\n", Hdr->ChannelsAnalog);
        fprintf( fp, "# ChannelsDigital     = %ld\n", Hdr->ChannelsDigital);
        fprintf( fp, "# ChannelsGps         = %ld\n", Hdr->ChannelsGps);
        fprintf( fp, "# ChannelsCounter     = %ld\n", Hdr->ChannelsCounter);
        fprintf( fp, "# ChannelDigIndex     = %ld\n", Hdr->ChannelDigIndex);
        fprintf( fp, "# ChannelMarkIndex    = %ld\n", Hdr->ChannelMarkIndex);
        fprintf( fp, "# ChannelObcIndex     = %ld\n", Hdr->ChannelObcIndex);
        fprintf( fp, "# ChannelPtsPerRecord = %ld\n", Hdr->ChannelPtsPerRecord);
        fprintf( fp, "# ChannelPtsPerFile   = %ld\n", Hdr->ChannelPtsPerFile);

        for ( i = 0 ; i < Hdr->Channels ; i++ ) {
                fprintf( fp, "# Channel %02ld          = %s\n",
                          Hdr->ChannelList[i].Channel,
                          Hdr->ChannelList[i].Title );
                }

        fprintf( fp, "# DigitalPerChan      = %ld\n", Hdr->DigitalPerChan);
        for ( i = 0 ; i < Hdr->DigitalPerChan ; i++ ) {
                fprintf( fp, "# Digital %02ld          = %s\n",
                          Hdr->DigitalList[i].Channel,
                          Hdr->DigitalList[i].Title );
                }
        
        fprintf( fp, "# AtodWordSize        = %ld\n", Hdr->AtodWordSize);
        fprintf( fp, "# AtodDataCoding      = %ld",   Hdr->AtodDataCoding);
        fprintf( fp, " (%s)\n", SrDatDataCodingString(Hdr->AtodDataCoding) );
        fprintf( fp, "# AtodSamplingRate    = %lf\n", Hdr->AtodSamplingRate);
        fprintf( fp, "# AtodSpsMeasured     = %lf\n", Hdr->AtodSamplingRateMeasured);
        fprintf( fp, "# AtodGain            = %lf\n", Hdr->AtodGain);
        fprintf( fp, "# AtodFilterCoeff     = %lf\n", Hdr->AtodFilterCoeff);
        fprintf( fp, "# AtodFilterScale     = %lf\n", Hdr->AtodFilterScale);

        fprintf( fp, "# TriggerType         = %ld",   Hdr->TriggerType);
        fprintf( fp, " (%s)\n", SrDatTriggerTypeString(Hdr->TriggerType) );
        fprintf( fp, "# TriggerParm1        = %ld", Hdr->TriggerParm1);
        fprintf( fp, " %s\n", PrepareTriggerParm1String(Hdr->TriggerType, Hdr->TriggerParm1) );
        fprintf( fp, "# TriggerParm2        = %ld", Hdr->TriggerParm2);
        fprintf( fp, " %s\n", PrepareTriggerParm2String(Hdr->TriggerType, Hdr->TriggerParm2) );
        
        fprintf( fp, "# CounterType         = %ld", Hdr->CounterType);
        fprintf( fp, " (%s)\n", SrDatCounterTypeString(Hdr->CounterType) );
        fprintf( fp, "# CounterFreqHi       = %ld\n", Hdr->CounterFreqHi);
        fprintf( fp, "# CounterFreq         = %ld\n", Hdr->CounterFreq);
        
        fprintf( fp, "# Counter2Type        = %ld", Hdr->Counter2Type);
        fprintf( fp, " (%s)\n", SrDatCounterTypeString(Hdr->Counter2Type) );
        fprintf( fp, "# Counter2FreqHi      = %ld\n", Hdr->Counter2FreqHi);
        fprintf( fp, "# Counter2Freq        = %ld\n", Hdr->Counter2Freq);
        
        fprintf( fp, "# LocationLatitude    = %lf\n", Hdr->LocationLatitude);
        fprintf( fp, "# LocationLongitude   = %lf\n", Hdr->LocationLongitude);
        fprintf( fp, "# LocationAltitude    = %lf\n", Hdr->LocationAltitude);
        fprintf( fp, "# LocationSource      = %ld",   Hdr->LocationSource);
        fprintf( fp, " (%s)\n", SrDatSourceString(Hdr->LocationSource) );
        
        for ( i = 0 ; i < SRDAT_DIM_TIME ; i++ ) {
                fprintf( fp, "# TimePoint[%d]:\n", i);
                
                fprintf( fp, "#    Type             = %ld",   Hdr->TimePoint[i].Type);
                fprintf( fp, " (%s)\n", SrDatTimeTypeString(Hdr->TimePoint[i].Type) );
                fprintf( fp, "#    SampleFromRun    = %ld\n", Hdr->TimePoint[i].SampleFromRun);
                fprintf( fp, "#    SampleFromFile   = %ld\n", Hdr->TimePoint[i].SampleFromFile);
                fprintf( fp, "#    GpsMarkIndex     = %ld\n", Hdr->TimePoint[i].GpsMarkIndex);
                
                fprintf( fp, "#    Year             = %ld\n", Hdr->TimePoint[i].Year);
                fprintf( fp, "#    Month            = %ld\n", Hdr->TimePoint[i].Month);
                fprintf( fp, "#    Day              = %ld\n", Hdr->TimePoint[i].Day);
                fprintf( fp, "#    YMDSource        = %ld",   Hdr->TimePoint[i].YMDSource);
                fprintf( fp, " (%s)\n", SrDatSourceString(Hdr->TimePoint[i].YMDSource) );
                
                fprintf( fp, "#    Hour             = %ld\n", Hdr->TimePoint[i].Hour);
                fprintf( fp, "#    Minute           = %ld\n", Hdr->TimePoint[i].Minute);
                fprintf( fp, "#    Second           = %ld\n", Hdr->TimePoint[i].Second);
                fprintf( fp, "#    Microsecond      = %ld\n", Hdr->TimePoint[i].Microsecond);
                fprintf( fp, "#    HMSSource        = %ld",   Hdr->TimePoint[i].HMSSource);
                fprintf( fp, " (%s)\n", SrDatSourceString(Hdr->TimePoint[i].HMSSource) );
                
                fprintf( fp, "#    Unused1          = %ld\n", Hdr->TimePoint[i].Unused1);
                }

        fprintf( fp, "# LocationUnused1     = %ld\n", Hdr->LocationUnused1);
        fprintf( fp, "# DigitalUnused1      = %ld\n", Hdr->DigitalUnused1);
        fprintf( fp, "# TriggerUnused1      = %ld\n", Hdr->TriggerUnused1);

}

//------------------------------------------------------------------------------
// ROUTINE: PrepareTriggerParm1String
// PURPOSE: Translate trigger parameter 1 integer constant into a string.
//------------------------------------------------------------------------------
char *PrepareTriggerParm1String( long TriggerType, long TriggerParm1 ) {
        
        if ( TriggerType == SRDAT_TRIGGERTYPE_THRESHOLD )
                return( "(Threshold)" );

        else if ( TriggerType == SRDAT_TRIGGERTYPE_DIGLEVEL ||
                  TriggerType == SRDAT_TRIGGERTYPE_DIGEDGE )
                return( "(Digital Input Line)" );
        else
                return( "" );
}


//------------------------------------------------------------------------------
// ROUTINE: PrepareTriggerParm2String
// PURPOSE: Translate trigger parameter 2 integer constant into a string.
//------------------------------------------------------------------------------
char *PrepareTriggerParm2String( long TriggerType, long TriggerParm2 ) {

        long channelBits;
        int  i, nchan;
        
        static char TmpStr[30];

        
        if ( TriggerType == SRDAT_TRIGGERTYPE_THRESHOLD ) {

                if ( TriggerParm2 == 0 )
                        sprintf( TmpStr, "(No channels)" );
                else {
                        channelBits = TriggerParm2;
                        nchan       = 0;
                        for ( i = 0 ; i < 32 ; i++ ) {
                                if ( channelBits & 0x1 )
                                        nchan++;
                                channelBits /= 2;
                                }
                        sprintf( TmpStr, "= 0x%lX (%d channels)",
                                 TriggerParm2, nchan );
                        }
                }

        else if ( TriggerType == SRDAT_TRIGGERTYPE_DIGLEVEL )
                sprintf( TmpStr, "(%s)", SrDatTriggerLevelString( TriggerParm2 ) );

        else if ( TriggerType == SRDAT_TRIGGERTYPE_DIGEDGE )
                sprintf( TmpStr, "(%s)", SrDatTriggerEdgeString( TriggerParm2 ) );

        else
                sprintf( TmpStr, "" );
        
        return( TmpStr );
}

//------------------------------------------------------------------------------
// ROUTINE: ShowTag
// PURPOSE: Write record tag information to the ASCII output file.
//------------------------------------------------------------------------------
void ShowTag( FILE *fp, SRDAT_TAG *Tag ) {

        char cTagType[5];
        char *TypeStr;

        
        // Check tag size
        
        if (sizeof(*Tag) != SRDAT_TAGSIZE)
                ErrorExec( "Tag size not right" );

        strncpy( cTagType,((char *)(&Tag->TagId)), 4);
        cTagType[4] = '\0';


        // Set up tag type string

        switch ( Tag->TagId ) {
                case SRDAT_TAGID_USBPACKET:
                        TypeStr = "USBxCH packet data";
                        break;
                case SRDAT_TAGID_USBANALOG:
                        TypeStr = "USBxCH analog data";
                        break;
                case SRDAT_TAGID_USBSERIAL:
                        TypeStr = "USBxCH serial data";
                        break;
                case SRDAT_TAGID_USBEQUIP:
                        TypeStr = "USBxCH equipment data";
                        break;
                case SRDAT_TAGID_EOF:
                        TypeStr = "End of File";
                        break;
                default:
                        TypeStr = "Unknown";
                        break;
                }
        

        // Show tag info

        fprintf( fp, "\nRecord Type: %s,  Tag: %s,  Bytes: %lu\n",
                                             TypeStr, cTagType, Tag->Nbytes );
        
}

//------------------------------------------------------------------------------
// ROUTINE: ShowData
// PURPOSE: Write record data to the ASCII output file.  The format of the data
//          depends on the record type.
//------------------------------------------------------------------------------
void ShowData( FILE *fpOutData, FILE *fpOutNmea, SRDAT_HDR *Hdr, void *Buffer,
               long DataType, unsigned long Nbytes ) {

        int   PacketSize, PacketType, Error;
        int   PacketCount, PacketCountSample, PacketCountStatus;
        long  ToShow;
        char *Unknown;
        SRUSBXCH_PACKET        *UsbPacket;
        SRDAT_USBXCH_ANALOG    *UsbAnalog;
        SRDAT_USBXCH_SERIAL    *UsbSerial;
        SRDAT_USBXCH_EQUIPMENT *UsbEquip;



        // Show data

        switch ( DataType ) {
                
            case SRDAT_TAGID_USBANALOG:

                    UsbAnalog = (SRDAT_USBXCH_ANALOG*)Buffer;
                    ToShow    = Nbytes;

                    // For all bytes to show, output data

                    while ( ToShow > 0 ) {

                            DisplayAnalogData( fpOutData, UsbAnalog, DP.CurrentPt, Hdr, 0 );

                            ToShow -= sizeof(SRDAT_USBXCH_ANALOG);
                            UsbAnalog++;
                            DP.CurrentPt++;

                            } // end of ToShow > 0

                    break;



            case SRDAT_TAGID_USBSERIAL:

                    UsbSerial = (SRDAT_USBXCH_SERIAL*)Buffer;
                    ToShow    = Nbytes;

                    while ( ToShow > 0 ) {

                            DisplaySerialData( fpOutNmea, UsbSerial );
                            
                            UsbSerial++;
                            ToShow -= sizeof(SRDAT_USBXCH_SERIAL);

                            } // end while ToShow > 0
                    break;


            case SRDAT_TAGID_USBEQUIP:

                    UsbEquip = (SRDAT_USBXCH_EQUIPMENT*)Buffer;
                    ToShow   = Nbytes;

                    while ( ToShow > 0 ) {

                            DisplayEquipData( fpOutNmea, UsbEquip );

                            UsbEquip++;
                            ToShow -= sizeof(SRDAT_USBXCH_EQUIPMENT);

                            } // end while ToShow > 0
                    break;

                    

            case SRDAT_TAGID_USBPACKET:

                    if ( !DP.Debug )
                            break;

                    UsbPacket = (SRUSBXCH_PACKET *)Buffer;

                    ToShow            = Nbytes;
                    PacketSize        = sizeof( SRUSBXCH_PACKET );
                    PacketCount       = 0;
                    PacketCountSample = 0;
                    PacketCountStatus = 0;

                    fprintf( fpOutData,"\n" );
                    fprintf( fpOutData,"%lu bytes -> %d packets\n\n", Nbytes, (((int)Nbytes)/PacketSize) );

                    while ( ToShow > 0 ) {

                            // Display packets

                            SrUsbXchPacketShow( fpOutData, UsbPacket, PacketCount, 1, &Error );

                            // Increment corresponding packet count

                            PacketType = SrUsbXchPacketType( UsbPacket );
                            if ( PacketType == SRDAT_USBPACKET_TYPE_SAMPLE )
                                    PacketCountSample++;
                            else if ( PacketType == SRDAT_USBPACKET_TYPE_STATUS )
                                    PacketCountStatus++;
                            
                            PacketCount++;
                            UsbPacket++;
                            ToShow  -= PacketSize;
                            }

                    fprintf( fpOutData, "Orig Packet Counts: Total %d, Data %d(*2), Status %d\n",
                             PacketCount, PacketCountSample, PacketCountStatus );

                    
                    UsbPacket = (SRUSBXCH_PACKET *)Buffer;

                    DisplayConvertedPackets( fpOutData, UsbPacket, &PacketCount,
                                             &PacketCountSample, &PacketCountStatus, Hdr );
                    break;




            default:

                    if ( !DP.Debug )
                            break;

                    Unknown = (char*)Buffer;
                    
                    DisplayUnknownData( fpOutData, Unknown, DataType, 10, Nbytes );
                        
                    break;

            } // end switch
}

//------------------------------------------------------------------------------
// ROUTINE: DisplayAnalogData
// PURPOSE: Output Analog structure.
//------------------------------------------------------------------------------
void DisplayAnalogData( FILE *fp, SRDAT_USBXCH_ANALOG *UsbxchAnalogData,
                        int TotalPt, SRDAT_HDR *Hdr, int ShowTitle ) {

        int    i, pos, Nchannels, PowerGoodC, PowerGoodH;
        char   TimeStrSec[MAXTIMESTR], TimeStrFracSec[MAXTIMESTR];
        char   TimeStrYmd[MAXTIMESTR], TimeStrHms[MAXTIMESTR];
        char  *Fmt, *ObcSetStr, ChannelString[MAXSTR], DigitalString[MAXSTR];
        double TempC, TempF;

        // Can pass null for analog data if only showing title
        
        if ( !fp || !Hdr || ( !ShowTitle && !UsbxchAnalogData ) )
                return;


        Nchannels = Hdr->ChannelsAnalog;
        
        
        if ( ShowTitle ) {

                // Prepare string with channel names

                if ( DP.CSV )
                        Fmt = ",%s";
                else
                        Fmt = "  %10s";

                pos = 0;
                for ( i = 0 ; i < Nchannels ; i++ ) {
                        sprintf( &ChannelString[pos], Fmt, Hdr->ChannelList[i].Title );
                        pos = strlen( ChannelString );
                        }


                // Write header/title info according to selected output format

                if ( !DP.CSV ) { // fixed width
                        
                        fprintf( fp, "# Dat2Asc\n# format=%s\n#\n", OutputFormatName[DP.OutFormat] );

                        if ( DP.Debug )
                                ShowHdr( fp, Hdr );
                        }
                
                if ( DP.OutFormat == OUTPUT_FORMAT_CSV_DBG )
                        fprintf( fp, "Pt#%s,D0,D1,D2,D3,PwrH,DegC,DegF,Seconds,Date(YMD),Time(HMS),FracSec,PwrC,San,Pps,Nmea,OBC,ObcSet\n", ChannelString );
                else if ( DP.OutFormat == OUTPUT_FORMAT_CSV_HEX )
                        fprintf( fp, "Pt#%s,Dig,PwrH,DegC,DegF,Seconds,Date(YMD),Time(HMS),FracSec\n", ChannelString );
                else if ( DP.OutFormat == OUTPUT_FORMAT_CSV )
                        fprintf( fp, "Pt#%s,D0,D1,D2,D3,PwrH,DegC,DegF,Seconds,Date(YMD),Time(HMS),FracSec\n", ChannelString );
                else if ( DP.OutFormat == OUTPUT_FORMAT_FW_DBG )
                        fprintf( fp, "#\n# Data:\n#   Pt#  %s  Dig PwrH DegC  DegF   Time in Seconds     Time as YMD HMS    Frac Sec PwrC San  Pps Nmea    OBC\n", ChannelString );
                else if ( DP.OutFormat == OUTPUT_FORMAT_FW_HEX )
                        fprintf( fp, "#   Pt#  %s  Dig PwrH DegC  DegF   Time in Seconds     Time as YMD HMS    Frac Sec\n", ChannelString );
                else // if ( DP.OutFormat == OUTPUT_FORMAT_FW )
                        fprintf( fp, "#   Pt#  %s D0 D1 D2 D3 PwrH DegC  DegF   Time in Seconds     Time as YMD HMS    Frac Sec\n", ChannelString );

                return;

                } // end if ShowTitle


        

        // Compute values for output below
        
        PowerGoodH = SrUsbXchIsPowerBitGood( DP.TS0.PowerInfo,
                                             USBXCH_POWER_HISTORY );
        
        PowerGoodC = SrUsbXchIsPowerBitGood( DP.TS0.PowerInfo,
                                             USBXCH_POWER_CURRENT );
        
        SrDatUsbTemperatureCompute( DP.TS0.Temperature, &TempC, &TempF );

        ComputeTime( TotalPt, TimeStrSec, TimeStrYmd, TimeStrHms, TimeStrFracSec );


        
        // Prepare analog channel data

        if ( DP.OutFormat == OUTPUT_FORMAT_FW_HEX ||
             DP.OutFormat == OUTPUT_FORMAT_FW_DBG )     
                Fmt = "0x%08lX  ";
        else if ( DP.OutFormat == OUTPUT_FORMAT_CSV_HEX )
                Fmt = ",0x%08lX";
        else if ( DP.OutFormat == OUTPUT_FORMAT_CSV ||
                  DP.OutFormat == OUTPUT_FORMAT_CSV_DBG )
                Fmt = ",%+ld";
        else // if ( DP.OutFormat == OUTPUT_FORMAT_FW || unknown
                Fmt = " %+9ld  ";
        pos = 0;
        for ( i = 0 ; i < Nchannels ; i++ ) {
                sprintf( &ChannelString[pos], Fmt, UsbxchAnalogData->AnalogData[i] );
                pos = strlen( ChannelString );
                }


        // Prepare digital data

        if ( DP.Hex ) {
                if ( DP.CSV )
                        Fmt = ",0x%02X";           // CSV_HEX
                else
                        Fmt = "0x%02X ";           // FW_HEX, FW_DBG
                sprintf( DigitalString, Fmt, UsbxchAnalogData->DigitalIn );
                }
        else {
                if ( DP.CSV )
                        Fmt = ",%d,%d,%d,%d";      // CSV, CSV_DBG
                else
                        Fmt = "%d  %d  %d  %d  ";  // FW
                sprintf( DigitalString, Fmt, (UsbxchAnalogData->DigitalIn & 0x01),
                                            ((UsbxchAnalogData->DigitalIn & 0x02)>>1),
                                            ((UsbxchAnalogData->DigitalIn & 0x04)>>2),
                                            ((UsbxchAnalogData->DigitalIn & 0x08)>>3)
                         );
                }


        // Prepare ObcSet string

        if ( DP.CSV ) {
                if ( DP.Debug ) {
                        if ( UsbxchAnalogData->OnBoardCount != 0x0 )
                                ObcSetStr = "<<"; // DBGCSV(#)
                        else
                                ObcSetStr = "--"; // DBGCSV(0)
                        }
                // CSV, HEXCSV
                }
        else {
                if ( DP.Debug  &&  UsbxchAnalogData->OnBoardCount != 0x0 )
                        ObcSetStr = "<<"; // DBGFIX(#)
                else
                        ObcSetStr = "  "; // FIXED, HEXFIX, DBGFIX(0)
                }


        
        // Output data as requested by selected format 

        if ( DP.OutFormat == OUTPUT_FORMAT_CSV_DBG )
                fprintf( fp, "%ld%s%s,%1d,%.1lf,%.1lf,%s,%s,%s,%s,%1d,%d,%d,%d,%d,%s\n",
                         TotalPt, ChannelString, DigitalString,
                         PowerGoodH, TempC, TempF,
                         TimeStrSec, TimeStrYmd, TimeStrHms, TimeStrFracSec,
                         PowerGoodC,
                         UsbxchAnalogData->Sanity,
                         UsbxchAnalogData->PpsToggle,
                         UsbxchAnalogData->GGACount,
                         UsbxchAnalogData->OnBoardCount,
                         ObcSetStr
                       );

        else if ( DP.OutFormat == OUTPUT_FORMAT_CSV_HEX  ||
                  DP.OutFormat == OUTPUT_FORMAT_CSV )
                fprintf( fp, "%ld%s%s,%1d,%.1lf,%.1lf,%s,%s,%s,%s\n",
                         TotalPt, ChannelString, DigitalString,
                         PowerGoodH, TempC, TempF,
                         TimeStrSec, TimeStrYmd, TimeStrHms, TimeStrFracSec
                       );
        
        else if ( DP.OutFormat == OUTPUT_FORMAT_FW_DBG )
                fprintf( fp, "%010ld %s%s %1d %5.1lf %5.1lf  %s  %s %s  %s  %1d  0x%02X 0x%02X 0x%02X 0x%08X %s\n",
                         TotalPt, ChannelString, DigitalString,
                         PowerGoodH, TempC, TempF,
                         TimeStrSec, TimeStrYmd, TimeStrHms, TimeStrFracSec,
                         PowerGoodC,
                         UsbxchAnalogData->Sanity,
                         UsbxchAnalogData->PpsToggle,
                         UsbxchAnalogData->GGACount,
                         UsbxchAnalogData->OnBoardCount,
                         ObcSetStr
                       );

        else // if ( DP.OutFormat == OUTPUT_FORMAT_FW_HEX ) ||
             // if ( DP.OutFormat == OUTPUT_FORMAT_FW )
                fprintf( fp, "%010ld %s%s %1d %5.1lf %5.1lf  %s  %s %s  %s\n",
                         TotalPt, ChannelString, DigitalString,
                         PowerGoodH, TempC, TempF,
                         TimeStrSec, TimeStrYmd, TimeStrHms, TimeStrFracSec
                       );
}

//------------------------------------------------------------------------------
// ROUTINE: DisplaySerialData
// PURPOSE: Output Serial structure.
//------------------------------------------------------------------------------
void DisplaySerialData( FILE *fp, SRDAT_USBXCH_SERIAL *UsbxchSerialData ) {

        int i;


        if ( !fp || !UsbxchSerialData )
                return;
        
        // Display serial NMEA data

       fprintf( fp, "%u\n", UsbxchSerialData->PpsEventNum );

        if ( UsbxchSerialData->NmeaCount > 0 ) {

                for ( i = 0 ; i < UsbxchSerialData->NmeaCount ; i++ ) {
                        fprintf( fp,"%s",
                                 &UsbxchSerialData->NmeaMsg[i*SRDAT_NMEA_MAX_SIZE]);
                        } // end for

                } // end if

                        
}

//------------------------------------------------------------------------------
// ROUTINE: DisplayEquipData
// PURPOSE: Output equipment structure to specified file.
//------------------------------------------------------------------------------
void DisplayEquipData( FILE *fp, SRDAT_USBXCH_EQUIPMENT *UsbxchEquipData ) {

        int    Empty, PartFull, Overflow;
        double TempC, TempF;

        if ( !fp || !UsbxchEquipData )
                return;


        // Convert raw temps to degrees C and F

        SrDatUsbTemperatureCompute( UsbxchEquipData->Temperature, &TempC, &TempF );
        SrUsbXchDramFlagsSplit( UsbxchEquipData->DramFlags, &Empty, &PartFull, &Overflow );


        // Display equipment data

        fprintf( fp, "%u\n",                     UsbxchEquipData->PpsEventNum );
        fprintf( fp, "EQUIP: Power %1d, ",       UsbxchEquipData->VoltageGood );
        fprintf( fp, "Temp 0x%04X, ",            UsbxchEquipData->Temperature );
        fprintf( fp, "DegC %5.1f, DegF %5.1f, ", TempC, TempF );
        fprintf( fp, "DramFlags EM/PF/OV %1X/%1X/%1X\n",
                                                 Empty, PartFull, Overflow );
}


//------------------------------------------------------------------------------
// ROUTINE: DisplayUnknownData
// PURPOSE: Output raw bytes of unknown structure 10 per line.
//------------------------------------------------------------------------------
void DisplayUnknownData( FILE *fp, char *Unknown,
                         int DataType, int MaxPerLine, unsigned long Nbytes ) {
        
        int           NumPerLine;
        unsigned long ii;
        
        fprintf( fp,"Unknown data type 0x%lX\n", DataType );

        NumPerLine = 0;

        for ( ii = 0 ; ii < Nbytes ; ii++ ) {
                fprintf( fp, "%02X ", (Unknown[ii] & 0xFF) );
                NumPerLine++;
                if (NumPerLine > MaxPerLine) {
                        fprintf( fp, "\n" );
                        NumPerLine = 1;
                        }
                }
        fprintf( fp, "\n" );
}

//------------------------------------------------------------------------------
// ROUTINE: DisplayConvertedPackets
// PURPOSE: Convert packets and output the results to specified file.
//------------------------------------------------------------------------------
void DisplayConvertedPackets( FILE *fp,
                              SRUSBXCH_PACKET *UsbPacketArray,
                              int *PacketCount,
                              int *PacketCountSample,
                              int *PacketCountStatus,
                              SRDAT_HDR *Hdr ) {

        int           i, Error;
        unsigned long NanalogReturned, NserialReturned, NequipReturned;
        unsigned long NanalogRequested, NserialRequested, NequipRequested;
        unsigned long NpacketsProcessed, NpacketsProcessedData, NpacketsProcessedStatus;
        unsigned long Nsamples;

        static int CurrentLastMsg = -1;
        static SRDAT_USBXCH_ANALOG    UsbXchAnalogData[MAXANALOG];
        static SRDAT_USBXCH_SERIAL    UsbXchSerialData[MAXSERIAL];
        static SRDAT_USBXCH_EQUIPMENT UsbXchEquipData[MAXEQUIP];


         if ( !fp || !UsbPacketArray || !PacketCount || 
              !PacketCountSample || !PacketCountStatus || !Hdr )
                 return;


        NanalogRequested = *PacketCountSample * 2;
        if ( NanalogRequested > MAXANALOG ) {
                printf( "WARNING: Size of data is larger than allocated SRDAT_USBXCH_ANALOG array\n" );
                printf( "Using only %lu points not all %lu\n", MAXANALOG, NanalogRequested );
                NanalogRequested = MAXANALOG;
                }

        NserialRequested = *PacketCountStatus;
        if ( NserialRequested > MAXSERIAL ) {
                printf( "WARNING: Size of data is larger than allocated SRDAT_USBXCH_SERIAL array\n" );
                printf( "Using only %lu points not all %lu\n", MAXSERIAL, NserialRequested );
                NserialRequested = MAXSERIAL;
                }

        NequipRequested = *PacketCountStatus;
        if ( NequipRequested > MAXEQUIP ) {
                printf( "WARNING: Size of data is larger than allocated SRDAT_USBXCH_EQUIPMENT array\n" );
                printf( "Using only %lu points not all %lu\n", MAXEQUIP, NequipRequested );
                NequipRequested = MAXEQUIP;
                }


        // Initialize SerialData and EquipData structures

        for ( i = 0 ; i < (int)NserialRequested ; i++ )
                SrDatUsbSerialInit( &UsbXchSerialData[i] );

        for ( i = 0 ; i < (int)NequipRequested ; i++ )
                SrDatUsbEquipInit( &UsbXchEquipData[i] );

        Nsamples = SrUsbXchConvertPackets(
                                            UsbPacketArray,
                                            (unsigned long)(*PacketCount),
                                            0,                              // Don't check sanity
                                            &CurrentLastMsg,

                                            UsbXchAnalogData,
                                            NanalogRequested,
                                            &NanalogReturned,

                                            UsbXchSerialData,
                                            NserialRequested,
                                            &NserialReturned,

                                            UsbXchEquipData,
                                            NequipRequested,
                                            &NequipReturned,

                                            &NpacketsProcessed,
                                            &NpacketsProcessedData,
                                            &NpacketsProcessedStatus,
                                            &Error
                                            );

        if ( NpacketsProcessed != (unsigned long)(*PacketCount) )
                printf("ShowData processed %lu packets from a total of %d\n",
                       NpacketsProcessed, *PacketCount );

        fprintf( fp, "Packet Counts: Total %d, Data %d, Status %d\n",
                 *PacketCount, *PacketCountSample, *PacketCountStatus );
        fprintf( fp, "Nread Counts: Analog %lu, Serial %lu, Equip %lu\n",
                 NanalogReturned, NserialReturned, NequipReturned );

//FIX - get consistent with size of data, ints or unsigned longs

        fprintf( fp, "   Pt#         CH00        CH01        CH02        CH03    San  Pps  Dig  GGA     OBC\n");

        for ( i = 0 ; i < (int)NanalogReturned ; i++ ) {
                DisplayAnalogData( fp, UsbXchAnalogData,
                                   DP.CurrentPt, Hdr, 0 );
                DP.CurrentPt++;
                }

        for ( i = 0 ; i < (int)NserialReturned ; i++ )
                DisplaySerialData( fp, UsbXchSerialData );

        for ( i = 0 ; i < (int)NequipReturned ; i++ )
                DisplayEquipData( fp, UsbXchEquipData );
}        

//------------------------------------------------------------------------------
// ROUTINE: ErrorCmd
// PURPOSE: Print a command line error message and quit.
//------------------------------------------------------------------------------
void ErrorCmd( char *msg ) {

        // Print a command line error message and quit.

        printf(

               "ERROR: Command line usage : %s \n\n ... Dat2Asc /? for help\n",
               msg

               );

        exit( 1 );  // < quit at errorlevel 1.

}

//------------------------------------------------------------------------------
// ROUTINE: ErrorExec
// PURPOSE: An execution error occurred.
//------------------------------------------------------------------------------
void ErrorExec( char *msg ) {

        // A fatal execution error occurred.

        printf( "\nERROR: %s \n\n", msg );

        exit( 2 );  // < quit at errorlevel 2.

}

//------------------------------------------------------------------------------
// ROUTINE: WarningExec
// PURPOSE: A non-fatal execution error occurred.
//------------------------------------------------------------------------------
void WarningExec( char *msg ) {

        // A non-fatal execution error occurred.

        printf( "\nWARNING: %s \n\n", msg );

}

//------------------------------------------------------------------------------
// ROUTINE: ProcessCommandLine
// PURPOSE: Extract information provided by the user on the program command line.
//          See the UsageMsg structure defined above for allowed options.
//------------------------------------------------------------------------------
int ProcessCommandLine( int argc, char *argv[] ) {

        int  i, ifmt;
        char MsgBuffer[MAXSTR];


        
        // No command line arguments gives the usage message.

        if ( argc == 1 )

                Usage( );


        // Check for too many arguments.

        if ( argc > 3 )

                ErrorCmd( "too many command line arguments" );


        // Scan for valid arguments.

        for ( i = 1 ; i < argc ; i++ ) {


                // Check for usage message request.

                if ( StrEqu( argv[i], "/?") )

                        Usage( );


                // Check for any allowed format.

                else if ( ifmt = MatchFormat( argv[i] ) ) {
                        
                        if ( DP.OutFormat == OUTPUT_FORMAT_NONE )
                                DP.OutFormat = ifmt;

                        else
                                ErrorCmd( "multiple format arguments" );
                        }

                

                // If not one of the above, it must be a filename ...

                else {

                        if ( ParseFileName( argv[i] ) ) {

                                if ( DP.InputFileFlag == 0 )

                                        DP.InputFileFlag = 1;

                                else

                                    ErrorCmd( "multiple filename arguments" );
                                }

                        else {

                                sprintf( MsgBuffer, "filename %s is too long",

                                                         ChopLong( argv[i] )

                                                         );

                                    ErrorCmd( MsgBuffer );

                                }

                        }

                }



        // An input filename is *required* on the command line.

        if ( !DP.InputFileFlag )

                ErrorCmd( "an input filename is required ..." );


        // If the input file base name is eight digits, turn on sequential.

        if ( StrIsEightDecimal( DP.InputFileBase ) ) {

                DP.Sequential = 1;  // < set the Sequential processing flag on


                // Get the starting filename as a binary integer.

                sscanf( DP.InputFileBase, "%d", &DP.nSequential );

                }
        



        // Fill input file name and verify that it exists

        sprintf( DP.InputFileName,  "%s.%s", DP.InputFileBase, DP.InputFileExt );
        
        if ( !SrFileExist( DP.FullPathName ) ) {
                sprintf( MsgBuffer, "File %s does not exist", DP.FullPathName );
                ErrorCmd( MsgBuffer );
                }



        // An output format is *required* on the command line.

        if ( DP.OutFormat == OUTPUT_FORMAT_NONE )

                ErrorCmd( "an output format is required ..." );


        // Set csv, hex, and debug helper variables based on selected format

        if ( DP.OutFormat == OUTPUT_FORMAT_CSV    ||
             DP.OutFormat == OUTPUT_FORMAT_CSV_HEX ||
             DP.OutFormat == OUTPUT_FORMAT_CSV_DBG )
                DP.CSV = 1;

        if ( DP.OutFormat == OUTPUT_FORMAT_FW_HEX ||
             DP.OutFormat == OUTPUT_FORMAT_CSV_HEX ||
             DP.OutFormat == OUTPUT_FORMAT_FW_DBG )  // fixed debug mode uses hex too
                DP.Hex = 1;
        
        if ( DP.OutFormat == OUTPUT_FORMAT_FW_DBG ||
             DP.OutFormat == OUTPUT_FORMAT_CSV_DBG )
                DP.Debug = 1;


        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: Usage
// PURPOSE: Print program usage message to the screen.
//------------------------------------------------------------------------------
void Usage( void ) {

        int i;

        // Print the usage message.

        for ( i = 0 ; UsageMsg[i] ; i++ )

                printf( "%s\n", UsageMsg[i] );

        exit( 0 );  // < quit at error level 0.
}

//------------------------------------------------------------------------------
// ROUTINE: MatchFormat
// PURPOSE: Check if argument is one of the allowed format names.  If not, 
//          return 0.  Otherwise, return the index of that matched.
//------------------------------------------------------------------------------
int MatchFormat( char *argument ) {

        int i;

        if ( !argument )
                return( 0 );

        for ( i = OUTPUT_FORMAT_NONE+1 ; i < OUTPUT_FORMAT_MAX ; i++ ) {

                if ( StrEqu( argument, OutputFormatName[i] ) )
                        return( i );

                } // end for i

        return( 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: StrEqu
// PURPOSE: Compare for exact equality between two strings.
//------------------------------------------------------------------------------
int StrEqu( char *s1, char *s2 ) {

        while ( *s1 && *s2 && (*s1 == *s2) )

                { s1++; s2++; }

        return ( (*s1 == '\0') && (*s2 == '\0') ? 1 : 0 );

}

//------------------------------------------------------------------------------
// ROUTINE: StrIsEightDecimal
// PURPOSE: Check a string for being eight decimal digits.
//------------------------------------------------------------------------------
int StrIsEightDecimal( char *s ) {

        int i;

        i = 0;

        while ( isdigit( *s ) )

                { s++; i++; }

        return ( (*s == '\0') && (i == 8) ? 1 : 0 );

}

//------------------------------------------------------------------------------
// ROUTINE: ParseFileName
// PURPOSE: Parse an input full pathname, determining any path and extension.
//------------------------------------------------------------------------------
int ParseFileName( char *fullpath ) {

        char *s, *path, *ext;
        int i;


        // Copy the command line input file to DP.FullPathName[].
        //
        //  ... and check to make sure not too long
        //
        //

        for ( s = fullpath , i = 0 ; *s && (i < MAXFILENAME) ; s++ , i++)

                DP.FullPathName[i] = *s;

        DP.FullPathName[i] = '\0';

        if ( i > (MAXFILENAME - 80) )

                return( 0); // don't even bother if too long ...




        // Scan for the last '\' to mark the filename path.
        //
        //  ... allow both '\' and '/' for Windows/Linux compatibility
        //

        path = NULL;

        for ( s = DP.FullPathName ; *s ; s++ ) {

                if ( (*s == '\\') || (*s == '/') )

                        path = s;

                } // path now == NULL, or points to last '\' ...



        // Scan from the path to the last '.' to mark the extension.

        ext = NULL;

        for ( s = ( path ? path + 1 : DP.FullPathName ) ; *s ; s++ ) {

                if ( *s == '.' )

                        ext = s;

                } // ext now == NULL, or points to last '.' ...





        // Copy out the input path into DP.InputFilePath[].
        //
        // ... the trailing '\' is included in DP.InputFilePath[]
        //

        if ( path ) {

                for ( s = DP.FullPathName , i = 0 ; s <= path ; s++ , i++ )

                        DP.InputFilePath[i] = *s;

                DP.InputFilePath[i] = '\0';

                }
        else

                DP.InputFilePath[0] = '\0';




        // Copy out the base filename into InputFileBase[].

        s = ( path ? path + 1 : DP.FullPathName );

        if ( ext ) { // gather up to the extension ...

                for ( i = 0 ; s < ext ; s++ , i++ )

                        DP.InputFileBase[i] = *s;

                }

        else { // ext == NULL

                for ( i = 0 ; *s ; s++ , i++ )

                        DP.InputFileBase[i] = *s;

                }

        DP.InputFileBase[i] = '\0';




        // Copy out the filename extension into InputFileExt[].

        if ( ext ) { // gather from the extension onward ...

                for ( s = ext + 1 , i = 0 ; *s ; s++ , i++ )

                        DP.InputFileExt[i] = *s;

                DP.InputFileExt[i] = '\0';

                }
        else

                DP.InputFileExt[0] = '\0';





        // Convert DP.InputFileExt[] to lowercase.

        for ( s = DP.InputFileExt ; *s ; s++ )

                *s = tolower( *s);



        // Point pFullPathNameBase to the base part of FullPathName[].

        DP.pFullPathNameBase = ( path ? path + 1 : DP.FullPathName );



        // Form the InputFileName[].

        sprintf( DP.InputFileName, "%s.%s", DP.InputFileBase, DP.InputFileExt);



        // Return success.
        //
        // Almost anything will be accepted for a filename if it is not too
        // long ... fopen will further qualify later ...
        //
        //

        return( 1);


}

//------------------------------------------------------------------------------
// ROUTINE: PrintRunTimeSummary
// PURPOSE: Display the final PP parameters to run with.
//------------------------------------------------------------------------------
void PrintRunTimeSummary( void ) {


        // Display the Dat2Asc parameters they are running with: 

        printf( "Beginning Dat2Asc with:\n"                               );
        printf( "\n"                                                      );
        printf( " Input FilePath: %s\n",   ChopLong( DP.InputFilePath )   );
        printf( " Input FileName: %s\n",   ChopLong( DP.InputFileBase )   );
        printf( "  Input FileExt: %s\n",   ChopLong( DP.InputFileExt  )   );
        printf( "  Output Format: %s\n",   OutputFormatName[DP.OutFormat] );
        printf( "     Sequential: %s\n",   DP.Sequential ? "ON" : "OFF"   );

    if ( DP.Sequential )

        printf( "         nStart: %08d\n", DP.nSequential                 );

    else

        printf( "         nStart: not applicable ... \n"                  );
        printf( "\n"                                                      );
}


//------------------------------------------------------------------------------
// ROUTINE: ChopLong
// PURPOSE: If a string is too long, make a chopped copy for display ...
//------------------------------------------------------------------------------
char *ChopLong( char *s ) {

        int i; static char MsgBuffer[MAXSTR];


        // Copy the string up to CHOPLENGTH.

#define CHOPLENGTH  25

        for ( i = 0 ; (i < CHOPLENGTH) && *s ; i++ , s++ )

                MsgBuffer[i] = *s;


        // Add '~' if more than CHOPLENGTH.

        if ( i == CHOPLENGTH )

                MsgBuffer[i++] = '~';


        // NULL terminate.

        MsgBuffer[i] = '\0';


        // Return the chopped string.
                
        return( MsgBuffer );

}

//------------------------------------------------------------------------------
// ROUTINE: FileReadTimeSample
// PURPOSE: Read the time sample file to get partial timestamp info.  Returns
//          1 on success, 0 otherwise.
//------------------------------------------------------------------------------
int FileReadTimeSample( FILE *fp, TS *TsPartial ) {

        int    Nscan, PpsCount, ObcPt, ObcCount;
        double ExtraSec;
        char   Line[MAXSTR];

        // Error check arguments

        if ( !fp || !TsPartial )
                return( 0 );

        
        // Read one line

        if ( !fgets( Line, MAXSTR-1, fp ) )
                return( 0 );


        // Read sample time info

        Nscan = sscanf( Line, FMT_TSAMP, // format defined in Step 1
                        &PpsCount,
                        &ObcPt,
                        &ObcCount,
                        &ExtraSec );
        
        if ( Nscan == 4 ) {
                TsPartial->PpsCount = PpsCount;
                TsPartial->Sample   = ObcPt;
                TsPartial->ObcCount = ObcCount;
                TsPartial->ExtraSec = ExtraSec;
                return( 1 );
                }
        else
                return( 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: FileReadTimeStatus
// PURPOSE: Read the time status file to get partial timestamp info.  Returns
//          1 on success, 0 otherwise.
//------------------------------------------------------------------------------
int FileReadTimeStatus( FILE *fp, TS *TsPartial, NI *NmeaInfo ) {

        int    Nscan, GpsCount, FullGpsCount;
        char   Line[MAXSTR];
        NI     NmeaTmp;

        // Error check arguments

        if ( !fp || !TsPartial || !NmeaInfo )
                return( 0 );

        
        // Read one line

        if ( !fgets( Line, MAXSTR-1, fp ) )
                return( 0 );


        // Read status time info

        Nscan = sscanf( Line, FMT_TSTAT,          // format defined in Step 1
                        &GpsCount,                // PpsCount
                        &NmeaTmp.SecSince1970,    // SecSince 1970
                        &NmeaTmp.Nsat,            // Sat
                        &NmeaTmp.YmdSource,       // Ysrc
                        &NmeaTmp.HmsSource,       // Hsrc
                        &FullGpsCount,            // PpsEvent
                        &NmeaTmp.Year,            // Time as YMD HMS ...
                        &NmeaTmp.Month,
                        &NmeaTmp.Day,
                        &NmeaTmp.Hour,
                        &NmeaTmp.Minute,
                        &NmeaTmp.Second,
                        &NmeaTmp.MicroSecond
                      );
        
        
        if ( Nscan == 13 ) {
                TsPartial->PpsCount = GpsCount;
                *NmeaInfo = NmeaTmp;
                return( 1 );
                }
        else
                return( 0 );
}


//------------------------------------------------------------------------------
// ROUTINE: FileReadTimeEquip
// PURPOSE: Read the time Equip file to get partial timestamp info.  Returns
//          1 on success, 0 otherwise.
//------------------------------------------------------------------------------
int FileReadTimeEquip( FILE *fp, TS *TsPartial ) {

        int  Nscan, PpsCount, PowerInfo, Temperature;
        char Line[MAXSTR];

        // Error check arguments

        if ( !fp || !TsPartial )
                return( 0 );

        
        // Read one line

        if ( !fgets( Line, MAXSTR-1, fp ) )
                return( 0 );


        // Read sample time info

        Nscan = sscanf( Line, FMT_TSEQ, // format defined in Step 1
                        &PpsCount,
                        &PowerInfo,
                        &Temperature );
        
        if ( Nscan == 3 ) {
                TsPartial->PpsCount    = PpsCount;
                TsPartial->PowerInfo   = PowerInfo;
                TsPartial->Temperature = Temperature;
                return( 1 );
                }
        else
                return( 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: FileUpdateTimeSample
// PURPOSE: Read the time sample file repeatedly until it returns partial
//          timestamp info that matches the RequiredPpsCount.  Returns
//          1 on success, 0 otherwise.
//------------------------------------------------------------------------------
int FileUpdateTimeSample( FILE *fp, TS *TsPartial, int RequiredPpsCount ) {

        if ( !fp || !TsPartial )
                return( 0 );

        while ( TsPartial->PpsCount < RequiredPpsCount ) {

                if ( !FileReadTimeSample( fp, TsPartial ) ) {
                        WarningExec( "Can't update PpsCount from TimeSample file." );
                        return( 0 );
                        }
                }

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: FileUpdateTimeStatus
// PURPOSE: Read the time Status file repeatedly until it returns partial
//          timestamp info that matches the RequiredPpsCount.  Returns
//          1 on success, 0 otherwise.
//------------------------------------------------------------------------------
int FileUpdateTimeStatus( FILE *fp, TS *TsPartial, int RequiredPpsCount,
                          NI *NmeaInfo ) {

        if ( !fp || !TsPartial || !NmeaInfo )
                return( 0 );

        while ( TsPartial->PpsCount < RequiredPpsCount ) {

                if ( !FileReadTimeStatus( fp, TsPartial, NmeaInfo ) ) {
                        WarningExec( "Can't update PpsCount from TimeStatus file." );
                        return( 0 );
                        }
                }

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: FileUpdateTimeEquip
// PURPOSE: Read the time equip file repeatedly until it returns partial
//          timestamp info that matches the RequiredPpsCount.  Returns
//          1 on success, 0 otherwise.
//------------------------------------------------------------------------------
int FileUpdateTimeEquip( FILE *fp, TS *TsPartial, int RequiredPpsCount ) {

        if ( !fp || !TsPartial )
                return( 0 );

        while ( TsPartial->PpsCount < RequiredPpsCount ) {

                if ( !FileReadTimeEquip( fp, TsPartial ) ) {
                        WarningExec( "Can't update PpsCount from TimeEquip file." );
                        return( 0 );
                        }
                }

        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: ProcessRawRecords
// PURPOSE: Write raw header and record bytes to the ASCII output file.
//------------------------------------------------------------------------------
void ProcessRawRecords( void ) {

    //  FILE      *DP.fpInput;   << see DP struct for this variable
        FILE      *fpRawOut;

        SRDAT_HDR Hdr;
        SRDAT_TAG Tag;

        unsigned long nTotalRecords;
        unsigned long nSamples;
        unsigned long nSkip;
        unsigned long nChan;
        int           Skip;

        static  char RawBuffer[20000];



        // Open the input and output files.

        fpRawOut = FileOpen( DP.FileNameRawOut, "w"  );


        // Read header from the input and write to the output file.

        ReadInputFileHdr( &Hdr ); // opens DP.fpInput
        ShowHdr( fpRawOut, &Hdr );
        fprintf( fpRawOut, "\n" );
        fclose( DP.fpInput );


        // Read and write data records until reaching the end

        nTotalRecords = 0L;
        nSamples      = 0L;
        nSkip         = 0;
        nChan         = (unsigned long)Hdr.ChannelsAnalog;

        // Loop for all input files

        while ( ReadInputFileHdr( &Hdr ) ) {    // opens DP.fpInput

                fprintf( fpRawOut, "File %s:\n", DP.InputFileName );
                
                while ( ReadInputFileRecord( &Tag, RawBuffer, &Skip )  ) {

                        if ( Skip ) {
                                fprintf( fpRawOut, "Skipping record %d (%lu bytes, ID = %s)\n",
                                         nTotalRecords, Tag.Nbytes, IdString( Tag.TagId ) );
                                nSkip++;
                                continue;
                                }

                        else {
                                ShowRawData( fpRawOut,
                                             RawBuffer,
                                             Tag.TagId,
                                             Tag.Nbytes,
                                             nTotalRecords,
                                             nChan,
                                             &nSamples );

                                nTotalRecords++;

                                if ( Tag.TagId == SRDAT_TAGID_EOF )
                                        break;
                                }

                        } // end while ReadInputFileRecord

                // Close current input file

                fclose( DP.fpInput );


                // Prepare for next one

                DP.nSequential++;
                if ( !PrepareFileNames() )
                        break;
                
                } // end while ReadInputFileHdr
        

        // Close the output file.

        fclose( fpRawOut );



        // Print RAW summary info.

        printf( "\n" );

        printf( "FINISHED Raw processing with:     \n"                    );
        printf( "                                  \n"                    );
        printf( "     Total records processed: %lu \n", nTotalRecords     );
        printf( "             Records skipped: %lu \n", nSkip             );
        printf( "             Last input file: %s  \n",
                                             ChopLong( DP.InputFileName ) );
        printf( "\n\n" );
        
        printf( "SUCCESS: final ASCII data >> %s\n\n", DP.FileNameRawOut );


        // Quit the program.

        exit( 0 );

}

//------------------------------------------------------------------------------
// ROUTINE: ShowRawData
// PURPOSE: Output raw record info into the specifed and opened file.
//------------------------------------------------------------------------------
void ShowRawData( FILE *fpRawOut,
                  char *Buffer,
                  unsigned long TagId,
                  unsigned long Nbytes,
                  unsigned long RecordNum,
                  unsigned long nChan,
                  unsigned long *nSamples
                  ) {

        unsigned long           ii, jj, StartPt;
        int                     i, j, Npts, Nser, Nequip;
        int                     PowerHistory, PowerCurrent;
        double                  TempC,TempF;
        SRDAT_USBXCH_ANALOG    *UsbAnalogData;
        SRDAT_USBXCH_SERIAL    *UsbSerialData;
        SRDAT_USBXCH_EQUIPMENT *UsbEquipData;


        if ( !fpRawOut || !Buffer )
                return;

        if ( nSamples )
                StartPt = *nSamples;
        else
                StartPt = 0L;

        // Output record tag
        
        fprintf( fpRawOut, "Record %d: %lu bytes, ID = %s\n",
                 RecordNum, Nbytes, IdString( TagId ) );

        // Output raw record bytes

        j = 0;
        for ( ii = 0 ; ii < Nbytes ; ii++ ) {
                fprintf( fpRawOut, "%02X ", (Buffer[ii] & 0x00FF) );
                j++;
                if ( j > 19 ) {
                        fprintf( fpRawOut, "\n" );
                        j = 0;
                        }
                }

        fprintf( fpRawOut, "\n" );



        // Output record bytes formatted appropriate for the record type

        if ( TagId == SRDAT_TAGID_USBANALOG ) {

                Npts = Nbytes / sizeof( SRDAT_USBXCH_ANALOG );
                UsbAnalogData = (SRDAT_USBXCH_ANALOG*)Buffer;

                fprintf( fpRawOut, "\nAnalog Data:\nPt (total)    (record)  " );
                for ( jj = 0 ; jj < nChan ; jj++ )
                        fprintf( fpRawOut, "  Chan%02d  ", jj );
                fprintf( fpRawOut, "PpsEvent  San Pps Dig GGA   OBC\n" );
                
                
                for ( i = 0 ; i < Npts ; i++ ) {

                        // Show pt number

                        fprintf( fpRawOut, "%10d  %10d  ", (StartPt+i), i );
                        
                        // Show channels
                        
                        for ( jj = 0 ; jj < nChan ; jj++ )
                                fprintf( fpRawOut, "0x%06lX  ",
                                          (MASKVAL & UsbAnalogData[i].AnalogData[jj]) );

                        // Show extras

                        fprintf( fpRawOut, "%08X  %02X  %02X  %02X  %02X %08X",
                                  UsbAnalogData[i].PpsEventNum,
                                  UsbAnalogData[i].Sanity,
                                  UsbAnalogData[i].PpsToggle,
                                  UsbAnalogData[i].DigitalIn,
                                  UsbAnalogData[i].GGACount,
                                  UsbAnalogData[i].OnBoardCount
                               );
                        
                        if ( UsbAnalogData[i].OnBoardCount == 0L )
                                fprintf( fpRawOut, "\n" );
                        else
                                fprintf( fpRawOut, " <<\n" );

                        } // end for i < Npts

                // Keep track of accumulated points

                if ( nSamples )
                        *nSamples += (unsigned long)Npts;

                fprintf( fpRawOut, "\n" );
                
                } // TagId == USBANALOG



        else if ( TagId == SRDAT_TAGID_USBSERIAL ) {

                Nser = Nbytes / sizeof( SRDAT_USBXCH_SERIAL );
                UsbSerialData = (SRDAT_USBXCH_SERIAL*)Buffer;

                fprintf( fpRawOut, "\n" );
                
                for ( i = 0 ; i < Nser ; i++ ) {

                        fprintf( fpRawOut, "Serial Data with PpsEvent = %u (0x%X) and %d NMEA strings\n",
                                      UsbSerialData[i].PpsEventNum,
                                      UsbSerialData[i].PpsEventNum,
                                      UsbSerialData[i].NmeaCount);
                        
                            for ( j = 0 ; j < UsbSerialData[i].NmeaCount ; j++ ) {
                                    fprintf( fpRawOut, "%s",
                                       &UsbSerialData[i].NmeaMsg[j*SRDAT_NMEA_MAX_SIZE]);
                                    } // end for j < NmeaCount

                        fprintf( fpRawOut, "\n" );

                        } // end for i < Nser

                } // TagId == USBSERIAL



        else if ( TagId == SRDAT_TAGID_USBEQUIP ) {

                Nequip = Nbytes / sizeof( SRDAT_USBXCH_EQUIPMENT );
                UsbEquipData = (SRDAT_USBXCH_EQUIPMENT*)Buffer;

                fprintf( fpRawOut, "\n" );
                
                for ( i = 0 ; i < Nequip ; i++ ) {

                        SrDatUsbTemperatureCompute( UsbEquipData[i].Temperature, &TempC, &TempF );

                        if ( SrUsbXchIsPowerBitGood( UsbEquipData[i].VoltageGood,
                                USBXCH_POWER_HISTORY ) )
                                PowerHistory = 1;
                        else
                                PowerHistory = 0;

                        if ( SrUsbXchIsPowerBitGood( UsbEquipData[i].VoltageGood,
                                USBXCH_POWER_CURRENT ) )
                                PowerCurrent = 1;
                        else
                                PowerCurrent = 0;


                        fprintf( fpRawOut, "Equipment Data with PpsEvent = %u (0x%X)\n",
                                                 UsbEquipData[i].PpsEventNum,
                                                 UsbEquipData[i].PpsEventNum );
                        fprintf( fpRawOut, "   Voltage Good = %d  (PwrH %d, PwrC %d)\n",
                                                 UsbEquipData[i].VoltageGood,
                                                 PowerHistory, PowerCurrent );
                        fprintf( fpRawOut, "   Temperature  = 0x%04X (Degrees %5.1lf C, %5.1lf F)\n",
                                                 UsbEquipData[i].Temperature,
                                                 TempC, TempF );
                        fprintf( fpRawOut, "\n" );

                        } // end for i < Nequip

                } // TagId == USBEQUIP
        
}

//------------------------------------------------------------------------------
// ROUTINE: IdString
// PURPOSE: Convert provided ID into
//------------------------------------------------------------------------------
char *IdString( unsigned long Id ) {

        strncpy( DP.cDatId,((char *)(&Id)), 4);
        DP.cDatId[4] = '\0';

        return( DP.cDatId );
}
