// FILE: DigitalIo.c
// COPYRIGHT: (c), Symmetric Research, 2009-2010
//
// This program demonstrates writing and reading the digital I/O bits
// that are available on the DB25 connector on the right hand side of the
// front panel of the USBxCH.
//
// Digital output bits on the USBxCH are latched by the hardware.  Write
// various output values with this program and test them with a voltmeter
// after the write.  Because they are latched they will remain stable.
// Output values are standard full scale 0v and +5v TTL.
//
// Another easy way to see the digital output is to wire up a solder cup
// D-Shell or ar TRM25E terminal board with 4 LEDs connected between the
// output bits and ground.  No other parts are required since current
// limiting resistors on the USBxCH board make it possible to power a T1
// LED directly off the digital output signals.
//
// For example, to see output bit 0, connect the negative LED lead to
// ground at DB25 pin 18 (or TRM25E position 10) and the positive LED lead
// to position 5 (or TRM25E position 9).  This LED should light whenever
// output bit 0 is set (1).
//
// An easy way to perform a loopback read test with this program is to
// prepare a test jig using a solder cup D-shell or a TRM25E terminal board
// and 4 pieces of wire.  Use the wire to connect each of the digital
// output bits to the corresponding digital input bit by connecting the
// following pairs of DB25 pins together: 1 + 5, 2 + 6, 3 + 7, 4 + 8
// If using a TRM25E, the euro block positions to connect are: 
// 1 + 9, 3 + 11, 5 + 13, 7 + 15
//
// With these connections, any value that is written to the output bits
// will be read back on the input bits and displayed by DigitalIo.exe.
//
// The following table shows the relationship between the digital I/O
// bits, the DB25 connector pins, and the TRM25E terminal board euro
// block positions.  For example, access output bit 2 is available by
// connecting to either pin #7 on the DB15 connector or euro block
// position 13 on the TRM25E.
//
//   INPUT bit    0   1   2   3
//   OUTPUT bit                   0   1   2   3 
//   Ground                                           G   G   G   G   G  ...  G 
//   DB pin #     1   2   3   4   5   6   7   8  ... 14  15  16  17  18  ... 25
//   TRM25E pos   1   3   5   7   9  11  13  15  ...  2   4   6   8  10  ... 24
//
//



#include <stdio.h>
#include <stdlib.h>  // for exit
#include <string.h>  // for strcmp

#include "SrHelper.h"
#include "SrUsbXch.h"



// Program copyright and screen title:

char *Copyright =

          "\n\nDigitalIo.exe : Copyright (c), Symmetric Research, 2009-2010\n\n";

char *ScreenTitle =
                   
          "USBxCH DigitalIo [/?] : (c) SR 2009 : Rev {2010/08/15}\n\n";



// Global variables

#define MAXSTR  256




// DIGITALIO RUN TIME PARAMETERS:
//
// Both the user specified command line parameters and fixed hardwired
// parameters are allocated here.  Access this structure with statments
// like: DP.nFiles.
//

struct DigitalIoParametersLayout {

         int AtodXchModel;
         int AtodGpsModel;
         int AtodUserCfgByte;
        char AtodDriverName[MAXSTR];

        } DigitalIoParameters = {


                // default command line parameter values:
                //
                //

                SRDAT_ATODMODEL_USB4CH,   // fixed AtodXchModel
                SRDAT_GPSMODEL_PCTIME,    // fixed AtodGpsModel
                USERCFG_DEFAULT_BYTE,     // fixed AtodUserCfgByte
                
                // variable AtodDriverName

                };

#define DP  DigitalIoParameters  // DP = shorthand for DigitalIo parameter struct



// Message text

char *UsageMsg[] = {

 "Writes and reads the USBxCH digital I/O                                    ",
 "                                                                           ",
 "USAGE: DigitalIo [driver] [/?]                                             ",
 "                                                                           ",
 "       driver = SrUsbXch# where # is 0 to 9, default SrUsbXch0             ",
 "           /? = show this help message                                     ",
 "                                                                           ",
 "Examples:                                                                  ",
 "                                                                           ",
 "    DigitalIo /?         < show help screen and quit                       ",
 "    DigitalIo SrUsbXch0  < test digital I/O on USBxCH 0                    ",
 "                                                                           ",
 "DigitalIo prompts for a four bit hex value.  It is written to the USBxCH   ",
 "DB25 digital I/O connector.  Output values are latched.  After output, the ",
 "digital input bits are read and displayed.  See DigitalIo.c for comments   ",
 "on making a simple loopback test jig with the TRM25E.                      ",
 "                                                                           ",
 
 NULL,
 };


char *ShowBinary[] = {
        "0000", "0001", "0010", "0011", "0100", "0101", "0110", "0111",
        "1000", "1001", "1010", "1011", "1100", "1101", "1110", "1111"
        };



// Function Prototypes

void ProcessCommandLine( int argc, char *argv[] );
void PrintRunTimeSummary( void );
void Usage( char *msg );
void PrintMsg( char *MsgArray[] );
int StrEqu( char *s1, char *s2 );
void ErrorCmd( char *msg );




// Main Digital I/O Program Code
//
//
// 
        
int main( int argc, char *argv[] ) {

        int       Error, Output, Input, Nscan;
        char      Response[15];
        DEVHANDLE UsbXchHandle;


        // Announce program version

        printf( ScreenTitle );
        
        
        // Process the command line arguments, if any.

        ProcessCommandLine( argc, argv );


        // Print out the final parameter summary before running.

        PrintRunTimeSummary( );


        // Open the USBxCH A/D driver.

        UsbXchHandle = SrUsbXchOpen( DP.AtodDriverName,
                                     DP.AtodXchModel,
                                     DP.AtodGpsModel,
                                     DP.AtodUserCfgByte,
                                     130.0,
                                     NULL,
                                    &Error
                                    );

        if ( !SrUsbXchIsOpenSuccess( Error ) ) {
                printf( "\nERROR: Could not open USBxCH device driver. (Err=%s)\n",
                          USBXCH_ERROR_MSG[Error] );
                exit( 1 );
                }

        

        // Loop until the user enters q.  Prompt for a value, output
        // it, read the digital I/O, and show the results.

        while ( 1 ) {

                // Prompt user for a value and process response.

                printf( "\nEnter hex output value or q to quit: " );
                SrFlushPrint();

                scanf( "%s", Response );
                if ( (Response[0] == 'q') || (Response[0] == 'Q') )
                        break;

                Nscan = sscanf( Response, "%X", &Output );
                if ( Nscan != 1 || Output < 0x0 || Output > 0xF ) {
                        printf( "Value must be between 0x0 and 0xF.  Please try again.\n\n" );
                        continue;
                        }

        
                // Write the digital output.

                SrUsbXchUserIoWr( UsbXchHandle, Output );



                // Read the digital input.

                SrUsbXchUserIoRd( UsbXchHandle, &Input );



                // Display the results.

                printf( "Digital output = 0x%X ( binary %s )\n", Output, 
                                                           ShowBinary[Output] );
                printf( "Digital input  = 0x%X ( binary %s )\n", Input, 
                                                           ShowBinary[Input] );

                } // end while

        

        // Close the USBxCH.

        SrUsbXchClose( UsbXchHandle );
        return( 0 );
}


//------------------------------------------------------------------------------
// ROUTINE: ProcessCommandLine
// PURPOSE: Gather the command line arguments, filling in global variables.
//------------------------------------------------------------------------------
void ProcessCommandLine( int argc, char *argv[] ) {

        int  i, DriverNum, DriverNameFound;
        char MsgBuffer[MAXSTR];


        // Set defaults

        SrStringCopy( DP.AtodDriverName, USBXCH_DEFAULT_DRIVER_NAME, MAXSTR );
        DriverNameFound = 0;

        
        // Verify number of arguments (none to 1)

        if ( argc < 1 || argc > 2 )
                Usage( "wrong number of arguments" );



        // Scan for valid arguments.

        for ( i = 1 ; i < argc ; i++ ) {


                // Check for a "/?" usage message request.

                if ( StrEqu( argv[i], "/?" ) )

                        Usage( "?" );



                // Check for a valid driver name

                else if ( SrUsbXchIsValidDriverName( argv[i], &DriverNum ) ) {
                        
                        SrStringCopy( DP.AtodDriverName, argv[i], MAXSTR );

                        if ( !DriverNameFound )
                                DriverNameFound = 1;

                        else
                               ErrorCmd( "multiple driver arguments" );
                        }


                // Unknown argument specified.

                else {
                        sprintf( MsgBuffer, "unknown argument > %s", argv[i] );

                            ErrorCmd( MsgBuffer );
                        }


                } // end for i

        return;
}

//------------------------------------------------------------------------------
// Routine: PrintRunTimeSummary
// PURPOSE: Display the final DP parameters to run with.
//------------------------------------------------------------------------------
void PrintRunTimeSummary( void ) {

        char *AtodXchModelName;
        char *AtodGpsModelName;


        // Map DP.AtodXchModel and DP.AtodGpsModel to string names ...

        AtodXchModelName = SrDatAtodModelString( DP.AtodXchModel );
        AtodGpsModelName = SrDatGpsModelString( DP.AtodGpsModel );
        
        printf( "  DigitalIo runtime summary ... \n\n");

        printf( "       DriverName:  %s\n", DP.AtodDriverName );
        printf( "         XchModel:  %s\n", AtodXchModelName );
        printf( "         GpsModel:  %s\n", AtodGpsModelName );
        printf( "\n");
}

//------------------------------------------------------------------------------
// ROUTINE: Usage
// PURPOSE: Display the usage message.
//------------------------------------------------------------------------------
void Usage( char *msg ) {

        // Select error line depending on 
        // 1) no message, 2) help requested, or 3) message given

        if ( !msg )
                printf( "\nERROR: Command line usage.\n\n" );

        else if ( strcmp( msg, "?" ) == 0 )
                ;

        else
                printf( "\nERROR: Command line usage (%s).\n\n", msg );

        PrintMsg( UsageMsg );
        exit( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: PrintMsg
// PURPOSE: Print an array of program status messages and continue on.
//------------------------------------------------------------------------------
void PrintMsg( char *MsgArray[] ) {

        int i;

        for ( i = 0 ; MsgArray[i] ; i++ )
                printf( "%s\n", MsgArray[i] );
}

//------------------------------------------------------------------------------
// ROUTINE: StrEqu
// PURPOSE: Compare for exact equality between two strings.
//------------------------------------------------------------------------------
int StrEqu( char *s1, char *s2 ) {

        while ( *s1 && *s2 && (*s1 == *s2) )

                { s1++; s2++; }

        return ( (*s1 == '\0') && (*s2 == '\0') ? 1 : 0 );

}

//------------------------------------------------------------------------------
// ROUTINE: ErrorCmd
// PURPOSE: Print a command line error message and quit.
//------------------------------------------------------------------------------
void ErrorCmd( char *msg ) {

        // Print a command line error message and quit.

        printf(

               "ERROR: Command line usage : %s \n\n ... DigitalIo /? for help\n",
               msg

               );

        exit( 1 );  // < quit at errorlevel 1.
}

