// FILE: EquipStatus.c
// COPYRIGHT: (c), Symmetric Research, 2009-2010
//
// This program reads status packets from the USBxCH and extracts the
// equipment status info.  It displays the power status and temperature.


// Include files:

#if defined( SROS_WINXP ) || defined( SROS_WIN2K )
#include <windows.h>
#elif defined( SROS_LINUX )
#include <unistd.h>      // for access, close function and STDIN_FILENO
#endif  // SROS_xxxxx

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>       // for time() and localtime() functions
#include <ctype.h>      // for isdigit(), isxdigit()

#include "SrUsbXch.h"   // for USBxCH user library functions ...
#include "SrHelper.h"   // for os independent mkdir, keyboard support ...





// Program copyright and screen title:

char *Copyright =

          "\n\nEquipStatus.exe : Copyright (c), Symmetric Research, 2009-2010\n\n";

char *ScreenTitle =

      "USBxCH Equipment Status Program: (c) SR 2009 : Rev {2010/08/15}\n\n";



#define MAXSTR 256      // dimension for message strings


// EQUIPSTATUS RUN TIME PARAMETERS:
//
// Both the user specified command line parameters and fixed hardwired
// parameters are allocated here.  Access this structure with statments
// like: EP.nFiles.
//

#define RUN_NONE         0
#define RUN_ONCE         1
#define RUN_FOREVER      2

struct EquipStatusParametersLayout {

         int RunMode;

         int AtodXchModel;
         int AtodGpsModel;
         int AtodUserCfgByte;
        char AtodDriverName[MAXSTR];

        } EquipStatusParameters = {


                // default command line parameter values:
                //
                //         RunMode = 0      -> not selected
                //                   1      -> once and then stop
                //                   2      -> forever, until 'q' aborts
                //
                //

                RUN_NONE,               // default RunMode is not selected



                // fixed driver name, GPS model, etc:
                //
                //   ( see SrUsbXch.h for MODEL define )
                //
                //

                SRDAT_ATODMODEL_USB4CH,   // fixed AtodXchModel
                SRDAT_GPSMODEL_PCTIME,    // fixed AtodGpsModel
                USERCFG_DEFAULT_BYTE,     // fixed AtodUserCfgByte
                
                // variable AtodDriverName

                };

#define EP  EquipStatusParameters  // EP = shorthand for EquipStatus parameter struct








// USAGE MESSAGE ARRAY:

char *UsageMsg[] = {

      "Read status info displaying the current power and temperature         ",
      "                                                                      ",
      "USAGE: EquipStatus [forever] [once] [driver] [/?]                     ",
      "                                                                      ",
      "     forever = run until 'q' keypress                                 ",
      "        once = run until one status displayed                         ",
      "      driver = SrUsbXch# where # is 0 to 9, default SrUsbXch0         ",
      "          /? = show help screen and quit                              ",
      "                                                                      ",
      "Examples:                                                             ",
      "                                                                      ",
      "    EquipStatus /?          < show help screen and quit               ",
      "    EquipStatus forever     < run forever, 'q' aborts                 ",
      "    EquipStatus once        < run once                                ",
      "                                                                      ",

      NULL
      };


#define MAXPACKET  50
SRUSBXCH_PACKET AtodPacketBuffer[MAXPACKET];


// Function prototypes:

   DEVHANDLE StartAcquisition( void );
        void PcTimeStamp( DEVHANDLE AtodHandle );
        void AcquireEquip( DEVHANDLE AtodHandle );
         int ProcessEquip( void );
        void StopAcquisition( DEVHANDLE AtodHandle );
        void ProcessCommandLine( int argc, char *argv[] );
        void PrintRunTimeSummary( void );
        void Usage( void );
        void PrintMsg( char *msg );
        void ErrorCmd( char *msg );
        void ErrorExec( char *msg );
         int StrEqu( char *s1, char *s2 );
         int CleanExit( int ErrLevel );



//------------------------------------------------------------------------------
// ROUTINE: main()
// PURPOSE: EquipStatus main entry point ...
//------------------------------------------------------------------------------
int main( int argc, char *argv[] ) {

        DEVHANDLE AtodHandle; int c;

        char MsgBuffer[MAXSTR];



        // Announce the program.

        printf( ScreenTitle );


        // Set keyboard behavior to allow for non blocked key checks.

        SrKeyboardMode( KEYMODE_NONBLOCK );


        // Process the command line arguments.

        ProcessCommandLine( argc, argv );  // fills in EP ...


        // Print out the final EP parameter summary before running.

        PrintRunTimeSummary( );




        // Open, initialize, and start the USBxCH.

        AtodHandle = StartAcquisition( /* with EP parameters */ );


        // Acquire equipment info:
        //

        while ( 1 ) { // infinite loop until break; ...


                // If using PCTIME, send a time stamp to the USBxCH.
                //  ... if GPS ... all time stamping is down on the USBxCH

                if ( EP.AtodGpsModel == SRDAT_GPSMODEL_PCTIME )

                        PcTimeStamp( AtodHandle );



                // Acquire equipment data from the USBxCH.

                AcquireEquip( AtodHandle );


                // Process equipment data, returning 0 to stop after a valid
                // time is found, unless continuous run mode is selected.

                if ( !ProcessEquip() )
                        break;



                // Allow a keypress to abort from continuous run ...

                if ( (c = SrGetKeyCheck()) != EOF ) {

                  // != EOF -> there is a keypress, so check it out ...

                        if ( c == 'q' || c == 'Q' ) {

                             PrintMsg( "\n ... Keypress 'q' abort ... \n");

                                break;

                                }

                        else {
                                sprintf( MsgBuffer, "To abort press %s ... \n", SR_QENTER );
                                PrintMsg( MsgBuffer );
                                }

                        }


                }



        // Stop the USBxCH and close the driver.

        StopAcquisition( AtodHandle );



        // Return from main:
        //
        // Linux *requires* main to be of type integer with return 0 for
        // success.  Windows doesn't care ...
        //
        //

        CleanExit( 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: StartAcquisition
// PURPOSE: Start the USBxCH running.
//------------------------------------------------------------------------------
#define SampleRate_s1 19.531250

DEVHANDLE StartAcquisition( void ) {

        DEVHANDLE AtodHandle;
        double    ActualSps;
        int       Error;

        char MsgBuffer[MAXSTR];

        
        // Initialize serial buffer parameters

        SerBufInit();


        // Open the USBxCH driver.

        AtodHandle = SrUsbXchOpen(

                                EP.AtodDriverName,
                                EP.AtodXchModel,
                                EP.AtodGpsModel,
                                EP.AtodUserCfgByte,
                                SampleRate_s1,

                                &ActualSps,
                                &Error

                                );

        if ( !SrUsbXchIsOpenSuccess( Error ) ) {

                sprintf( MsgBuffer,

                         "SrUsbXchOpen error: %s", USBXCH_ERROR_MSG[Error] );

                        ErrorExec( MsgBuffer);
                }



        // Start the USBxCH.

        if ( EP.RunMode == RUN_FOREVER )
                sprintf( MsgBuffer, "Acquisition starting.  Press %s to abort ... \n\n", SR_QENTER );
        else
                sprintf( MsgBuffer, "Acquisition starting ... \n\n" );

        PrintMsg( MsgBuffer );


        if ( !SrUsbXchStart( AtodHandle, &Error ) ) {
                
                SrUsbXchClose( AtodHandle );
                AtodHandle = BAD_DEVHANDLE;

                sprintf( MsgBuffer,

                         "SrUsbXchStart error: %s", USBXCH_ERROR_MSG[Error] );

                        ErrorExec( MsgBuffer);
                }



        // Return the driver handle.

        return( AtodHandle );
}

//------------------------------------------------------------------------------
// ROUTINE: PcTimeStamp
// PURPOSE: Send a PC Time stamp and PPS down to the USBxCH.
//------------------------------------------------------------------------------
void PcTimeStamp( DEVHANDLE AtodHandle ) {

        int Error; double CurrentTime; char MsgBuffer[MAXSTR];

        static double LastTime = (double)0.0;


        // Get current PC time

        if ( !SrGetPcTime( &CurrentTime,
                          NULL, NULL, NULL, NULL, NULL, NULL, NULL ) )

                ErrorExec( "SrGetPcTime error.  Could not get current PC time\n" );


        // If more than a second has passed,
        // send a PC time stamp down to the USBxCH.

        if ( CurrentTime > (LastTime + 1.0) ) {

                LastTime = CurrentTime; // update last time

                SrUsbXchSendTimeStamp( AtodHandle, &Error );

                if ( Error != USBXCH_ERROR_NONE ) {

                    sprintf( MsgBuffer, "SrUsbXchSendTimeStamp error: %d", Error );

                        ErrorExec( MsgBuffer );

                        }
                }

}

//------------------------------------------------------------------------------
// ROUTINE: AcquireEquip
// PURPOSE: Read equipment data from the USBxCH.
//------------------------------------------------------------------------------
void AcquireEquip( DEVHANDLE AtodHandle ) {

        SRUSBXCH_PACKET *pBuffer;
          unsigned long nRequested;
          unsigned long nReturned;
                    int Error;
                   char MsgBuffer[MAXSTR];



        // Loop on GetData until nRequested packets have been read.

           pBuffer = AtodPacketBuffer;
        nRequested = 1;

        while ( nRequested > 0 ) {


                SrUsbXchGetDataAsPackets(

                        AtodHandle,
                        pBuffer,
                        nRequested,  // < # packets requested

                        &nReturned,  // < # packets returned
                        &Error

                        );

                if ( Error != USBXCH_ERROR_NONE ) {

                        sprintf( MsgBuffer,

                                 "SrUsbXchGetDataAsPackets error: %d", Error );

                            ErrorExec( MsgBuffer );
                        }


                   pBuffer += nReturned;
                nRequested -= nReturned;

                }
}


//------------------------------------------------------------------------------
// ROUTINE: ProcessEquip
// PURPOSE: Extract power status and temperature from the equipment data and
//          display it.
//------------------------------------------------------------------------------
int ProcessEquip( void ) {

           int Error, errNmea, nChar, nSecond, EqFlag;
          char NmeaMsgString[SRDAT_NMEA_MAX_SIZE], MsgBuffer[MAXSTR];
          char *ResultString;
        double TempC, TempF;
        
        SRDAT_USBXCH_EQUIPMENT EquipData;



        // Process STATUS packets and ignore all others

        if ( SrUsbXchPacketType( &AtodPacketBuffer[0] ) !=
             SRDAT_USBPACKET_TYPE_STATUS )
                
                return( 1 ); // Continue



        // Put packet characters into working serial buffer

        if ( !SrDatUsbSerialFillBuffer( &AtodPacketBuffer[0], 0, &Error ) ) {
                sprintf( MsgBuffer,
                         "SrDatUsbSerialFillBuffer error: %d", Error );
                ErrorExec( MsgBuffer );
                }



        // Parse serial buffer characters into NMEA strings and equipment structure

        while ( 1 ) {

                errNmea = SrDatSerialReadNmea(
                                        SRDAT_NMEA_MAX_SIZE,
                                        NmeaMsgString,
                                        NULL,                   // don't check MSGID
                                        &nChar,                 // ignoring this value
                                        &nSecond,
                                        &EqFlag,
                                        &EquipData
                                       );

                nChar = SerBufNumChar();

                // Display the equipment data, if any

                if ( EqFlag == 1 ) {

                        // Prepare power info

                        if ( SrUsbXchIsPowerBitGood( EquipData.VoltageGood,
                                                     USBXCH_POWER_HISTORY ) )
                                ResultString = "GOOD";
                        else
                                ResultString = " BAD";


                        // Prepare temperature info
                        
                        SrDatUsbTemperatureCompute( EquipData.Temperature,
                                                    &TempC, &TempF );

                        
                        // Display it

                        sprintf( MsgBuffer,
                                 "Power is %s (0x%02X), Temperature is %5.1lf C, %3.0lf F (0x%X)\n",
                                 ResultString, EquipData.VoltageGood,
                                 TempC, TempF, EquipData.Temperature );
                        PrintMsg( MsgBuffer );


                        if ( EP.RunMode == RUN_ONCE )
                                return( 0 );
                        }


                // Got error while parsing, so exit loop

                if ( errNmea != SRDAT_NMEA_ERR_NONE )
                        break;


                // No more serial characters to parse, so exit loop

                else if ( nChar == 0 )
                        break;

                } // end while 1

        
        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: StopAcquisition
// PURPOSE: Stop the USBxCH and close the driver.
//------------------------------------------------------------------------------
void StopAcquisition( DEVHANDLE AtodHandle ) {

         SrUsbXchStop( AtodHandle, NULL );
        SrUsbXchClose( AtodHandle );
}

//------------------------------------------------------------------------------
// ROUTINE: ProcessCommandLine
// PURPOSE: Gather the command line arguments, filling in EP.
//------------------------------------------------------------------------------
void ProcessCommandLine( int argc, char *argv[] ) {

        int i, DriverNum, DriverNameFound;
        char MsgBuffer[MAXSTR];



        // Set defaults

        SrStringCopy( EP.AtodDriverName, USBXCH_DEFAULT_DRIVER_NAME, MAXSTR );
        DriverNameFound = 0;


        // EquipStatus with no arguments at all gives the usage message.

        if ( argc == 1 )

                Usage( );


        // Check for too many arguments.

        if ( argc > 3 )

                ErrorCmd( "too many command line arguments" );



        // Scan for valid arguments.

        for ( i = 1 ; i < argc ; i++ ) {



                // Check for a "/?" usage message request.

                if ( StrEqu( argv[i], "/?") )

                        Usage( );



                // Check for a valid driver name

                else if ( SrUsbXchIsValidDriverName( argv[i], &DriverNum ) ) {
                        
                        SrStringCopy( EP.AtodDriverName, argv[i], MAXSTR );

                        if ( !DriverNameFound )
                                DriverNameFound = 1;

                        else
                               ErrorCmd( "multiple driver arguments" );
                        }


                // Check for a run mode of "forever" request.

                else if ( StrEqu( argv[i], "forever" ) ) {

                        if ( EP.RunMode == RUN_NONE )
                                EP.RunMode = RUN_FOREVER;

                        else

                               ErrorCmd( "multiple run mode arguments.  Use forever OR once." );
                        }



                // Check for a run mode of "once" request.

                else if ( StrEqu( argv[i], "once" ) ) {

                        if ( EP.RunMode == RUN_NONE )
                                EP.RunMode = RUN_ONCE;

                        else

                               ErrorCmd( "multiple run mode arguments.  Use forever OR once." );
                        }



                // Unknown argument specified.

                else {
                        sprintf( MsgBuffer, "unknown argument > %s", argv[i] );

                            ErrorCmd( MsgBuffer );
                        }


                }

        if ( EP.RunMode == RUN_NONE )
                ErrorCmd( "run mode argument required.  Use forever OR once." );

}

//------------------------------------------------------------------------------
// Routine: PrintRunTimeSummary
// PURPOSE: Display the final EP parameters to run with.
//------------------------------------------------------------------------------
void PrintRunTimeSummary( void ) {

        char *AtodXchModelName;
        char *AtodGpsModelName;


        // Map EP.AtodXchModel and EP.AtodGpsModel to string names ...

        AtodXchModelName = SrDatAtodModelString( EP.AtodXchModel );
        AtodGpsModelName = SrDatGpsModelString( EP.AtodGpsModel );


        // Sampling rate and driver parameters.

        printf( "  EquipStatus runtime summary ... \n\n");

        printf( "             DriverName:  %s\n",      EP.AtodDriverName  );
        printf( "               XchModel:  %s\n",      AtodXchModelName   );
        printf( "               GpsModel:  %s\n",      AtodGpsModelName   );
        printf( "            UserCfgByte:  0x%02X\n",  EP.AtodUserCfgByte );

    if ( EP.RunMode == RUN_FOREVER )
        printf( "                    Run:  forever\n"                     );
    
    else // ( EP.RunMode == RUN_ONCE )
        printf( "                    Run:  once\n"                        );


        printf( "\n" );

}

//------------------------------------------------------------------------------
// ROUTINE: Usage
// PURPOSE: Display the usage message.
//------------------------------------------------------------------------------
void Usage( void ) {

        int i;


        // Print the usage message.

        for ( i = 0 ; UsageMsg[i] ; i++ )

                printf( "%s\n", UsageMsg[i] );


        CleanExit( 0 );  // < quit at errorlevel 0.

}

//------------------------------------------------------------------------------
// ROUTINE: PrintMsg
// PURPOSE: Print a program status message and continue on.
//------------------------------------------------------------------------------
void PrintMsg( char *msg ) {

        // Print a program status message.

        printf( msg );

}

//------------------------------------------------------------------------------
// ROUTINE: ErrorCmd
// PURPOSE: Print a command line error message and quit.
//------------------------------------------------------------------------------
void ErrorCmd( char *msg ) {

        // Print a command line error message and quit.

        printf(

               "ERROR: Command line usage : %s \n\n ... EquipStatus /? for help\n",
               msg

               );

        CleanExit( 1 );  // < quit at errorlevel 1.

}

//------------------------------------------------------------------------------
// ROUTINE: ErrorExec
// PURPOSE: An execution error occured.
//------------------------------------------------------------------------------
void ErrorExec( char *msg ) {

        // A fatal execution error occured.

        printf( "ERROR: %s \n\n", msg );

        CleanExit( 2 );  // < quit at errorlevel 2.

}

//------------------------------------------------------------------------------
// ROUTINE: StrEqu
// PURPOSE: Compare for exact equality between two strings.
//------------------------------------------------------------------------------
int StrEqu( char *s1, char *s2 ) {

        while ( *s1 && *s2 && (*s1 == *s2) )

                { s1++; s2++; }

        return ( (*s1 == '\0') && (*s2 == '\0') ? 1 : 0 );

}

//------------------------------------------------------------------------------
// ROUTINE: CleanExit
// PURPOSE: Leave the program, while ensuring the keyboard mode is restored.
//------------------------------------------------------------------------------
int CleanExit( int ErrLevel ) {
        
        SrKeyboardMode( KEYMODE_BLOCK );  // Restore blocking keyboard behavior.

        exit( ErrLevel );
}
