// FILE: NmeaTime.c
// COPYRIGHT: (c), Symmetric Research, 2009-2010
//
// This program reads GPS NMEA data from the USBxCH and extracts the
// current NMEA time.  The time is accurate to roughly one second and
// can be used to set the PC time.


// Include files:

#if defined( SROS_WINXP ) || defined( SROS_WIN2K )
#include <windows.h>
#elif defined( SROS_LINUX )
#include <unistd.h>      // for access, close function and STDIN_FILENO
#endif  // SROS_xxxxx

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>       // for time() and localtime() functions
#include <ctype.h>      // for isdigit(), isxdigit()

#include "SrUsbXch.h"   // for USBxCH user library functions ...
#include "SrHelper.h"   // for os independent mkdir, keyboard support ...





// Program copyright and screen title:

char *Copyright =

          "\n\nNmeaTime.exe : Copyright (c), Symmetric Research, 2010\n\n";

char *ScreenTitle =

      "USBxCH NMEA Time Program: (c) SR 2010 : Rev {2010/08/15}\n\n";


#define MAXSTR 256               // dimension for message strings


// NMEATIME RUN TIME PARAMETERS:
//
// Both the user specified command line parameters and fixed hardwired
// parameters are allocated here.  Access this structure with statments
// like: NP.nFiles.
//

struct NmeaTimeParametersLayout {

        struct TimeTableRecord *pTimeTable;
        
         int RunForever;
         int SetPcTime;
         int UserCfgByte;
         int AtodGpsModel;
         int AtodXchModel;

        char AtodDriverName[MAXSTR];

        } NmeaTimeParameters = {


                // default command line parameter values:
                //
                //      RunForever = 0      -> not forever, but stop after valid time
                //       SetPcTime = 0      -> do not set PC time from GPS time
                //     UserCfgByte = 0x00   -> standard GPS config
                //              gn = g2     -> Garmin
                //
                //

                NULL,                   // optional "gn"
                0,                      // default RunForever
                0,                      // default SetPcTime
                USERCFG_DEFAULT_BYTE,   // default UserCfgByte
                SRDAT_GPSMODEL_PCTIME,  // default "gn" AtodGpsModel

                SRDAT_ATODMODEL_USB4CH, // fixed AtodXchModel
                                        //   ( see SrUsbXch.h for MODEL define )
                
                // variable AtodDriverName
             
                };

#define NP  NmeaTimeParameters  // NP = shorthand for NmeaTime parameter struct



// TIME SOURCE TABLE:
//
// These are the possible time sources for Blast: PC Time, GPS Time ...
//
//

struct TimeTableRecord {

        char  *GnString;        // < allowed command line "gn" string
        int   AtodGpsModel;     // < GPS model, NONE, PCTIME, GARMIN etc
        char  *UsageDesc;       // < usage message description

        } TimeTable[] = {

          { "g1",  SRDAT_GPSMODEL_PCTIME  , "PCTIME   ( NMEA = ZDA )      << default" },
          { "g2",  SRDAT_GPSMODEL_GARMIN  , "GARMIN   ( NMEA = RMC, GGA )           " },
          { "g3",  SRDAT_GPSMODEL_TRIMBLE , "TRIMBLE  ( NMEA = ZDA, GGA )           " },

          // end of table ...

          { NULL, 0, NULL },

          };


#define TR  struct TimeTableRecord  // TR = shorthand for time source record






// USAGE MESSAGE ARRAY:

char *UsageMsg[] = {

      "Read and display GPS NMEA strings and optionally set PC time          ",
      "                                                                      ",
      "USAGE: NmeaTime [gn] [forever] [set] [0xUC] [driver] [/?]             ",
      "                                                                      ",
      "          gn = time source                                            ",
      "     forever = run until 'q' keypress                                 ",
      "         set = set PC time from first valid NMEA time                 ",
      "        0xUC = User Config Byte as 0x.. hex number                    ",
      "      driver = SrUsbXch# where # is 0 to 9, default SrUsbXch0         ",
      "          /? = show help screen and quit                              ",
      "                                                                      ",
      "                                                                      ",
      "  Time source table: gn selects the time source ...                   ",
      "                                                                      ",
//    "    GG   ->  DESC                                                     ",
      NULL,
      "    %s   ->  %s        ",

      "                                                                      ",
      "                                                                      ",
      "Examples:                                                             ",
      "                                                                      ",
      "    NmeaTime /?          < show help screen and quit                  ",
      "    NmeaTime g2          < run with Garmin receiver until first valid ",
      "    NmeaTime g2 set      < set PC time from NMEA time and exit        ",
      "    NmeaTime forever     < run forever with PcTime, 'q' aborts        ",
      "    NmeaTime g3 forever  < run forever with Trimble receiver          ",
      "                                                                      ",

      NULL
      };



#define MAXPACKET  50
SRUSBXCH_PACKET AtodPacketBuffer[MAXPACKET];


// Function prototypes:

   DEVHANDLE StartAcquisition( void );
        void PcTimeStamp( DEVHANDLE AtodHandle );
        void AcquireNmea( DEVHANDLE AtodHandle );
         int ProcessNmea( void );
         int ExtractNmeaInfo( char *NmeaGroup, int NmeaCount, NI *NmeaInfo );
        void SetPcTimeFromNmeaTime( NI *NmeaInfo );
        void StopAcquisition( DEVHANDLE AtodHandle );
          TR *TimeTableLookUp( char *s );
        void ProcessCommandLine( int argc, char *argv[] );
        void PrintRunTimeSummary( void );
        void Usage( void );
        void PrintMsg( char *msg );
        void ErrorCmd( char *msg );
        void ErrorExec( char *msg );
         int StrEqu( char *s1, char *s2 );
         int StrIsHex( char *s );
         int CleanExit( int ErrLevel );



//------------------------------------------------------------------------------
// ROUTINE: main()
// PURPOSE: NmeaTime main entry point ...
//------------------------------------------------------------------------------
int main( int argc, char *argv[] ) {

        DEVHANDLE AtodHandle; int c;

        char MsgBuffer[MAXSTR];



        // Announce the program.

        printf( ScreenTitle );


        // Set keyboard behavior to allow for non blocked key checks.

        SrKeyboardMode( KEYMODE_NONBLOCK );


        // Process the command line arguments.

        ProcessCommandLine( argc, argv );  // fills in NP ...


        // Print out the final NP parameter summary before running.

        PrintRunTimeSummary( );




        // Open, initialize, and start the USBxCH.

        AtodHandle = StartAcquisition( /* with NP parameters */ );


        // Acquire NMEA info:
        //


        while ( 1 ) { // infinite loop until break; ...


                // If using PCTIME, send a time stamp to the USBxCH.
                //  ... if GPS ... all time stamping is down on the USBxCH

                if ( NP.AtodGpsModel == SRDAT_GPSMODEL_PCTIME )

                        PcTimeStamp( AtodHandle );



                // Acquire NMEA data from the USBxCH.

                AcquireNmea( AtodHandle );


                // Process NMEA data, returning 0 to stop after a valid
                // time is found, unless continuous run mode is selected.

                if ( !ProcessNmea() )
                        break;



                // Allow a keypress to abort from continuous run ...

                if ( (c = SrGetKeyCheck()) != EOF ) {

                  // != EOF -> there is a keypress, so check it out ...

                        if ( c == 'q' || c == 'Q' ) {

                             PrintMsg( "\n ... Keypress 'q' abort ... \n");

                                break;

                                }

                        else {
                                sprintf( MsgBuffer, "To abort press %s ... \n", SR_QENTER );
                                PrintMsg( MsgBuffer );
                                }

                        }


                }



        // Stop the USBxCH and close the driver.

        StopAcquisition( AtodHandle );



        // Return from main:
        //
        // Linux *requires* main to be of type integer with return 0 for
        // success.  Windows doesn't care ...
        //
        //

        CleanExit( 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: StartAcquisition
// PURPOSE: Start the USBxCH running.
//------------------------------------------------------------------------------
#define SampleRate_s1 19.531250

DEVHANDLE StartAcquisition( void ) {

        DEVHANDLE AtodHandle;
        double    ActualSps;
        int       Error;

        char MsgBuffer[MAXSTR];

        
        // Initialize serial buffer parameters

        SerBufInit();



        // Open the USBxCH driver.

        AtodHandle = SrUsbXchOpen(

                                NP.AtodDriverName,
                                NP.AtodXchModel,
                                NP.AtodGpsModel,
                                NP.UserCfgByte,
                                SampleRate_s1,

                                &ActualSps,
                                &Error

                                );

        if ( !SrUsbXchIsOpenSuccess( Error ) ) {

                sprintf( MsgBuffer,

                         "SrUsbXchOpen error: %s", USBXCH_ERROR_MSG[Error] );

                        ErrorExec( MsgBuffer);
                }



        // Start the USBxCH.

        if ( NP.RunForever )
                sprintf( MsgBuffer, "Acquisition starting.  Press %s to abort ... \n\n", SR_QENTER );
        else
                sprintf( MsgBuffer, "Acquisition starting ... \n\n" );

        PrintMsg( MsgBuffer );



        if ( !SrUsbXchStart( AtodHandle, &Error ) ) {
                
                SrUsbXchClose( AtodHandle );
                AtodHandle = BAD_DEVHANDLE;

                sprintf( MsgBuffer,

                         "SrUsbXchStart error: %s", USBXCH_ERROR_MSG[Error] );

                        ErrorExec( MsgBuffer);
                }



        // Return the driver handle.

        return( AtodHandle );


}

//------------------------------------------------------------------------------
// ROUTINE: PcTimeStamp
// PURPOSE: Send a PC Time stamp and PPS down to the USBxCH.
//------------------------------------------------------------------------------
void PcTimeStamp( DEVHANDLE AtodHandle ) {

        int Error; double CurrentTime; char MsgBuffer[MAXSTR];

        static double LastTime = (double)0.0;


        // Get current PC time

        if ( !SrGetPcTime( &CurrentTime,
                          NULL, NULL, NULL, NULL, NULL, NULL, NULL ) )

                ErrorExec( "SrGetPcTime error.  Could not get current PC time\n" );


        // If more than a second has passed,
        // send a PC time stamp down to the USBxCH.

        if ( CurrentTime > (LastTime + 1.0) ) {

                LastTime = CurrentTime; // update last time

                SrUsbXchSendTimeStamp( AtodHandle, &Error );

                if ( Error != USBXCH_ERROR_NONE ) {

                    sprintf( MsgBuffer, "SrUsbXchSendTimeStamp error: %d", Error );

                        ErrorExec( MsgBuffer );

                        }
                }

}

//------------------------------------------------------------------------------
// ROUTINE: AcquireNmea
// PURPOSE: Read NMEA data from the USBxCH.
//------------------------------------------------------------------------------
void AcquireNmea( DEVHANDLE AtodHandle ) {

        SRUSBXCH_PACKET *pBuffer;
          unsigned long nRequested;
          unsigned long nReturned;
                    int Error;
                   char MsgBuffer[MAXSTR];



        // Loop on GetData until nRequested packets have been read.

           pBuffer = AtodPacketBuffer;
        nRequested = 1;

        while ( nRequested > 0 ) {


                SrUsbXchGetDataAsPackets(

                        AtodHandle,
                        pBuffer,
                        nRequested,  // < # packets requested

                        &nReturned,  // < # packets returned
                        &Error

                        );

                if ( Error != USBXCH_ERROR_NONE ) {

                        sprintf( MsgBuffer,

                                 "SrUsbXchGetDataAsPackets error: %d", Error );

                            ErrorExec( MsgBuffer );
                        }


                   pBuffer += nReturned;
                nRequested -= nReturned;

                }
}


//------------------------------------------------------------------------------
// ROUTINE: ProcessNmea
// PURPOSE: Determine NmeaTime and set PC time if requested.
//------------------------------------------------------------------------------
int ProcessNmea( void ) {

         int Error, errNmea, iNmea, nChar, nSecond, EqFlag;
        char NmeaMsgString[SRDAT_NMEA_MAX_SIZE], MsgBuffer[MAXSTR];
          NI NmeaInfo;

         SRDAT_USBXCH_EQUIPMENT EquipData;

         static int nNmea = 0;
         static int nSecondLast = -1;
         static int ValidTime = 0;
         
         static char NmeaGroup[SRDAT_NMEA_MAX_BUFF];



        // Process STATUS packets and ignore all others

        if ( SrUsbXchPacketType( &AtodPacketBuffer[0] ) !=
             SRDAT_USBPACKET_TYPE_STATUS )
                
                return( 1 ); // Continue



        // Put packet characters into working serial buffer

        if ( !SrDatUsbSerialFillBuffer( &AtodPacketBuffer[0], 0, &Error ) ) {
                sprintf( MsgBuffer,
                         "SrDatUsbSerialFillBuffer error: %d", Error );
                ErrorExec( MsgBuffer );
                }



        // Parse serial buffer characters into NMEA strings

        while ( 1 ) {

                errNmea = SrDatSerialReadNmea(
                                        SRDAT_NMEA_MAX_SIZE,
                                        NmeaMsgString,
                                        NULL,                   // don't check MSGID
                                        &nChar,                 // ignoring this value
                                        &nSecond,
                                        &EqFlag,
                                        &EquipData
                                       );

                nChar = SerBufNumChar();

                // We got a good NMEA string

                if ( errNmea == SRDAT_NMEA_ERR_NONE ) {

                        // It's the first string in a group, print a blank line to
                        // separate the next NMEA group.  Then extract NMEA info and,
                        // if that goes well, set the PC time and maybe exit immediately.

                        if ( nSecond != -1 ) {

                                printf( "\n" );
                                
                                if ( (nSecondLast != nSecond )  &&
                                     (nSecondLast != -1)  ) {

                                        if ( ExtractNmeaInfo( NmeaGroup, nNmea, &NmeaInfo ) ) {
                                                
                                                ValidTime = 1;

                                                SetPcTimeFromNmeaTime( &NmeaInfo );

                                                if ( !NP.RunForever )
                                                        return( 0 ); // we're done
                                                }

                                        }

                                nNmea       = 1;
                                nSecondLast = nSecond;
                                }

                        // It's the trailing string in a group
                        
                        else {
                                if (nSecondLast == -1)
                                        printf( "\n" );
                                nNmea++;
                                }


                        // Save NmeaMsgString into the NmeaGroup for later processing

                        iNmea = (nNmea-1) * SRDAT_NMEA_MAX_SIZE;
                        if ( nNmea <= SRDAT_NMEA_MAX_TYPE )
                                strcpy( &NmeaGroup[iNmea], NmeaMsgString );
                        else {
                                sprintf( MsgBuffer,
                                         "Too many NMEA messages per second, expecting 2, got at least %d", SRDAT_NMEA_MAX_TYPE );
                                ErrorExec( MsgBuffer);
                                }


                        // Display the string

                        printf( "%s", NmeaMsgString );
                        }


                // No more serial characters to parse, so exit loop

                else if ( nChar == 0 )
                        break;

                // Got error while parsing, so exit loop

                else // ( errNmea != SRDAT_NMEA_ERR_NONE  &&  nChar != 0 )
                        break;


                } // end while 1


        
        // Decide whether or not to continue

        if ( NP.RunForever )    // if running forever ...

                return( 1 );

        else if ( !ValidTime )  // if still waiting for valid time ...

                return( 1 );

        else {                  // else done

                PrintMsg( "Stopping after valid time found.\n");
                return( 0 );
                }

}

//------------------------------------------------------------------------------
// ROUTINE: ExtractNmeaInfo
// PURPOSE: 
//------------------------------------------------------------------------------
int ExtractNmeaInfo( char *NmeaGroup, int NmeaCount, NI *NmeaInfo ) {

        int Ok;

        if ( !NmeaGroup || NmeaCount < 0 || !NmeaInfo )
                return( 0 );


        // Get the best NMEA info possible from the provided group of NMEA strings

        Ok = SrDatRequestNmeaInfo2( NmeaGroup, NmeaCount,
                                    &NmeaInfo->Year,
                                    &NmeaInfo->Month,
                                    &NmeaInfo->Day,
                                    &NmeaInfo->YmdSource,
                                    &NmeaInfo->YmdIsValid,
                                    &NmeaInfo->Hour,
                                    &NmeaInfo->Minute,
                                    &NmeaInfo->Second,
                                    &NmeaInfo->MicroSecond,
                                    &NmeaInfo->HmsSource,
                                    &NmeaInfo->HmsIsValid,
                                    &NmeaInfo->Latitude,
                                    &NmeaInfo->Longitude,
                                    &NmeaInfo->LocSource,
                                    &NmeaInfo->LocIsValid,
                                    &NmeaInfo->Altitude,
                                    &NmeaInfo->PosSource,
                                    &NmeaInfo->PosIsValid,
                                    &NmeaInfo->Nsat,
                                    &NmeaInfo->SatSource,
                                    &NmeaInfo->SatIsValid
                                  );

        // If all went well, combine the YMD HMS values into seconds since 1970
        
        if ( Ok ) {
                Ok = SrDatSecTimeCombine( NmeaInfo->Year, NmeaInfo->Month, NmeaInfo->Day,
                                          NmeaInfo->Hour, NmeaInfo->Minute, NmeaInfo->Second, 
                                          NmeaInfo->MicroSecond, &NmeaInfo->SecSince1970 );

                NmeaInfo->SecIsValid = Ok;
                
                if ( Ok && NmeaInfo->YmdIsValid && NmeaInfo->HmsIsValid )

                        return( 1 );

                // else printf( "Time/date is not valid and not printed to file\n" );

                }

        // else printf("SrDatRequestNmeaInfo2 failed\n" );

        return( 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: SetPcTimeFromNmeaTime
// PURPOSE: Set the PC time using the supplied NMEA time.
//------------------------------------------------------------------------------
void SetPcTimeFromNmeaTime( NI *NmeaInfo ) {

         int Ok;
        char MsgBuffer[MAXSTR];

        // Error check

        if ( !NmeaInfo )
                return;

        
        // Check if PC time should be set.

        if ( NP.SetPcTime ) {

                NP.SetPcTime = 0;       // Only set time once

                // Increment extracted time by one second to account for
                // pipeline delay

                NmeaInfo->SecSince1970 += (double)1.0;

                // Set the PC time
                
                Ok = SrSetPcTime( NmeaInfo->SecSince1970,    // SecSince 1970
                                  0, 0, 0, 0, 0, 0, 0L,     // YMD HMS .us
                                  0                         // From Sec not YMD
                                  );

                // Update YMD HMS to match incremented seconds since 1970 time

                SrSecTimeSplit( NmeaInfo->SecSince1970,
                                &NmeaInfo->Year, &NmeaInfo->Month, &NmeaInfo->Day,
                                &NmeaInfo->Hour, &NmeaInfo->Minute, &NmeaInfo->Second,
                                &NmeaInfo->MicroSecond );

                // Show the time to the user

                sprintf( MsgBuffer,
                         "Set PC time to %4d/%02d/%02d %02d:%02d:%02d\n",
                         NmeaInfo->Year, NmeaInfo->Month, NmeaInfo->Day,
                         NmeaInfo->Hour, NmeaInfo->Minute, NmeaInfo->Second );

                PrintMsg( MsgBuffer );
                }
}

//------------------------------------------------------------------------------
// ROUTINE: StopAcquisition
// PURPOSE: Stop the USBxCH and close the driver.
//------------------------------------------------------------------------------
void StopAcquisition( DEVHANDLE AtodHandle ) {

         SrUsbXchStop( AtodHandle, NULL );
        SrUsbXchClose( AtodHandle );
}


//------------------------------------------------------------------------------
// ROUTINE: TimeTableLookUp
// PURPOSE: Compare s against the time source table "gn" names.
//------------------------------------------------------------------------------
TR *TimeTableLookUp( char *s ) {

        TR *t;


        // Scan the time source table for a match with s.

        for ( t = TimeTable ; t->GnString ; t++ ) {

                if ( StrEqu( s, t->GnString ) )

                        return( t );  // > success, got a match ...
                }


        return( NULL );  // > failure, hit end of table without a match ...


}

//------------------------------------------------------------------------------
// ROUTINE: ProcessCommandLine
// PURPOSE: Gather the command line arguments, filling in NP.
//------------------------------------------------------------------------------
void ProcessCommandLine( int argc, char *argv[] ) {

        int i, UserCfg, DriverNum, DriverNameFound;
        TR *t;

        char MsgBuffer[MAXSTR];


        // Set defaults

        SrStringCopy( NP.AtodDriverName, USBXCH_DEFAULT_DRIVER_NAME, MAXSTR );
        DriverNameFound = 0;


        // NmeaTime with no arguments at all gives the usage message.

        if ( argc == 1 )

                Usage( );


        // Check for too many arguments.

        if ( argc > 5 )

                ErrorCmd( "too many command line arguments" );



        // Scan for valid arguments.

        for ( i = 1 ; i < argc ; i++ ) {



                // Check for a "/?" usage message request.

                if ( StrEqu( argv[i], "/?") )

                        Usage( );



                // Check for a valid driver name

                else if ( SrUsbXchIsValidDriverName( argv[i], &DriverNum ) ) {
                        
                        SrStringCopy( NP.AtodDriverName, argv[i], MAXSTR );

                        if ( !DriverNameFound )
                                DriverNameFound = 1;

                        else
                               ErrorCmd( "multiple driver arguments" );
                        }


                // Check for a "run" continuous request.

                else if ( StrEqu( argv[i], "forever" ) ) {

                        if ( !NP.RunForever )
                                NP.RunForever = 1;

                        else

                               ErrorCmd( "multiple \"forever\" arguments" );
                        }



                // Check for a "set" PC time request.

                else if ( StrEqu( argv[i], "set" ) ) {

                        if ( !NP.SetPcTime )
                                NP.SetPcTime = 1;

                        else

                               ErrorCmd( "multiple \"set\" arguments" );
                        }



                // Check for a UserConfig byte.

                else if ( StrIsHex( argv[i] ) ) {

                        sscanf( argv[i], "0x%lX", &UserCfg );

                        if ( NP.UserCfgByte == 0x00 ) {

                                if ( UserCfg <= 0x0FF )
                                        NP.UserCfgByte = UserCfg;
                                else

                  ErrorCmd( "invalid User Config Byte, must be 0x00 to 0xFF" );

                                }

                        else
                                ErrorCmd( "multiple 0x.. User Config Bytes" );

                        }




                // Check for a "gn" time source.

                else if ( t = TimeTableLookUp( argv[i] ) ) {

                        if ( NP.pTimeTable == NULL )
                                NP.pTimeTable = t;

                        else
                                ErrorCmd( "multiple \"gn\" time sources" );

                        }



                // Unknown argument specified.

                else {
                        sprintf( MsgBuffer, "unknown argument > %s", argv[i] );

                            ErrorCmd( MsgBuffer );
                        }


                }



        // Set the "gn" time source if they specified one.

        if ( NP.pTimeTable )

                NP.AtodGpsModel = NP.pTimeTable->AtodGpsModel;

                   // else AtodGpsModel already = initializer default above ...


        // Don't allow both forever and set options at the same time

        if ( NP.RunForever && NP.SetPcTime )
                ErrorCmd( "Choose forever or set, but not both" );


        // Don't allow setting PcTime when using that as the time source

        if ( NP.AtodGpsModel == SRDAT_GPSMODEL_PCTIME  &&  NP.SetPcTime )
                ErrorCmd( "Set PC time not allowed when PC is time source (g1)" );

}

//------------------------------------------------------------------------------
// Routine: PrintRunTimeSummary
// PURPOSE: Display the final NP parameters to run with.
//------------------------------------------------------------------------------
void PrintRunTimeSummary( void ) {

        char *AtodXchModelName;
        char *AtodGpsModelName;


        // Map NP.AtodXchModel and NP.AtodGpsModel to string names ...

        AtodXchModelName = SrDatAtodModelString( NP.AtodXchModel );
        AtodGpsModelName = SrDatGpsModelString( NP.AtodGpsModel );


        // Sampling rate and driver parameters.

        printf( "  NmeaTime runtime summary ... \n\n");

        printf( "             DriverName:  %s\n",      NP.AtodDriverName );
        printf( "               XchModel:  %s\n",      AtodXchModelName  );
        printf( "               GpsModel:  %s\n",      AtodGpsModelName  );
        printf( "            UserCfgByte:  0x%02X\n",  NP.UserCfgByte    );

    if ( NP.RunForever )
        printf( "                    Run:  forever\n"                    );
    else
        printf( "                    Run:  until first valid time\n"     );

    if ( NP.SetPcTime )
        printf( "            Set PC Time:  yes\n"                        );
    else
        printf( "            Set PC Time:  no\n"                         );
    
        printf( "\n" );

}

//------------------------------------------------------------------------------
// ROUTINE: Usage
// PURPOSE: Display the usage message.
//------------------------------------------------------------------------------
void Usage( void ) {

        int i; TR *t;



        // Print the top section of the usage message.

        for ( i = 0 ; UsageMsg[i] ; i++ )

                printf( "%s\n", UsageMsg[i] );



        // Print the allowed time source table selections.

        for ( ++i , t = TimeTable ; t->GnString ; t++ ) {
        
                printf( UsageMsg[i],  // < fixed format string

                          t->GnString,
                          t->UsageDesc

                          );

                printf( "\n");

                }


        // Print the bottom section of the usage message.

        for ( ++i ; UsageMsg[i] ; i++ )

                printf( "%s\n", UsageMsg[i] );



        CleanExit( 0 );  // < quit at errorlevel 0.


}

//------------------------------------------------------------------------------
// ROUTINE: PrintMsg
// PURPOSE: Print a program status message and continue on.
//------------------------------------------------------------------------------
void PrintMsg( char *msg ) {

        // Print a program status message.

        printf( msg );

}

//------------------------------------------------------------------------------
// ROUTINE: ErrorCmd
// PURPOSE: Print a command line error message and quit.
//------------------------------------------------------------------------------
void ErrorCmd( char *msg ) {

        // Print a command line error message and quit.

        printf(

               "ERROR: Command line usage : %s \n\n ... NmeaTime /? for help\n",
               msg

               );

        CleanExit( 1 );  // < quit at errorlevel 1.

}

//------------------------------------------------------------------------------
// ROUTINE: ErrorExec
// PURPOSE: An execution error occured.
//------------------------------------------------------------------------------
void ErrorExec( char *msg ) {

        // A fatal execution error occured.

        printf( "ERROR: %s \n\n", msg );

        CleanExit( 2 );  // < quit at errorlevel 2.

}

//------------------------------------------------------------------------------
// ROUTINE: StrEqu
// PURPOSE: Compare for exact equality between two strings.
//------------------------------------------------------------------------------
int StrEqu( char *s1, char *s2 ) {

        while ( *s1 && *s2 && (*s1 == *s2) )

                { s1++; s2++; }

        return ( (*s1 == '\0') && (*s2 == '\0') ? 1 : 0 );

}

//------------------------------------------------------------------------------
// ROUTINE: StrIsHex
// PURPOSE: Check a string for being a '0x' hex number.
//------------------------------------------------------------------------------
int StrIsHex( char *s ) {

        if ( *s == '0' ) { s++;

                if ( *s == 'x' || *s == 'X' ) { s++;

                        while ( isxdigit( *s) )
                                s++;

                        if ( *s == '\0' )

                                return( 1);

                        }
                }

        return ( 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: CleanExit
// PURPOSE: Leave the program, while ensuring the keyboard mode is restored.
//------------------------------------------------------------------------------
int CleanExit( int ErrLevel ) {
        
        SrKeyboardMode( KEYMODE_BLOCK );  // Restore blocking keyboard behavior.

        exit( ErrLevel );
}
