// FILE: SetDid.c
// COPYRIGHT: (c), Symmetric Research, 2010
//
// This program is used to set the USBxCH VID and PID number to the
// proper SR values and then check them.
//
//

#include <stdio.h>
#include <stdlib.h>


#if defined( SROS_WIN2K ) || defined( SROS_WINXP )
#include <windows.h>     // windef.h -> winnt.h -> defines GENERIC_READ etc

#elif defined( SROS_LINUX )
#include <string.h>      // for strncopy, strncmp, strcmp

#endif // SROS_xxx


#include "SrDefines.h"   // for function type, dev handle, 64 bit, + other defines
#include "SrUsbXch.h"    // for SrUsbXch library functions
#include "SrUsbXchDrv.h" // for SrUsbXchDrv defines and structures like USB_DEVICE_DESCRIPTOR
#include "SrHelper.h"    // for os independent mkdir, keyboard support ...


#if defined( SROS_WIN2K ) || defined( SROS_WINXP )

#include <PSHPACK1.H> // required for close packing of USB structures
struct _USB_DEVICE_DESCRIPTOR {
        unsigned char bLength;
        unsigned char bDescriptorType;
        unsigned short bcdUSB;
        unsigned char bDeviceClass;
        unsigned char bDeviceSubClass;
        unsigned char bDeviceProtocol;
        unsigned char bMaxPacketSize0;
        unsigned short idVendor;
        unsigned short idProduct;
        unsigned short bcdDevice;
        unsigned char iManufacturer;
        unsigned char iProduct;
        unsigned char iSerialNumber;
        unsigned char bNumConfigurations;
};
#include <POPPACK.H> // Undoes PSHPACK1.H

typedef struct _USB_DEVICE_DESCRIPTOR         USB_DEVICE_DESCRIPTOR;

#else // if defined( SROS_LINUX )
#include <linux/types.h>        /* __u8 etc */

struct usb_device_descriptor {
        __u8  bLength;
        __u8  bDescriptorType;

        __le16 bcdUSB;
        __u8  bDeviceClass;
        __u8  bDeviceSubClass;
        __u8  bDeviceProtocol;
        __u8  bMaxPacketSize0;
        __le16 idVendor;
        __le16 idProduct;
        __le16 bcdDevice;
        __u8  iManufacturer;
        __u8  iProduct;
        __u8  iSerialNumber;
        __u8  bNumConfigurations;
} __attribute__ ((packed));

typedef struct usb_device_descriptor         USB_DEVICE_DESCRIPTOR;

#endif // SROS_xxx


// Program copyright and screen title:

char *Copyright =

        "\n\nSetDid.exe : Copyright (c), Symmetric Research, 2010\n\n";

char *ScreenTitle =

  "USBxCH SetDid Setting Program: (c) SR 2010 : Rev {2010/09/05}\n\n";


#define MAXSTR        256  // size for all strings



// This MUST match CMD_EEPROM_SET_DID in SrUsbXch.c and agree with the 8051 code

#define USB8051_CMD_EEPROM_SET_DID  0x30 


#define CMD_NONE        0
#define CMD_SHOW        1
#define CMD_SET         2
#define CMD_MAX         3

#define DID_UNUSED     -1

#define MAX_RESET_OPEN_TRIES    50



// SetDid RUN TIME PARAMETERS:
//
// Both command line and fixed hardwired parameters are located here.
//    ... access this structure with statments like: DP.Cmd ...
//
//

struct SetDidParametersLayout {

        int  Cmd;
        int  DidValue;
        DEVHANDLE UsbXchHandle;
        
        int  DidInUse[MAX_DEVICES];
        char AtodDriverName[MAXSTR];
        char AtodDriverNameOld[MAXSTR];
        char MsgBuffer[MAXSTR];

        } SetDidParameters = {
        
                // Initial values:

                CMD_NONE,                    // Command show/set, default = none
                0,                           // DID value,        default = 0
                BAD_DEVHANDLE,               // USBxCH handle,    default = bad
                
                };


#define DP  SetDidParameters  // < DP = shorthand for SetDidParameters ...



// USAGE MESSAGE ARRAY:

char *UsageMsg[] = {

    "Change the USB DID (device id) on a USBxCH board                           ",
    "                                                                           ",
    "USAGE: SetDid [driver] [0-9] [showall] [/?]                                ",
    "                                                                           ",
    "  driver = SrUsbXch(DID), where DID # is 0 to 9     << default = SrUsbXch0 ",
    "     0-9 = new DID # for the physical board EEPROM                         ",
    " showall = scan for all currently installed drivers                        ",
    "      /? = show help screen and quit                                       ",
    "                                                                           ",
    " Examples:                                                                 ",
    "                                                                           ",
    "    SetDid /?               < show help screen and quit                    ",
    "    SetDid 1                < change SrUsbXch0 board to SrUsbXch1          ",
    "    SetDid SrUsbXch1 2      < change SrUsbXch1 to SrUsbXch2                ",
    "    SetDid showall          < scan and list all SrUsbXch                   ",
    "                                                                           ",

    NULL
    };

extern long SrUsbXchLastDriverError;

// Prototypes

   int CommandShow( void );
   int CommandSet( void );

   int DoCheckActive( void );
   int DoSetDid( DEVHANDLE UsbXchHandle );
   int DoReset( DEVHANDLE UsbXchHandle );
   int DoVerifyDid( DEVHANDLE UsbXchHandle );

  void OpenUsbXch( void );
  void CloseUsbXch( void );
   
  void ProcessCommandLine( int argc, char *argv[] );
  void Usage( void );
   int StrEqu( char *s1, char *s2 );
  void ErrorCmd( char *msg );
  void ErrorExec( char *msg );
  int CleanExit( int ErrLevel );


// Prototypes from SrUsbXch.c

   int SrUsbXchOsDriverIoctl( DEVHANDLE UsbXchHandle,
                           unsigned long IoCtlCode,
                           void *pValueIn, unsigned long InSize,
                           void *pValueOut, unsigned long  OutSize,
                           unsigned long *ReturnSize );

//------------------------------------------------------------------------------
// ROUTINE: main()
// PURPOSE: SetDid main entry point ...
//------------------------------------------------------------------------------
int main( int argc, char *argv[] ) {

         int i;


        // Announce the program.

        printf( ScreenTitle );


        // Set keyboard behavior to allow for non blocked key checks.

        SrKeyboardMode( KEYMODE_NONBLOCK );


        // Process the command line arguments.

        ProcessCommandLine( argc, argv );


        // Initialize values

        for ( i = 0 ; i < MAX_DEVICES ; i++ )
                DP.DidInUse[i] = DID_UNUSED;



        // Execute selected command

        if ( DP.Cmd == CMD_SHOW )

                CommandShow( );

        else if ( DP.Cmd == CMD_SET ) 
                        
                CommandSet( );

        else 
                printf( "Unknown command %d ...\n\n", DP.Cmd );

        
        CleanExit( 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: CommandShow
// PURPOSE: Display a list of available SrUsbXch# drivers.
//------------------------------------------------------------------------------
int CommandShow( void ) {

        int i, cnt, Nactive;

        printf( "Showing available drivers ...\n\n" );
        

        // Check for active drivers

        Nactive = DoCheckActive( );


        // Print appropriate title line

        if ( Nactive == 0 )
                printf( "No drivers available\n" );
        else
                printf( "  Count   Driver Name\n" );


        cnt = 0;
        
        for ( i = 0 ; i < MAX_DEVICES ; i++ ) {
                if ( DP.DidInUse[i] != DID_UNUSED ) {
                        printf( "    %d      %s%d\n",
				  cnt, SRUSBXCH_DRIVER_NAME, DP.DidInUse[i] );
                        cnt++;
                        }
                }


        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: CommandSet
// PURPOSE: Write DID to USBxCH EEPROM.
//------------------------------------------------------------------------------
int CommandSet( void ) {

        int i, Result, Nactive;


        printf( "Setting DID to %d ...\n\n", DP.DidValue );


        // Verify requested DID is not active or in use

        printf( "Checking currently active drivers ...\n\n" );
        
        Nactive = DoCheckActive( );

        for ( i = 0 ; i < MAX_DEVICES ; i++ ) {
                if ( DP.DidInUse[i] == DP.DidValue ) {
                        printf( "FAILURE:  Requested DID of %d is already in use\n", DP.DidValue );
                        return( 0 );
                        }
                }


        
        printf( "Opening driver %s ...\n\n", DP.AtodDriverName );
        
        OpenUsbXch( );

        
        // Set the DID
        
        Result = DoSetDid( DP.UsbXchHandle );
        if ( !Result )
                ErrorExec( "Could not set DID\n" );



        // Reset board

        printf( "Cycling board power so new DID becomes active ...\n" );
        
        Result = DoReset( DP.UsbXchHandle );
        if ( !Result )
                ErrorExec( "Could not reset board\n" );



        // Check DID

        printf( "Verifying DID from device descriptor  ...\n\n" );

        Result = DoVerifyDid( DP.UsbXchHandle );
        if ( !Result )
                ErrorExec( "Could not verify DID\n" );


        CloseUsbXch();


        printf( "SUCCESS:  DID now is %d.  Use driver name %s.\n",
                  DP.DidValue, DP.AtodDriverName );
        
        return( Result );
}


//------------------------------------------------------------------------------
// ROUTINE: DoCheckActive
// PURPOSE: Check which SrUsbXch# drivers are active by trying to open them.
//          It first tries a minimal open which verifies a driver of that name
//          exists.  If that open succeeds, it procees on to do a normal open
//          which verifies the USB port, board, and power are all good and shows 
//          the driver is ready for use.  Returns the number of active drivers
//          found (0 to 10).
//------------------------------------------------------------------------------
int DoCheckActive( void ) {

        char      DriverName[MAXSTR];
        int       i, Count, Error;
        double    ActualSps;
        DEVHANDLE UsbXchHandle;


        // Initialize count variable

        Count = 0;


        // Loop over all possible driver names
        
        for ( i = 0 ; i < MAX_DEVICES ; i++ ) {

                // Prepare driver name

                sprintf( DriverName, "%s%d", SRUSBXCH_DRIVER_NAME, i );
                printf( "   %s ... ", DriverName );  SrFlushPrint();


                // Test driver for simple open

                UsbXchHandle = SrUsbXchMinimalOpen( DriverName, &Error );


                // If simple open is successful

                if ( ( SrUsbXchIsOpenSuccess( Error ) ) &&
                     ( UsbXchHandle != BAD_DEVHANDLE )   ) {

                        // Close driver and add to active list

                        SrUsbXchClose( UsbXchHandle );
                        DP.DidInUse[i] = i;
                        Count++;


                        // Now test driver for normal open 

                        UsbXchHandle = SrUsbXchOpen( DriverName,
                                                     SRDAT_ATODMODEL_USB4CH,
                                                     SRDAT_GPSMODEL_PCTIME,
                                                     USERCFG_DEFAULT_BYTE,
                                                     USBXCH_SPS_DEFAULT,

                                                     &ActualSps,
                                                     &Error
                                                    );

                        // If normal open is successful

                        if ( ( SrUsbXchIsOpenSuccess( Error ) ) &&
                             ( UsbXchHandle != BAD_DEVHANDLE )   ) {

                                // Close driver and announce success

                                SrUsbXchClose( UsbXchHandle );
                                printf( "Working\n" );
                                }

                        // Else normal open failed even though simple was successful

                        else
                                printf( "Available but has a PROBLEM\n" );

                        } // end if simple open successful


                // Else simple open failed, so list as unused

                else {
                        DP.DidInUse[i] = DID_UNUSED;
                        printf( "not installed\n" );
                        }
                
                } // end for i < MAX_DEVICES (first loop)

        printf( "\n" );

        return( Count );
}

//------------------------------------------------------------------------------
// ROUTINE: DoSetDid
// PURPOSE: Write DID to USBxCH EEPROM.
//------------------------------------------------------------------------------
int DoSetDid( DEVHANDLE UsbXchHandle ) {

        unsigned long BytesSent, Ncmd;
        int           UsbError;
        char          Values[3];

        // Set up command values including data bytes and send to USBxCH

        Values[0] = USB8051_CMD_EEPROM_SET_DID;
        Values[1] =  (DP.DidValue       & 0x00FF); // LSB (DID lo)
        Values[2] = ((DP.DidValue >> 8) & 0x00FF); // MSB (DID hi)
        Ncmd = 3L;
        BytesSent = SrUsbXchSendCommand( UsbXchHandle, Values, Ncmd, &UsbError );

        if ( UsbError == USBXCH_ERROR_NONE )
                return( 1 );
        else
                return( 0 );
}


//------------------------------------------------------------------------------
// ROUTINE: DoReset
// PURPOSE: Reset the board by cycling its power.  After the reset, the board
//          should have a new DID and therefore require a new name when
//          accessing it.
//------------------------------------------------------------------------------
int DoReset( DEVHANDLE UsbXchHandle ) {

        int Trys;


        // Reset the board
        
        DP.UsbXchHandle = SrUsbXchPowerReset( UsbXchHandle );


        
        // Save old driver name and make new one

        strcpy( DP.AtodDriverNameOld, DP.AtodDriverName );
        sprintf( DP.AtodDriverName, "%s%d", SRUSBXCH_DRIVER_NAME, DP.DidValue );



        // Wait for reset to finish and driver to reload

        printf( "  Waiting a few seconds for driver reload (%s to abort)\n  ", SR_QENTER );

        Trys = 0;

        while ( !SrUsbXchIsResetComplete( DP.AtodDriverName ) ) {

                // Allow 'q' to abort

                if ( SrGetKeyCheck() == 'q' ) {
                        printf( "\nFAILURE:  Reset stopped by keypress\n" );
                        return( 0 );
                        }

                // Cause abort after too many tries

                else if ( Trys > MAX_RESET_OPEN_TRIES ) {
                        printf( "\nFAILURE:  Reset stopped after %d tries to full open driver\n", Trys );
                        return( 0 );
                        }

                // Otherwise wait a little longer

                else {
                        printf( "." ); SrFlushPrint();
                        SrSleep( 500 );       // half second
                        Trys++;
                        }

                } // end while

        printf( "\n\n" );


        // Now that it is ready again, open UsbXch driver
        
        OpenUsbXch( );


        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: DoVerifyDid
// PURPOSE: Read the device descriptor and verify the DID is set correctly.
//------------------------------------------------------------------------------
int DoVerifyDid( DEVHANDLE UsbXchHandle ) {

        USB_DEVICE_DESCRIPTOR DeviceDescriptor;
        int                   Result;
        BOOL                  Ok;
        unsigned long         BytesSent;


        // Get device descriptor
        
        Ok = SrUsbXchOsDriverIoctl( UsbXchHandle,
                                    IOCTL_USBXCH_GET_DEVICE_DESCRIPTOR,
                                    NULL,
                                    0,
                                    &DeviceDescriptor,
                                    sizeof( USB_DEVICE_DESCRIPTOR ),
                                    &BytesSent );

        if ( Ok  &&  (DeviceDescriptor.bcdDevice == DP.DidValue) )
                Result = 1;
        else
                Result = 0;

        return( Result );
}

//------------------------------------------------------------------------------
// ROUTINE: OpenUsbXch
// PURPOSE: 
//------------------------------------------------------------------------------
void OpenUsbXch( void ) {

        int    Error;
        double ActualSps;


        // Open USBXCH driver

        DP.UsbXchHandle = SrUsbXchOpen(
                                       DP.AtodDriverName,
                                       SRDAT_ATODMODEL_USB4CH,  // AtodXchModel
                                       SRDAT_GPSMODEL_PCTIME,   // AtodGpsModel
                                       USERCFG_DEFAULT_BYTE,    // AtodUserCfg
                                       USBXCH_SPS_DEFAULT,      // Sampling rate

                                       &ActualSps,
                                       &Error
                                      );
        
        if ( ( DP.UsbXchHandle == BAD_DEVHANDLE ) ||
             ( !SrUsbXchIsOpenSuccess( Error ) )   ) {
                sprintf( DP.MsgBuffer,
                         "Could not open USB driver %s.\n          %s (Err %d)",
                         DP.AtodDriverName, USBXCH_ERROR_MSG[Error], Error );
                ErrorExec( DP.MsgBuffer );
                }
}

//------------------------------------------------------------------------------
// ROUTINE: CloseUsbXch
// PURPOSE: 
//------------------------------------------------------------------------------
void CloseUsbXch( void ) {

        // Close USBXCH driver

        if ( !SrUsbXchClose( DP.UsbXchHandle ) ) {
                printf( "FAILURE:  Could not close USB driver %s\n\n", DP.AtodDriverName );
                CleanExit( 1 );
                }

        DP.UsbXchHandle = BAD_DEVHANDLE;
}


//------------------------------------------------------------------------------
// ROUTINE: ProcessCommandLine
// PURPOSE: Gather the command line arguments, filling in PP.
//------------------------------------------------------------------------------
void ProcessCommandLine( int argc, char *argv[] ) {

        int  i, NewDid, DriverNum, DriverNameFound;
        char MsgBuffer[MAXSTR];


        // Set defaults

        SrStringCopy( DP.AtodDriverName, USBXCH_DEFAULT_DRIVER_NAME, MAXSTR );
        DriverNameFound = 0;




        // No command line arguments gives the usage message.

        if ( argc == 1 )

                Usage( );


        // Check for too many arguments.

        if ( argc > 3 )

                ErrorCmd( "too many command line arguments" );


        // Scan for valid arguments.

        for ( i = 1 ; i < argc ; i++ ) {


                // Check for usage message request.

                if ( StrEqu( argv[i], "/?") )

                        Usage( );


                // Check for driver number


                else if ( SrUsbXchIsValidDriverName( argv[i], &DriverNum ) ) {

                        SrStringCopy( DP.AtodDriverName, argv[i], MAXSTR );

                        if ( !DriverNameFound )
                                DriverNameFound = 1;
                        else
                                ErrorCmd( "multiple driver arguments" );
                        }


                // Check for the showall command.

                else if ( StrEqu( argv[i], "showall" ) ) {

                        if ( DP.Cmd == CMD_NONE )
                                DP.Cmd = CMD_SHOW;
                        else
                                ErrorCmd( "multiple command arguments" );
                        }


                // Check for numeric argument ("set" command)

                else if ( sscanf( argv[i], "%d", &NewDid ) == 1 ) {

                        if ( NewDid < 0 || NewDid > 9 )
                                ErrorCmd( "invalid numeric argument, new DID must be 0 to 9" );

                        if ( DP.Cmd == CMD_NONE ) {
                                DP.Cmd = CMD_SET;
                                DP.DidValue = NewDid;
                                }
                        else
                                ErrorCmd( "multiple command arguments" );
                        }


                // Unknown command

                else {
                        sprintf( MsgBuffer, "unknown command %s given", argv[i] );
                        ErrorCmd( MsgBuffer );
                        }
                
                }


        if ( DP.Cmd == CMD_NONE )
                ErrorCmd( "at least one command is required" );
}

//------------------------------------------------------------------------------
// ROUTINE: Usage
// PURPOSE: Display the usage message.
//------------------------------------------------------------------------------
void Usage( void ) {

        int i;


        // Print the usage message.

        for ( i = 0 ; UsageMsg[i] ; i++ )

                printf( "%s\n", UsageMsg[i] );

        CleanExit( 0 );  // < quit at errorlevel 0.
}

//------------------------------------------------------------------------------
// ROUTINE: StrEqu
// PURPOSE: Compare for exact equality between two strings.
//------------------------------------------------------------------------------
int StrEqu( char *s1, char *s2 ) {

        while ( *s1 && *s2 && (*s1 == *s2) )

                { s1++; s2++; }

        return ( (*s1 == '\0') && (*s2 == '\0') ? 1 : 0 );

}

//------------------------------------------------------------------------------
// ROUTINE: ErrorCmd
// PURPOSE: Print a command line error message and quit.
//------------------------------------------------------------------------------
void ErrorCmd( char *msg ) {

        // Print a command line error message and quit.

        printf(

               "ERROR: Command line usage : %s \n\n ... SetDid /? for help\n",
               msg

               );

        CleanExit( 1 );  // < quit at errorlevel 1.
}

//------------------------------------------------------------------------------
// ROUTINE: ErrorExec
// PURPOSE: A fatal execution error occured.
//------------------------------------------------------------------------------
void ErrorExec( char *msg ) {

        // Close any open driver

        if ( DP.UsbXchHandle != BAD_DEVHANDLE )
                CloseUsbXch();


        // Announce error and exit

        printf( "\nFAILURE:  %s \n\n", msg );

        CleanExit( 2 );  // < quit at errorlevel 2.
}

//------------------------------------------------------------------------------
// ROUTINE: CleanExit
// PURPOSE: Leave the program, while ensuring the keyboard mode is restored.
//------------------------------------------------------------------------------
int CleanExit( int ErrLevel ) {

        SrKeyboardMode( KEYMODE_BLOCK );  // Restore blocking keyboard behavior.

        exit( ErrLevel );
}
