// FILE: Blast.c
// COPYRIGHT: (c), Symmetric Research, 2009-2010
//
// Blast is a USBxCH command line acquisition program using minimum PC
// CPU and disk resources.  It saves its data to disk without any on
// screen display of the data.  Blast achieves the highest sampling
// rates possible with the USBxCH on any given computer.
//


// Include files:

#if defined( SROS_WINXP ) || defined( SROS_WIN2K )
#include <windows.h>
#include <io.h>         // for low level file disk read/write functions
#include <direct.h>     // for getcwd function
#define OsOpen   _open
#define OsWrite  _write
#define OsClose  _close
#define OsGetCWD _getcwd

#elif defined( SROS_LINUX )
#include <unistd.h>      // for access, close function and STDIN_FILENO
#define OsOpen   open
#define OsWrite  write
#define OsClose  close
#define OsGetCWD getcwd
#endif  // SROS_xxxxx

#include <stdio.h>
#include <stdlib.h>
#include <time.h>       // for time() and localtime() functions
#include <fcntl.h>      // for _O file open constants
#include <sys/stat.h>   // for _S file share constants
#include <ctype.h>      // for isdigit(), isxdigit()

#include "SrUsbXch.h"   // for USBxCH user library functions ...
#include "SrHelper.h"   // for os independent mkdir, keyboard support ...





// Program copyright and screen title:

char *Copyright =

          "\n\nBlast.exe : Copyright (c), Symmetric Research, 2009-2010\n\n";

char *ScreenTitle =

      "USBxCH Blast Acquisition Program: (c) SR 2010 : Rev {2010/08/15}\n\n";



#define MAXSTR  256    // << dimension used for most string buffers


// BLAST RUN TIME PARAMETERS:
//
// Both the user specified command line parameters and fixed hardwired
// parameters are allocated here.  Access this structure with statments
// like: BP.nFiles.
//
// The sampling rate is specified with a "sn" string like s5" on the
// command line.  pRateTable -> selected value in RateTable[].
//
// The default gn parameter is given at the bottom of the
// ProcessCommandLine function.

struct BlastParametersLayout {

        struct RateTableRecord *pRateTable;
        struct TimeTableRecord *pTimeTable;

         int nFiles;
         int KeyPressToRun;
         int UserCfgByte;

         int AtodXchModel;
        char AtodDriverName[MAXSTR];

        } BlastParameters = {


                // default command line parameter values:
                //
                //          nFiles = 0      -> output files forever
                //   KeyPressToRun = 1      -> keypress to start
                //     UserCfgByte = 0x00   -> standard GPS config
                //              gn = g1     -> PC Time
                //
                //    ( see SrUsbXch.h for UC default define )
                //
                //

                NULL,                   // < "sn" required, no default ...
                NULL,                   // optional "gn"
                0,                      // default nFiles
                1,                      // default KeyPressToRun
                USERCFG_DEFAULT_BYTE,   // default UserCfgByte



                // fixed driver name, GPS model, etc:
                //
                //   ( see SrUsbXch.h for MODEL define )
                //
                //

                SRDAT_ATODMODEL_USB4CH,   // fixed AtodXchModel
                
                // variable AtodDriverName


                };

#define BP  BlastParameters  // BP = shorthand for blast parameter struct





// SAMPLING RATE TABLE AND FILE SIZES:
//
// These are the allowed sampling rates for Blast, along with the file
// sizes that will be created at each rate.  This table maps the "sn"
// command line rate into the actual values that will be used.
//
// FilePackets controls the file size.  The corresponding FileBytes and
// FileSeconds are computed and filled in at run time.
//
// The FilePackets values have been selected so output files are created
// every few seconds ...
//
//

struct RateTableRecord {

        char  *SnString;       // < allowed command line "sn" string
        double SamplingRate;   // < SPS rate
        int    FilePackets;    // < sets file size at the SPS rate

        int    FileBytes;      // < computed, 64 bytes/packet
        int    FileSeconds;    // < computed from SPS and files size

        } RateTable[] = {

                // always keep FilePackets < MAXPACKET ...

                { "s1",      19.531250,     47,   0,   0  },
                { "s2",      32.552083,     64,   0,   0  },
                { "s3",      39.062500,     78,   0,   0  },
                { "s4",      65.104167,     97,   0,   0  },
                { "s5",      78.125000,    117,   0,   0  },
                { "s6",     130.208333,    130,   0,   0  },
                { "s7",     651.041667,    651,   0,   0  },
                { "s8",    1302.083333,   1302,   0,   0  },
                { "s9",    2604.166667,   1302,   0,   0  },
                { "s10",   4882.812500,   2441,   0,   0  },
                { "s11",   9765.625000,   4882,   0,   0  },

                // end of table ...

                {  NULL,             0,      0,   0,   0  },

                };


#define RR  struct RateTableRecord  // RR = shorthand for sampling rate record





// TIME SOURCE TABLE:
//
// These are the possible time sources for Blast: PC Time, GPS Time ...
//
//

struct TimeTableRecord {

        char  *GnString;        // < allowed command line "gn" string
        int   AtodGpsModel;     // < GPS model, NONE, PCTIME, GARMIN etc
        char  *GpsModelName;    // < GPS model name
        char  *UsageDesc;       // < usage message description

        } TimeTable[] = {

          { "g0",  SRDAT_GPSMODEL_NONE   , "NONE"   , "NONE                         " },
          { "g1",  SRDAT_GPSMODEL_PCTIME , "PCTIME" , "PCTIME   << default          " },
          { "g2",  SRDAT_GPSMODEL_GARMIN , "GARMIN" , "GARMIN   ( NMEA = RMC, GGA ) " },
          { "g3",  SRDAT_GPSMODEL_TRIMBLE, "TRIMBLE", "TRIMBLE  ( NMEA = ZDA, GGA ) " },

          // end of table ...

          { NULL, 0, NULL, NULL },

          };


#define TR  struct TimeTableRecord  // TR = shorthand for time source record






// USAGE MESSAGE ARRAY:

char *UsageMsg[] = {

      "Acquires USBxCH data and saves it to disk as .pak files              "
      "                                                                     ",
      "USAGE: Blast sn [gn] [nFiles] [nokeypress] [0xUC] [driver] [/?]      ",
      "                                                                     ",
      "          sn = sampling rate (required)                              ",
      "          gn = time source                                           ",
      "                                                                     ",
      "      nFiles = # files to save as a decimal number and then quit     ",
      "               must be > 0, nothing = run forever ...                ",
      "                                                                     ",
      "  nokeypress = turn off keypress to run/start                        ",
      "        0xUC = User Config Byte as 0x.. hex number                   ",
      "      driver = SrUsbXch# where # is 0 to 9, default SrUsbXch0        ",
      "          /? = show help screen and quit                             ",
      "                                                                     ",
      "                                                                     ",
      "  Sampling rate table: (64 bytes/packet)                             ",
      "                                                                     ",
      "    sn   ->       SPS(Hz)      PACKETS/FILE     BYTES    ~ SECONDS   ",
      "                                                                     ",
//    "    SSS      NNNNN.NNNNNN        DDDDD        DDDDDDD         D      ",
      NULL,
      "    %-3s      %12.6lf        %5d        %7d         %1d      ",

      "                                                                     ",
      "                                                                     ",
      "  Time source table: gn selects the time source ...                  ",
      "                                                                     ",
//    "    GG   ->  DESC                                                    ",
      NULL,
      "    %s   ->  %s        ",

      "                                                                     ",
      "                                                                     ",
      "Examples:                                                            ",
      "                                                                     ",
      "    Blast /?             < show help screen and quit                 ",
      "    Blast s3             < acquire forever at 39 Hz, keypress aborts ",
      "    Blast s3 16          < acquire 16 files at 39 Hz                 ",
      "    Blast s5 0x14        < acquire at 78 Hz with UC = 0x14           ",
      "    Blast s8 nokeypress  < acquire with no keypress to start         ",
      "                                                                     ",

      NULL
      };



// LARGE DATA BUFFER:
//
// Large memory buffer for USBxCH packets on their way out to disk.
//
// This buffer must be large enough to hold the entire number of pakets
// for a complete file at the highest sampling rate.  Dimension
// MAXPACKET > max( RateTable[i].FilePackets ) ...
//
//

#define MAXPACKET  5000
SRUSBXCH_PACKET AtodPacketBuffer[MAXPACKET];




// YMDHMS directory name buffer:
//
// Buffer for the YMDHMS output data directory name.
//
//

char YMDHMS[MAXSTR];




// PAK output files will be written using these open and share flags:
//
//

#if defined( SROS_WINXP ) || defined( SROS_WIN2K )
#define OFLAGS (_O_WRONLY|_O_CREAT|_O_TRUNC|_O_BINARY)
#define SFLAGS (_S_IWRITE)
#else
#define OFLAGS (O_WRONLY|O_CREAT|O_TRUNC)
#define SFLAGS (S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH)
#endif // SROS_xxx




// Prototypes:

   DEVHANDLE StartAcquisition( void );
        void PcTimeStamp( DEVHANDLE AtodHandle );
        void AcquireData( DEVHANDLE AtodHandle );
         int WriteData( void );
        void ToggleYellowLed( DEVHANDLE AtodHandle );
        void StopAcquisition( DEVHANDLE AtodHandle );

        void ComputeRateTable( void );
          RR *RateTableLookUp( char *s );
          TR *TimeTableLookUp( char *s );
        void ProcessCommandLine( int argc, char *argv[] );
        void MakeDirYmdHms( void );
        void PrintRunTimeSummary( void );
        void FormatRunTimeSummary( FILE *fp );
        void Usage( void );
        void PrintMsg( char *msg );
        void ErrorCmd( char *msg );
        void ErrorExec( char *msg );
         int StrEqu( char *s1, char *s2 );
         int StrIsDecimal( char *s );
         int StrIsHex( char *s );
         int CleanExit( int ErrLevel );





//------------------------------------------------------------------------------
// ROUTINE: main()
// PURPOSE: Blast main entry point ...
//------------------------------------------------------------------------------
int main( int argc, char *argv[] ) {

        DEVHANDLE AtodHandle; int c;

        char MsgBuffer[MAXSTR];



        // Announce the program.

        printf( ScreenTitle );


        // Set keyboard behavior to allow for non blocked key checks.

        SrKeyboardMode( KEYMODE_NONBLOCK );


        // Fill in the computed RateTable[] fields.

        ComputeRateTable();


        // Process the command line arguments.

        ProcessCommandLine( argc, argv );  // fills in BP ...


        // Create an output data directory from current YMDHMS.

        MakeDirYmdHms( );


        // Print out the final BP parameter summary before running.

        PrintRunTimeSummary( );




        // Open, initialize, and start the USBxCH.

        AtodHandle = StartAcquisition( /* with BP parameters */ );



        // Acquire and output data files:
        //
        // AcquireData reads data from the USBxCH, saving it in the
        // global array AtodPacketBuffer.  AcquireData does not return
        // until FilePackets have been read.
        //
        // WriteData writes the entire AtodPacketBuffer to the current
        // output file.
        //
        //

        while ( 1 ) { // infinite loop until break; ...



                // If using PCTIME, send a time stamp to the USBxCH.
                //
                //  ... if GPS ... all time stamping is down on the USBxCH
                //

                if ( BP.pTimeTable->AtodGpsModel == SRDAT_GPSMODEL_PCTIME )

                        PcTimeStamp( AtodHandle );



                // Acquire data from the USBxCH.

                AcquireData( AtodHandle );


                // Write the data to the hard disk.

                if ( !WriteData( ) )  // will fail at BP.nFiles ...
                        break;



                // Allow an early keypress abort ...

                if ( (c = SrGetKeyCheck()) != EOF ) {

                  // != EOF -> there is a keypress, so check it out ...

                        if ( c == 'q' || c == 'Q' ) {

                             PrintMsg( "\n ... Keypress 'q' abort ... \n");

                                break;

                                }

                        else {
                                sprintf( MsgBuffer, "To abort press %s ... \n", SR_QENTER );
                                PrintMsg( MsgBuffer );
                                }

                        }



                // Toggle the yellow led to show program is running.

                ToggleYellowLed( AtodHandle );


                }



        // Stop the USBxCH and close the driver.

        StopAcquisition( AtodHandle );




        // Restore blocking keyboard behavior.

        SrKeyboardMode( KEYMODE_BLOCK );
        

        // Return from main:
        //
        // Linux *requires* main to be of type integer with return 0 for
        // success.  Windows doesn't care ...
        //
        //

        return( 0 );

}


//------------------------------------------------------------------------------
// ROUTINE: StartAcquisition
// PURPOSE: Start the USBxCH running.
//------------------------------------------------------------------------------
DEVHANDLE StartAcquisition( void ) {

        DEVHANDLE AtodHandle;
        double    ActualSps;
        int       i, Error;

        char MsgBuffer[MAXSTR];

        
        
        // Open the USBxCH driver.

        AtodHandle = SrUsbXchOpen(

                                BP.AtodDriverName,
                                BP.AtodXchModel,
                                BP.pTimeTable->AtodGpsModel,
                                BP.UserCfgByte,
                                BP.pRateTable->SamplingRate,

                                &ActualSps,
                                &Error

                                );

        if ( !SrUsbXchIsOpenSuccess( Error ) ) {

                if ( Error == USBXCH_ERROR_DRIVER_NOT_OPEN ) {

                        for ( i = 0 ; UsbXchErrorMsgDetailDriverNotOpen[i] ; i++ )

                                PrintMsg( UsbXchErrorMsgDetailDriverNotOpen[i] );
                        }

                sprintf( MsgBuffer,

                         "SrUsbXchOpen error (%d): %s", Error, USBXCH_ERROR_MSG[Error] );

                        ErrorExec( MsgBuffer);
                }



        // Wait for a keypress to start.

        if ( BP.KeyPressToRun ) {

                sprintf( MsgBuffer, "Press ENTER to start acquisition, or %s to quit ... ", SR_QENTER );
                PrintMsg( MsgBuffer );
                SrFlushPrint();

                if ( SrGetKeyWait() != SR_ENTERKEY )   // quit if not ENTER ...
                        CleanExit( 1 );

                }


        // Start the USBxCH.

        PrintMsg( "Acquisition starting ... \n\n");

        if ( !SrUsbXchStart( AtodHandle, &Error ) ) {
                
                sprintf( MsgBuffer,

                         "SrUsbXchStart error (%d): %s", Error, USBXCH_ERROR_MSG[Error] );

                        ErrorExec( MsgBuffer);
                }



        // Initalize the red and yellow LEDs to off ...

        SrUsbXchUserLed( AtodHandle, 0, 0 );



        // Return the driver handle.

        return( AtodHandle );


}

//------------------------------------------------------------------------------
// ROUTINE: PcTimeStamp
// PURPOSE: Send a PC Time stamp and PPS down to the USBxCH.
//------------------------------------------------------------------------------
void PcTimeStamp( DEVHANDLE AtodHandle ) {

        int Error; char MsgBuffer[MAXSTR];


        // Send a PC time stamp down to the USBxCH.

        SrUsbXchSendTimeStamp( AtodHandle, &Error );

        if ( Error != USBXCH_ERROR_NONE ) {

            sprintf( MsgBuffer, "SrUsbXchSendTimeStamp error (%d): %s",
                     Error, USBXCH_ERROR_MSG[Error] );

                ErrorExec( MsgBuffer );

                }

}

//------------------------------------------------------------------------------
// ROUTINE: AcquireData
// PURPOSE: Read nFilePackets from the USBxCH.
//------------------------------------------------------------------------------
void AcquireData( DEVHANDLE AtodHandle ) {

        SRUSBXCH_PACKET *pBuffer;
          unsigned long nRequested;
          unsigned long nReturned;

        int Error;
   
        char MsgBuffer[MAXSTR];



        // Loop on GetData until FilePackets have been read.
        //
        //  nRequested = # of packets requested on each call to GetData
        //
        //

           pBuffer = AtodPacketBuffer;
        nRequested = BP.pRateTable->FilePackets;

        while ( nRequested > 0 ) {


                SrUsbXchGetDataAsPackets(

                        AtodHandle,
                        pBuffer,
                        nRequested,  // < # packets requested

                        &nReturned,  // < # packets returned
                        &Error

                        );

                if ( Error != USBXCH_ERROR_NONE ) {

                        sprintf( MsgBuffer,
                                 "SrUsbXchGetDataAsPackets error (%d): %s",
                                 Error, USBXCH_ERROR_MSG[Error] );

                            ErrorExec( MsgBuffer );
                        }



                   pBuffer += nReturned;
                nRequested -= nReturned;

                }


}

//------------------------------------------------------------------------------
// ROUTINE: WriteData
// PURPOSE: Write out acquired USB packets to disk.
//------------------------------------------------------------------------------
int WriteData( void ) {

        unsigned long  TotalBytes;
        char           DataFileName[MAXSTR], MsgBuffer[MAXSTR];
        int            fh;

        static int     FilesWritten = 0;

        // Get the output file size in bytes.

        TotalBytes = BP.pRateTable->FileBytes;


        // Form the YMDHMS/nnnnnnnn.pak sequential output file name.

        sprintf( DataFileName, "%s/%08d.pak", YMDHMS, FilesWritten );


        // Print a message with the file name.

        sprintf( MsgBuffer, "Writing file: %s ... \n", DataFileName );

                PrintMsg( MsgBuffer);


        // Open the output file.

        fh = OsOpen( DataFileName, OFLAGS, SFLAGS );

        if ( fh == -1 ) {

                sprintf( MsgBuffer, "Could not open file: %s", DataFileName );

                        ErrorExec( MsgBuffer);
                }


        // Write out the large AtodPacketBuffer array all at once.

        if ( OsWrite( fh, AtodPacketBuffer, TotalBytes ) != TotalBytes )

                ErrorExec( "WriteData failed to write out requested data" );


        // Close the output file.

        OsClose( fh );


        // Increment the number of files written.

        FilesWritten++;


        // Check if the nFiles limit has been reached.

        if ( BP.nFiles == 0 )  // < then writing files forever ...

                return( 1 );

        else if ( FilesWritten == BP.nFiles ) {

                sprintf( MsgBuffer,

                     "File limit reached: %d files written\n", FilesWritten );

                        PrintMsg( MsgBuffer);

                return( 0 );

                }

        else // ( FilesWritten < BP.nFiles ) so continue on ...

                return( 1 );

}

//------------------------------------------------------------------------------
// ROUTINE: ToggleYellowLed
// PURPOSE: Toggle the yellow LED.
//------------------------------------------------------------------------------
void ToggleYellowLed( DEVHANDLE AtodHandle ) {

        static int YellowLed = 0;

        YellowLed ^= 1;

        SrUsbXchUserLed( AtodHandle, YellowLed, /* red led > */ 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: StopAcquisition
// PURPOSE: Stop the USBxCH and close the driver.
//------------------------------------------------------------------------------
void StopAcquisition( DEVHANDLE AtodHandle ) {

         SrUsbXchStop( AtodHandle, NULL );
        SrUsbXchClose( AtodHandle );

}

//------------------------------------------------------------------------------
// ROUTINE: ComputeRateTable
// PURPOSE: Fill in the computed RateTable[] fields.
//------------------------------------------------------------------------------
void ComputeRateTable( void ) {

        RR *r;


        // Scan the rate table, filling in the computed fields.

        for ( r = RateTable ; r->SnString ; r++ ) {


                // Compute FileBytes.
                //
                // File size is controlled by the number of packets in
                // each output file, with 64 bytes per USB packet.
                //

                r->FileBytes = r->FilePackets * SRDAT_USBPACKET_SIZE;


                // Compute FileSeconds.
                //
                // Approximate number of seconds of data in each output
                // file.  This is an estimate, assuming all packets are
                // analog, ignoring some are GPS NMEA packets, etc.
                //
                // Analog packets have two A/D sample points per packet.
                // Round to the nearest integer.
                //

                r->FileSeconds =

                        (int)( ((r->FilePackets*2) / r->SamplingRate) + 0.5 );

        }


}

//------------------------------------------------------------------------------
// ROUTINE: RateTableLookUp
// PURPOSE: Compare s against the sampling rate table "sn" names.
//------------------------------------------------------------------------------
RR *RateTableLookUp( char *s ) {

        RR *r;


        // Scan the sampling rate table for a match with s.

        for ( r = RateTable ; r->SnString ; r++ ) {

                if ( StrEqu( s, r->SnString ) )

                        return( r );  // > success, got a match ...
                }


        return( NULL );  // > failure, hit end of table without a match ...


}

//------------------------------------------------------------------------------
// ROUTINE: TimeTableLookUp
// PURPOSE: Compare s against the time source table "gn" names.
//------------------------------------------------------------------------------
TR *TimeTableLookUp( char *s ) {

        TR *t;


        // Scan the time source table for a match with s.

        for ( t = TimeTable ; t->GnString ; t++ ) {

                if ( StrEqu( s, t->GnString ) )

                        return( t );  // > success, got a match ...
                }


        return( NULL );  // > failure, hit end of table without a match ...


}

//------------------------------------------------------------------------------
// ROUTINE: ProcessCommandLine
// PURPOSE: Gather the command line arguments, filling in BP.
//------------------------------------------------------------------------------
void ProcessCommandLine( int argc, char *argv[] ) {

        int i, nFiles, UserCfg, DriverNum, DriverNameFound;
        RR *r;
        TR *t;

        char MsgBuffer[MAXSTR];



        // Set defaults

        SrStringCopy( BP.AtodDriverName, USBXCH_DEFAULT_DRIVER_NAME, MAXSTR );
        DriverNameFound = 0;


        // Blast with no arguments at all gives the usage message.

        if ( argc == 1 )

                Usage( );


        // Check for too many arguments.

        if ( argc > 6 )

                ErrorCmd( "too many command line arguments" );



        // Scan for valid arguments.

        for ( i = 1 ; i < argc ; i++ ) {



                // Check for a "/?" usage message request.

                if ( StrEqu( argv[i], "/?") )

                        Usage( );



                // Check for a valid driver name

                else if ( SrUsbXchIsValidDriverName( argv[i], &DriverNum ) ) {
                        
                        SrStringCopy( BP.AtodDriverName, argv[i], MAXSTR );

                        if ( !DriverNameFound )
                                DriverNameFound = 1;

                        else
                               ErrorCmd( "multiple driver arguments" );
                        }


                // Check for a "nokeypress" to run request.

                else if ( StrEqu( argv[i], "nokeypress" ) ) {

                        if ( BP.KeyPressToRun == 1 )
                                BP.KeyPressToRun = 0;

                        else

                               ErrorCmd( "multiple \"nokeypress\" arguments" );
                        }



                // Check for a "sn" sampling rate.

                else if ( r = RateTableLookUp( argv[i] ) ) {

                        if ( BP.pRateTable == NULL )
                                BP.pRateTable = r;

                        else
                                ErrorCmd( "multiple \"sn\" sample rates" );

                        }



                // Check for a UserConfig byte.

                else if ( StrIsHex( argv[i] ) ) {

                        sscanf( argv[i], "0x%lX", &UserCfg );

                        if ( BP.UserCfgByte == 0x00 ) {

                                if ( UserCfg <= 0x0FF )
                                        BP.UserCfgByte = UserCfg;
                                else

                  ErrorCmd( "invalid User Config Byte, must be 0x00 to 0xFF" );

                                }

                        else
                                ErrorCmd( "multiple 0x.. User Config Bytes" );

                        }




                // Check for a "gn" time source.

                else if ( t = TimeTableLookUp( argv[i] ) ) {

                        if ( BP.pTimeTable == NULL )
                                BP.pTimeTable = t;

                        else
                                ErrorCmd( "multiple \"gn\" sample rates" );

                        }



                // Check for a nFiles number of output files.

                else if ( StrIsDecimal( argv[i] ) ) {

                        sscanf( argv[i], "%ld", &nFiles);

                        if ( BP.nFiles == 0 ) {

                                if ( nFiles > 0)
                                        BP.nFiles = nFiles;
                                else

                                    ErrorCmd( "invalid nFiles, must be > 0" );
                                }

                        else
                                ErrorCmd( "multiple nFiles specified");

                        }



                // Unknown argument specified.

                else {
                        sprintf( MsgBuffer, "unknown argument > %s", argv[i] );

                            ErrorCmd( MsgBuffer );
                        }


                }



        // Ensure at least the "sn" sampling rate argument was specified.

        if ( !BP.pRateTable )

                ErrorCmd( "\"sn\" sampling rate is required" );



        // Set the default "gn" time source if they did not specified one.

        if ( !BP.pTimeTable ) {

                if ( t = TimeTableLookUp( "g1" ) )

                        BP.pTimeTable = t;

                else
                        ErrorCmd( "\"gn\" time source default is bad" );
                }


}

//------------------------------------------------------------------------------
// ROUTINE: MakeDirYmdHms
// PURPOSE: Create the YMDHMS data directory.
//------------------------------------------------------------------------------
void MakeDirYmdHms( void ) {

            time_t UtcTimeInSecondsSince_1970;
        struct tm *LocTime;


        // Get the current time.

                       time( &UtcTimeInSecondsSince_1970 );
        LocTime = localtime( &UtcTimeInSecondsSince_1970 );


        // Prepare the directory name: YYYY-MM-DD-at-HH-MM-SS
        //
        //    YMDHMS = ( year, month, day, hours, minuts, seconds)
        //
        //

        sprintf( YMDHMS, "%04d-%02d-%02d-at-%02d-%02d-%02d",

                                        LocTime->tm_year+1900,
                                        LocTime->tm_mon+1,
                                        LocTime->tm_mday,
                                        LocTime->tm_hour,
                                        LocTime->tm_min,
                                        LocTime->tm_sec

                                        );

        // Make the directory.

        if ( SrMkDir( YMDHMS ) == -1 /* == failure */ )

                ErrorExec( "Unable to create YMDHMS output data directory" );


}


//------------------------------------------------------------------------------
// ROUTINE: PrintRunTimeSummary
// PURPOSE: Display the final BP parameters to run with.
//------------------------------------------------------------------------------

void PrintRunTimeSummary( void ) {

        char filename[MAXSTR];
        FILE *fp;


        // Display the BP parameters they are running with,
        //
        //     and also save a copy to a report file ...
        //
        //


        // screen display ...

        FormatRunTimeSummary( stdout );


        // also save a YMDHMS report file copy ...

        sprintf( filename, "%s/%s", YMDHMS, "000 Blast Run Summary.rpt" );

          fp = fopen( filename, "w");

          if ( !fp )

              ErrorExec( "unable to open \"000 Blast Run Summary.rpt\" " );


        FormatRunTimeSummary( fp );

        fclose( fp );


}

void FormatRunTimeSummary( FILE *fp ) {

        RR *r;
        TR *t;

        char *AtodXchModelName;
        char *CurrDir;

        
        // Set up table pointers
        
        r = BP.pRateTable;
        t = BP.pTimeTable;

        
        // Map AtodXchModel to string name ...

        AtodXchModelName = SrDatAtodModelString( BP.AtodXchModel );


        // Allocate space and return current directory name

        CurrDir = OsGetCWD( NULL, 0 );
        if ( !CurrDir )
                ErrorExec( "Could not allocate space for current directory name\n" );


        // Sampling rate and driver parameters.

        fprintf( fp, "  Blast runtime summary ... \n\n");

        fprintf( fp, "       Sampling Rate sn:  %s    ( %lf Hz )\n",  r->SnString, r->SamplingRate );
        fprintf( fp, "         Time Source gn:  %s    ( %s )\n",      t->GnString, t->GpsModelName );
        fprintf( fp, "            UserCfgByte:  0x%02X\n",            BP.UserCfgByte               );
        fprintf( fp, "             DriverName:  %s\n",                BP.AtodDriverName            );
        fprintf( fp, "               XchModel:  %s\n",                AtodXchModelName             );
        fprintf( fp, "\n");


        // Output file parameters.

        fprintf( fp, "  YMDHMS data directory:  %s\n",                YMDHMS                       );
        fprintf( fp, "             Created in:  %s\n",                CurrDir                      );


    if ( BP.nFiles )

        fprintf( fp, "                 nFiles:  %d\n",                BP.nFiles                    );

    else // BP.nFiles == 0 -> run forever

        fprintf( fp, "                 nFiles:  0 (run forever)\n"                                 );


        fprintf( fp, "            FilePackets:  %d \n",               r->FilePackets               );
        fprintf( fp, "              FileBytes:  %d \n",               r->FileBytes                 );
        fprintf( fp, "            FileSeconds:  %d \n",               r->FileSeconds               );
        fprintf( fp, "\n" );


        // Clean up allocated directory name space
        
        free( CurrDir );
}

//------------------------------------------------------------------------------
// ROUTINE: Usage
// PURPOSE: Display the usage message.
//------------------------------------------------------------------------------
void Usage( void ) {

        int i; RR *r; TR *t;



        // Print the top section of the usage message.

        for ( i = 0 ; UsageMsg[i] ; i++ )

                printf( "%s\n", UsageMsg[i] );



        // Print the allowed sampling rate table selections.

        for ( ++i , r = RateTable ; r->SnString ; r++ ) {

                printf( UsageMsg[i],  // < fixed format string

                          r->SnString,
                          r->SamplingRate,
                          r->FilePackets,
                          r->FileBytes,
                          r->FileSeconds

                          );

                printf( "\n" );

                }



        // Print the allowed time source table selections.

        for ( ++i ; UsageMsg[i] ; i++ )

                printf( "%s\n", UsageMsg[i] );


        for ( ++i , t = TimeTable ; t->GnString ; t++ ) {
        
                printf( UsageMsg[i],  // < fixed format string

                          t->GnString,
                          t->UsageDesc

                          );

                printf( "\n");

                }


        // Print the bottom section of the usage message.

        for ( ++i ; UsageMsg[i] ; i++ )

                printf( "%s\n", UsageMsg[i] );



        CleanExit( 0 );  // < quit at errorlevel 0.


}

//------------------------------------------------------------------------------
// ROUTINE: PrintMsg
// PURPOSE: Print a program status message and continue on.
//------------------------------------------------------------------------------
void PrintMsg( char *msg ) {

        // Print a program status message.

        printf( msg );

}

//------------------------------------------------------------------------------
// ROUTINE: ErrorCmd
// PURPOSE: Print a command line error message and quit.
//------------------------------------------------------------------------------
void ErrorCmd( char *msg ) {

        // Print a command line error message and quit.

        printf(

               "ERROR: Command line usage : %s \n\n ... Blast /? for help\n",
               msg

               );

        CleanExit( 1 );  // < quit at errorlevel 1.

}

//------------------------------------------------------------------------------
// ROUTINE: ErrorExec
// PURPOSE: An execution error occured.
//------------------------------------------------------------------------------
void ErrorExec( char *msg ) {

        // A fatal execution error occured.

        printf( "ERROR: %s \n\n", msg );

        CleanExit( 2 );  // < quit at errorlevel 2.

}

//------------------------------------------------------------------------------
// ROUTINE: StrEqu
// PURPOSE: Compare for exact equality between two strings.
//------------------------------------------------------------------------------
int StrEqu( char *s1, char *s2 ) {

        while ( *s1 && *s2 && (*s1 == *s2) )

                { s1++; s2++; }

        return ( (*s1 == '\0') && (*s2 == '\0') ? 1 : 0 );

}

//------------------------------------------------------------------------------
// ROUTINE: StrIsDecimal
// PURPOSE: Check a string for being a decimal number.
//------------------------------------------------------------------------------
int StrIsDecimal( char *s ) {

        while ( isdigit( *s ) )

                s++;

        return ( (*s == '\0') ? 1 : 0 );

}

//------------------------------------------------------------------------------
// ROUTINE: StrIsHex
// PURPOSE: Check a string for being a '0x' hex number.
//------------------------------------------------------------------------------
int StrIsHex( char *s ) {

        if ( *s == '0' ) { s++;

                if ( *s == 'x' || *s == 'X' ) { s++;

                        while ( isxdigit( *s) )
                                s++;

                        if ( *s == '\0' )

                                return( 1);

                        }
                }

        return ( 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: CleanExit
// PURPOSE: Leave the program, while ensuring the keyboard mode is restored.
//------------------------------------------------------------------------------
int CleanExit( int ErrLevel ) {
        
        SrKeyboardMode( KEYMODE_BLOCK );  // Restore blocking keyboard behavior.

        exit( ErrLevel );
}
